/** @file gsIETIAssembler.h

    @brief


    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): C. Hofer
    Created on: 2014-12-03
*/


#pragma once

#include <gsAssembler/gsAssemblerBase2.h>
//#include <gsAssembler/gsAssemblerBase.h>
#include <gsAssembler/gsAssemblerOptions.h>
#include <gsPde/gsBoundaryConditions.h>
#include <gsCore/gsBoxTopology.h>

#include <gsUtils/gsSortedVector.h>
#include <gsAssembler/gsIETIUtils.h>

namespace gismo {

template<typename T> class gsIETIJumpOperator;

/**
 * @brief gsIETIAssembler is the main assembling routine for the IETI method.
 * It assembles the stiffness matrices on each patch and calculates the LU factorization,
 * it also provides the bookkeeping for different splittings.

 */
template<typename T>
class gsIETIAssembler
{
public:
    /// \brief  the method used for facorizing the matrices
#if defined(GISMO_WITH_PARDISO)
    typedef typename gsSparseSolver<T>::PardisoLU sparseLUfact; //Threadsave
    typedef typename gsSparseSolver<T>::PardisoLLT sparseSPDfact; //Threadsave
#elif defined(_OPENMP)
    typedef typename gsSparseSolver<T>::LU sparseLUfact; //Threadsave
    typedef typename gsSparseSolver<T>::LU sparseSPDfact; //Threadsave
#elif(defined(GISMO_WITH_SUPERLU))
    typedef typename gsSparseSolver<T>::SuperLU sparseLUfact; //Not Threadsave
    typedef typename gsSparseSolver<T>::SuperLU sparseSPDfact; //Not Threadsave
#else
    typedef typename gsSparseSolver<T>::LU sparseLUfact; //Threadsave
    typedef typename gsSparseSolver<T>::LU sparseSPDfact; //Threadsave
#endif
    /// \brief  shortcut
    typedef std::pair<unsigned, unsigned> patchDof;

    friend class gsIETIJumpOperator<T>;
protected:
    /// \brief  shortcut for assembling matrices
    typedef Eigen::Triplet<T> Trip;

    /**
     * @brief getEdgesFromBoundary given a 3D patch side, this method calculates all 4 Edges to this side
     * @param side the given patchSide of a 3D cube
     * @param e the vector containing the 4 Edges
     */
    virtual void getEdgesFromBoundary(const patchSide& side ,std::vector<Edge>& e)
    {
        gsMatrix<unsigned>* bI = m_basis.front().basis(side.patch).boundary(side);
        gsBasis<T>* bbasis = m_basis.front().basis(side.patch).boundaryBasis(side);
        bool hasDirichlet;

        for(int i=1;i<5;i++)
        {
            boxSide side2D(i);
            gsMatrix<unsigned>* bbI = bbasis->boundary(side2D);

            int n = bbI->rows();
            hasDirichlet=false;

            if(n>2)
            {
                for(int j=1;j<n-1;j++)
                {
                    if(!m_locDofsMapper[side.patch][0].is_free((*bI)((*bbI)(j,0),0)))
                    {
                        hasDirichlet = true;
                        break;
                    }
                }
                e.push_back(Edge(hasDirichlet,n,orient(side2D.index()-1,side.index()), side, side2D));
            }

            delete bbI;
        }
        delete bbasis;
        delete bI;
    }


    /**
     * @brief compCalc calculates the index in a consecutive number of the
     *  dofs of a given index \a idx on patch \a np for a component \a c.
     *  e.g. the second dof of the third component on patch 1 results in 300th dof.
     *  (np = 0, idx = 1, c = 2 -> return 299)
     * @param np the patch number
     * @param idx the index of the dof of the given component
     * @param c the component
     * @return  the consecutive index
     */
    virtual inline unsigned compCalc(int np,unsigned idx, int c) const
    {
        return c*m_basis.front()[np].size() +idx;
    }

    /**
     * @brief compCalcBack given a consecutive dof number on a patch, the function
     * calculates the index w.r.t. to its component. In order to get the component
     * use getComp(..).
     * e.g. the 300th dof on patch 1 has index 2
     * (np =0, idx  = 299 -> return 1)
     * @param np the patch number
     * @param idx the consecutive index of the dof
     * @return  the index of the dof w.r.t. its component
     */
    virtual inline unsigned compCalcBack(int np,unsigned idx) const
    {
        return idx%m_basis.front()[np].size();
    }

    /**
     * @brief getComp given a consecutive dof number on a patch, the function calculates
     * its component. In order to get the index use compCalcBack(..).
     * e.g. the 300th dof on patch 1 has component 3
     * (np = 0, idy = 299 -> return 2)
     * @param np the patch number
     * @param idx the consecutive index of the dof
     * @return  the component of the dof
     */
    virtual inline int getComp(int np,unsigned idx) const
    {
        return idx/m_basis.front()[np].size();
    }

    /**
     * @brief setInfo Populates the entries of the gsIETIInfo structure. Needs to be called after
         init_patchInterfaceSides();
         init_primalDofs();
         init_boundaryDofs(isMinimalEnergy);
         init_jumpOperatorData();

        and called before
         init_remainingDofs();

     * @param isMinimalEnergy which algorithm is used
     */
    virtual void setInfo(bool isMinimalEnergy);


    /**
     * @brief init_mappers initializes the mappers for the bookkeeping. Needs to be called before all other init methods.
     */
    virtual void init_mappers();

    /**
     * @brief init_jumpOperatorData initializes the data for the jump operator B,
     * i.e. gsIETIJumpOperator. Needs to be called before setInfo(..);
     */
    virtual void init_jumpOperatorData();

    /**
     * @brief init_patchInterfaceSides Initializes the sides for each patch lieing on an interface.
     * Needs to be called before setInfo(..);
     */
    virtual void init_patchInterfaceSides();

    /**
     * @brief init_primalDofs Initializes the info for the primal variables, needs to be called after
     *       init_patchInterfaceSides();
     *
     *  and called before
     *      init_boundaryDofs(isMinimalEnergy);
     *      setInfo(bool isMinimalEnergy);
     *      init_remainingDofs();
     *
     *
     * Global Primal Dof are ordered in the following way:
     *
     *[c1,c2,c3,....,cN,f1,...,fM,e1,e2,...,eK]
     *
     * where ci are vertex values, fi are face averages and ei are edge averages
     *
     * in 2D there are no faces, just edges, but they are computed like the faces. (just interfaces)
     * (an edge is always a 2D object and a face always a 3D object, independent of the object dimension)
     *
     **/
    virtual void init_primalDofs();

    /**
     * @brief init_remainingDofs Initializes the splitting for primal and remaining dofs. This is only needed in case of
     *  minimumEnergy is false.
     *  Has to be called after all init_* functions and setInfo.
     */
    virtual void init_remainingDofs();

    /**
     * @brief init_boundaryDofs Initializes the splitting in boundary and remaining dofs. Has to be called after
     *    init_primalDofs()
     *  and called before
     *    setInfo(..)
     *    init_boundaryDofs();
     *
     * @param isMinimumEnergy which algorithm is used
     */
    virtual void init_boundaryDofs();


    /**
     * @brief assembleLUofKC assembles the LU factorization of the blockmatrix [K C^T;C 0].
     *
     * due to stability reasons for very jumping coefficients, we store the LU factorization of:
     * [ K , \alpha C^T; \alpha C^T, 0], hence, if you solve with a rhs, you should use:
     * [F1 ,\alpha F2].
     * @param tripletList a vector of triplets representing the matrix K, already ordered as
     *          [Kbb Kbi; Kib Kii].
     */
    virtual void assembleLUofKC(gsSparseMatrix<T>& matrix, std::size_t np);

    /**
     * @brief assembleKrrKrpKpp builds the submatrices Krp, Kpp from the given vector of Triplets and stores a LU factorization of Krr
     * @param tripletList the vector of triplets representing the matrix K, already ordered asn
     *          [Kpp Kpr; Krp Krr].
     */
    virtual void assembleKrrKrpKpp(gsSparseMatrix<T>& matrix, std::size_t np);

    /**
     * @brief assembleKiiKibKbb builds the submatrices  Kib, Kbb from the given vector of Triplets and stores a LU factorization of Kii.
     * Required for preconditioning for all cases and for processing the rhs and the solution in case of minimal energy.
     * @param tripletList a vector of triplets representing the matrix K, already ordered as
     *          [Kbb Kbi; Kib Kii].
     */
    virtual void assembleKiiKibKbb(gsSparseMatrix<T>& matrix, std::size_t np);

    /**
     * @brief assembleSpp Assembles the schur complement w.r.t. the primal variables.
     *  In case of minimum Energy, it is the sumbmatrix  w.r.t. the primal varibales of the schur complement S of K w.r.t. to b.
     *  In the other case it is calculated as Kpp - Kpr* Krr^-1 * Krp
     *
     *  Needs to be called after
     *      assembleC();
     *      assembleLUofKC(tripletList);
     *   or
     *      assembleKrrKrpKpp(tripletList);
     *
     */
    virtual void assembleSpp(std::vector<gsMatrix<T> >& spp_loc);
    virtual void assembleSppLoc(std::vector<gsMatrix<T> >& spp_loc, std::size_t np);

    /**
     * @brief assembleC assembes the constraint matrix C for the minimal Energy method, which is used to enforce the primal constraints
     */
    virtual void assembleC(std::size_t np);

    /**
     * @brief assembleRhs Assembles the rhs accoring to the desired splitting and/or transformation to the schur complement.
     *   Needs to be called after ALL assemble routines
     *
     * @param rhs_loc the patch-local contributions of the rhs
     */
    virtual void assembleRhs(const std::vector<gsMatrix<T> >& rhs_loc, std::size_t np);

    /**
     * @brief assembleRhsFreeElim Assembles the right hand side for the lagrange multipliers, which connect a eliminated and free dof
     */
    virtual void assembleRhsFreeElim();

    /**
     * @brief getInterfaceAverageValue This functions calculates for one patchSide of the parameter domain the interface average
     * for all nonzero basis function on this side
     * @param side The given patch side (2D or 3D)
     * @param average a vector of length boundaryBasis(side).size() ,containig for each entry the average.
     * @param c the component of the basis (for scalar problems c = 0)
     * Note that for basis on the boundary of side, the average is not computed, the corresponding entry is 0
     */
    virtual void getInterfaceAverageValue(const patchSide& side, gsMatrix<T>& average, int c = 0);

    /**
     * @brief getFaceEdgeAverageValue This functions calculates for one Edge of a 3D cube in the parameter domain the average
     * for all nonzero basis function on this edge. For Edge averages of a 2D domain see getInterfaceAverageValue(..).
     * @param edge the given edge (only for 3D)
     * @param average  a vector of length edge.number ,containig for each entry the average.
     * @param c the component of the basis (for scalar problems c = 0)
     * Note that for basis on the boundary of side, the average is not computed, the corresponding entry is 0
     */
    virtual void getFaceEdgeAverageValue(const Edge& edge,gsMatrix<T>& average, int c = 0,bool eliminateCorners=true);

    /**
     * @brief checkPrimalDofStrategy The function checks if the primal dof strategy is appropriate for the desired
     * problem and used algorithm. This function changes the member m_primalDofstrat.
     * @param isMinimalEnergy true if minimal energy algorithm is used
     * @param dim the dimension of the desired object
     */
    virtual void checkPrimalDofStrategy(bool isMinimalEnergy,int dim);

    /// \brief  populates the orient matrix
    virtual void InstantiateEdge()
    {
        orient.resize(4,7);

        orient.col(static_cast<int>(boundary::west))<< 4,6,1,9;
        orient.col(static_cast<int>(boundary::east))<< 5,7,2,10;
        orient.col(static_cast<int>(boundary::south))<< 4,5,0,8;
        orient.col(static_cast<int>(boundary::north))<< 6,7,3,11;
        orient.col(static_cast<int>(boundary::front))<<1,2,0,3;
        orient.col(static_cast<int>(boundary::back))<< 9,10,8,11;
    }

    virtual void makeReorderingBoundInt(gsSparseMatrix<T>& matrix, std::size_t np)
    {
        gsVector<index_t> permVector(matrix.cols());
        for(index_t i = 0; i< matrix.cols();i++)
        {
            permVector[i] = m_glob2BoundInteriorIndex[np][i];
            if(m_globIsBoundIndex[np][i] ==false)
                permVector[i]+= info.dofsB[np];//shift the interior ones
        }

        Eigen::PermutationMatrix<Dynamic,Dynamic,index_t> perm(permVector);
        matrix=matrix.twistedBy(perm);
    }
    virtual void makeReorderingPrimalRem(gsSparseMatrix<T>& matrix, std::size_t np)
    {
        gsVector<index_t> permVector(matrix.cols());
        for(index_t i = 0; i< matrix.cols();i++)
        {
            permVector[i] = m_glob2PrimalRemainingIndex[np][i];
            if(m_globIsPrimalIndex[np][i] ==false)
                permVector[i]+= info.dofsP[np];//shift the remaining ones
        }

        Eigen::PermutationMatrix<Dynamic,Dynamic,index_t> perm(permVector);
        matrix=matrix.twistedBy(perm);
    }


public:

    /**
     * @brief gsIETIAssembler The constructor of the IETI assembler
     * @param patches the multi patch representation
     * @param bases the basis corresponding to the patch
     * @param bconditions the boundary conditions
     * @param rhs the right hand side as gsFunction<T>
     * @param isMinimalEnergy should the minimum energy method be used (default is true)
     * @param strat_ the strategy for selecting the primal variables (default is primalDofMethod::all)
     * @param dirStrategy strategy for incormporating the dirichlet boundary (default is dirichlet::elimination)
     */
    gsIETIAssembler(gsAssemblerBase2<T> &assembler,
                    bool isMinimalEnergy = true,
                    primalDofMethod::strategy strat_ = primalDofMethod::B,
                    bool extraTimings = false);


    /// \brief  destructor
    virtual ~gsIETIAssembler()
    {
        freeAll(m_LU_KC);
        freeAll(m_LU_Krr);
        freeAll(m_LU_Kii);
        for(std::size_t np =0; np<info.numberPatches;np++)
        {
            std::vector<boundary_condition<T> > bc_all = m_bc_loc[np].allConditions();
            for(typename std::vector<boundary_condition<T> >::iterator it = bc_all.begin(); it!= bc_all.end();it++)
            {
                gsFunction<T>* ptr = (*it).function().get();
                delete ptr;
            }
        }

        delete m_LU_Spp;
    }

    /// \brief  returns the info structure
    virtual const gsIETIInfo & getInfo() const {return info;}

    /**
     * @brief solveKC solves [K ,C^T;C , 0]* sol = rhs. In order to obtain numerical stability, parts of the matrix
     * are scaled, this implies also a scaling of the rhs. Note, that the rhs is changed during the function call,
     * but the change is reverted at the end. So in fact, rhs stays unchanged.
     * @param np the actual patch
     * @param rhs the right hand side
     * @param sol the solution
     */
    virtual void solveKC(unsigned np, gsMatrix<T>& rhs, gsMatrix<T>& sol) const
    {
        GISMO_ASSERT(info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as true");
        sol.setZero(rhs.rows(),rhs.cols());

        if(rhs.rows()==0)
            return;

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
                rhs.col(col) = m_scalingsKC[np].LeftScaling().cwiseProduct(rhs.col(col));


        sol=m_LU_KC[np]->solve(rhs);

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
            {
                sol.col(col) =  m_scalingsKC[np].RightScaling().cwiseProduct(sol.col(col));
                rhs.col(col) = m_scalingsKC[np].LeftScaling().cwiseQuotient(rhs.col(col));
            }
    }

    /**
     * @brief solveKii solves Kii* sol = rhs. In order to obtain numerical stability, parts of the matrix
     * are scaled, this implies also a scaling of the rhs. Note, that the rhs is changed during the function call,
     * but the change is reverted at the end. So in fact, rhs stays unchanged.
     * @param np the actual patch
     * @param rhs the right hand side
     * @param sol the solution
     */
    virtual void solveKii(unsigned np, gsMatrix<T>& rhs, gsMatrix<T>& sol) const
    {
        sol.setZero(rhs.rows(),rhs.cols());
        if(rhs.rows()==0)
            return;

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
                rhs.col(col) = m_scalingsKii[np].LeftScaling().cwiseProduct(rhs.col(col));


        sol=m_LU_Kii[np]->solve(rhs);

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
            {
                sol.col(col) =  m_scalingsKii[np].RightScaling().cwiseProduct(sol.col(col));
                rhs.col(col) = m_scalingsKii[np].LeftScaling().cwiseQuotient(rhs.col(col));
            }
    }

    /**
     * @brief solveKr solves Krr* sol = rhs. In order to obtain numerical stability, parts of the matrix
     * are scaled, this implies also a scaling of the rhs. Note, that the rhs is changed during the function call,
     * but the change is reverted at the end. So in fact, rhs stays unchanged.
     * @param np the actual patch
     * @param rhs the right hand side
     * @param sol the solution
     */
    virtual void solveKrr(unsigned np, gsMatrix<T>& rhs, gsMatrix<T>& sol) const
    {
        GISMO_ASSERT(!info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as false");

        sol.setZero(rhs.rows(),rhs.cols());
        if(rhs.rows()==0)
            return;

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
                rhs.col(col) = m_scalingsKrr[np].LeftScaling().cwiseProduct(rhs.col(col));


        sol=m_LU_Krr[np]->solve(rhs);

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
            {
                sol.col(col) =  m_scalingsKrr[np].RightScaling().cwiseProduct(sol.col(col));
                rhs.col(col) = m_scalingsKrr[np].LeftScaling().cwiseQuotient(rhs.col(col));
            }
    }

    /**
     * @brief solveKC solves [K ,C^T;C , 0]* sol = rhs. In order to obtain numerical stability, parts of the matrix
     * are scaled, this implies also a scaling of the rhs. Note, that the rhs is changed during the function call,
     * but the change is reverted at the end. So in fact, rhs stays unchanged.
     * @param np the actual patch
     * @param rhs the right hand side
     * @param sol the solution
     */
    virtual void solveSpp(gsMatrix<T>& rhs, gsMatrix<T>& sol) const
    {
        sol.setZero(rhs.rows(),rhs.cols());
        if(rhs.rows()==0)
            return;

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
                rhs.col(col) = m_scalingSpp.LeftScaling().cwiseProduct(rhs.col(col));


        sol=m_LU_Spp->solve(rhs);

        if(m_needRescaling)
            for(int col=0;col<rhs.cols();col++)
            {
                sol.col(col) =  m_scalingSpp.RightScaling().cwiseProduct(sol.col(col));
                rhs.col(col) = m_scalingSpp.LeftScaling().cwiseQuotient(rhs.col(col));
            }
    }

    /// \brief   accessor to the LU factorization of Spp. Use this with care, Spp is scaled. use instead solveSpp(...).
    //virtual sparseLUfact*  getLUofSpp() const {return m_LU_Spp;}

    /// \brief  returns LU factorization of KC for one patch. Use this with care, KC is scaled. use instead solveKC(...).
    //virtual sparseLUfact* getLUofKC(int patch) const  {GISMO_ASSERT(info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as true"); return m_LU_KC[patch];}

    /// \brief  returns the conraint matrix C for one patch
    virtual const gsSparseMatrix<T> & getC(int patch) const {GISMO_ASSERT(info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as true"); return m_C[patch];}

    /// \brief  return the LU factorization of Krr for one patch. Use this with care, Krr is scaled. use instead solveKrr(...).
    //virtual sparseLUfact* getLUofKrr(int patch) const  {GISMO_ASSERT(!info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as false");return m_LU_Krr[patch];}

    /// \brief  return the matrix Krp for one patch
    virtual const gsSparseMatrix<T> & getKrp(int patch) const {GISMO_ASSERT(!info.isMinimumEnergy, "Not available for this algorithm, choose minimumEnergy as false");return m_Krp[patch];}

    /// \brief  return the LU factorization of Kii for one patch. Use this with care, Kii is scaled. use instead solveKii(...).
    //virtual sparseLUfact* getLUofKii(int patch) const {return m_LU_Kii[patch];}

    /// \brief  return the matrix Kib for one patch
    virtual const gsSparseMatrix<T> & getKib(int patch) const {return m_Kib[patch];}

    /// \brief  return the matrix Kbb for one patch
    virtual const gsSparseMatrix<T> & getKbb(int patch) const {return m_Kbb[patch];}

    /// \brief  return the matrix Kbb
    virtual const std::vector<gsSparseMatrix<T> >& getKbb() const {return m_Kbb;}

    /**
     * @brief assemble assembles all required matrices and LU factorizations and the corresponding right hand sides.
     *
     * This functions calls:
     *      assembleKiiKibKbb(tripletList);
     *      assembleSpp();
     *      assembleRhs(rhs_loc);
     *
     * In case of Minimum energy also:
     *      assembleC();
     *      assembleLUofKC(tripletList);
     *
     * and otherwise:
     *      assembleKrrKrpKpp(tripletList);
     *
     */
    virtual void assemble(bool needRescaling=true);

    /**
     * @brief numberRhs returns the number of used rhs
     */
    virtual unsigned numberRhs() { return info.nRhs; }

    /**
     * @brief numberLagrangeMult returns the number of lagrange multiplier, i.e. the size of the solution vector used in the PCG.
     */
    virtual unsigned numberLagrangeMult() { return info.lagrangeMult;}

    /**
     * @brief getDirRhs returns ICO nCoupledElimDofs() == true, a vector for the rhs of the system corresponding to B.
     * @return the rhs vector b.
     */
    virtual const gsMatrix<T>& getDirRhs() const {return m_rhs_dir;}

    /**
     * @brief nCoupledElimDofs checks if there are any lagrange multip. which couple a dirichlet and neumann boundary.
     * @return true if there are none.
     */
    virtual bool nCoupledElimDofs() const {return m_freeElimLagr.empty();}

    /**
     * @brief getRawRhs   gives you in case of minimal energy the corresponding components of fb - Kbi * Kii^-1 fi
     * and for not minimal energy fp, fr  (fp entries of f corresponding to the vertices)
     * @param rhs_p the primal component of the corresponding rhs
     * @param rhs_d the remaining component of the corresponding rhs
     */
    virtual void getRawRhs(gsMatrix<T> & rhs_p,std::vector<gsMatrix<T> >& rhs_d) const{
        rhs_p = m_rhs_p;
        rhs_d = m_rhs_d;
    }

    /**
     * @brief processSolution takes the results of IETISolver and returns from the solution (xP,(x2)_i) the
     * solution on the whole domain as a gsMatrix<T>. The output is the same as calling the solver without IETI.
     * IETISolver::calulateSolution needs to be called before. In order to receive a gsField on the whole domain,
     * use the assembler specific constructSolution routine.
     *
     * @param uP primal component of the solution
     * @param u2 ICO minimalEnergy u2 is the solution on the interface and else the solution on the patch.
     *   (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     * @return solVec the solution of the system in matrix form.
     */
    virtual void processSolution(const gsMatrix<T>& uP, const std::vector<gsMatrix<T> >& u2, gsMatrix<T>& solVec) const;

    /**
     * @brief processSolution takes the results of IETISolver and returns from the solution (xP,(x2)_i) the
     * solution on the whole domain as gsField*. IETISolver::calulateSolution needs to be called before.
     * ASSUMES A SPECIFIC TREATMENT OF THE SOLUTION VECTOR, THIS MAY NOT WORK FOR EACH ASSEMBLER. Use instead
     * processSolution(const gsMatrix<T>& uP, const std::vector<gsMatrix<T> >& u2, gsMatrix<T>& solVec);
     *
     * @param uP primal component of the solution
     * @param u2 ICO minimalEnergy u2 is the solution on the interface and else the solution on the patch.
     *   (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     * @return the solution as gsField on the whole domain
     */
    virtual gsField<T>* processSolution(const gsMatrix<T>& uP, const std::vector<gsMatrix<T> >& u2, int rhs=0) const;

    /**
     * @brief embedding transforms a representation of the solution vector from a primal and remaining to patch wise values on the interface
     * @param xP the global primal component ( totalPdofs x nRhs)
     * @param x2 the local remaining/dual component for each patch (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     * @param w the patch wise values on the interface  (dofsB[np] x nRhs)
     */
    virtual void embedding(gsMatrix<T> const & xP,std::vector<gsMatrix<T> >const & x2,std::vector<gsMatrix<T> >& w) const;
    virtual void embedding(gsMatrix<T> const & xP, gsMatrix<T> const & x2, std::size_t np, gsMatrix<T>& w) const;
    /**
     * @brief embeddingTrans transforms a representation of the solution vecor form a patch wise interface values to global primal component and a remaining/dual component
     * @param w the patch wise values on the interface  (dofsB[np] x nRhs)
     * @param uP the global primal component ( totalPdofs x nRhs)
     * @param u2 the local remaining/dual component for each patch (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     */
    virtual void embeddingTrans(const std::vector<gsMatrix<T> >& w,gsMatrix<T>&  uP, std::vector<gsMatrix<T> > &  u2) const;
    virtual void embeddingTrans(const gsMatrix<T> & w,std::size_t np, gsMatrix<T>&  uP, gsMatrix<T> &  u2) const;
    /**
     * @brief assemblePrimal assembles the patch contributions of the primal variables to the global representation
     * @param wi patch local contribution (dofsP[patch] x nRhs)
     * @param patch the actual patch
     * @param uP the updated global primal variable vector (totalPdofs x nRhs)
     */
    virtual void inline assemblePrimal(gsMatrix<T> const & wi, int patch, gsMatrix<T>& uP) const
    {
#ifdef _OPENMP
        if(omp_get_level()!=0)
#else
        if(false)
#endif
        {
            for(int rhs=0; rhs<wi.cols();rhs++)
                for(unsigned i=0; i<info.dofsP[patch];i++)
#pragma omp atomic
                    uP(m_pDofsLoc2Glob[patch][i],rhs) += wi(i,rhs);
        }
        else
            for(unsigned i=0; i<info.dofsP[patch];i++)
                uP.row(m_pDofsLoc2Glob[patch][i]) += wi.row(i);
    }


    /**
     * @brief distributePrimal distributes the global primal variable vector to a certain patch
     * @param xP the global primal variable vector (totalPdofs x nRhs)
     * @param patch the actual patch
     * @param xPi patch local contribution (dofsP[patch] x nRhs)
     */
    virtual void inline distributePrimal(gsMatrix<T>const & xP,int patch,gsMatrix<T> & xPi) const
    {
        xPi.setZero(info.dofsP[patch],xP.cols());
        for(unsigned i=0; i<info.dofsP[patch];i++) {
            xPi.row(i) += xP.row(m_pDofsLoc2Glob[patch][i]);
        }
    }

protected:
    gsIETIAssembler() {}


protected:

    /// @brief pointer to the used assembler
    gsAssemblerBase2<T>* m_assembler;

    /// @brief The multipatch domain
    gsMultiPatch<T> m_patches;

    /// \brief   The discretization bases corresponding to \a m_patches and to
    /// the number of solution fields that are to be computed
    /// m_basis[i]: The multi-basis for unknown i
    std::vector< gsMultiBasis<T> > m_basis;

    /// \brief backup of the "full" boundary conditions
    gsBoundaryConditions<T> m_bConditions;

    /// \brief  An info Structure
    gsIETIInfo info;

    /// \brief  contains global information of the primal dofs to the system (for each component)
    std::vector<gsDofMapper> m_primalDofMapper;

    /// \brief  global standart mapper represent the matched patches (for each component)
    std::vector<gsDofMapper> m_stdMapper;

    /// \brief  local dof mapper, map a basis function of a patch to the real (where elimination happend) index (locally) and for each component .
    /// m_locDofsMapper[np] -> all mappers on patch np.
    std::vector<std::vector< gsDofMapper> > m_locDofsMapper;

    /// \brief  contains information of coupled edges in 3D
    gsDofMapper m_edgeMapper;


    /// \brief  strategy for selecting the primal Dofs
    primalDofMethod m_primalDofstrat;

    ///The stored primal dof mappings dont have a dirichlet boundary, so you dont have to care.
    //(these maps works for uPi -> uP)

    /// \brief  local number of patch k to the global number (in uP).  length: (nPatches)
    std::vector< std::vector<unsigned > > m_pDofsLoc2Glob;

    //for all stored dofs it is assumed that the dirichlet boundary is not eliminated!!!!
    //use mapper.index(i) to get the real index in the matrix.

    /// \brief  contains the local numbers of Vertex Primal dofs for each patch
    std::vector< gsSortedVector<unsigned> > m_primalVdofs;

    /// \brief  Contains the Edges for each patch in 3D
    std::vector< gsSortedVector<Edge > > m_primalEdges;

    /// \brief  contains the local number of the remaining dofs for each patch
    std::vector< std::vector<unsigned > > m_remDofs;

    /// \brief  contains the local numbers of interface dofs for each patch
    std::vector< gsSortedVector<unsigned > > m_boundDofs;

    /// \brief  on a patch k contains for a remaining dof i true if it is also on the interface, i.e. is in m_boundDofs[k]
    std::vector<std::vector<bool> >m_remDofIsAlsoBoundDof;

    /// \brief  contains for each patch the the interface sides
    std::vector< std::vector<patchSide> > m_patchISides;

    /// \brief  boundary conditions for each patch
    std::vector<gsBoundaryConditions<T> >  m_bc_loc;

    /// \brief  maps the global (patchwise) dof to the local interface index or interior index
    std::vector< std::vector<unsigned> > m_glob2BoundInteriorIndex;

    /// \brief  contains true if a global(patchwise) index is a interface index
    std::vector< std::vector<bool> > m_globIsBoundIndex;

    /// \brief  maps a global index to the repective primal or remaining index (ONLY FOR noMinimalEnergy)
    std::vector< std::vector<unsigned> > m_glob2PrimalRemainingIndex;

    /// \brief  contains true if a global index is a primal one (ONLY FOR noMinimalEnergy)
    std::vector< std::vector<bool> > m_globIsPrimalIndex;

    /// \brief  lagrange mult -> the pair of patchDofs
    std::vector<std::pair<patchDof,patchDof> > m_lagrangeTable;

    /// \brief  global interface dof -> merged nodes+patches
    std::map<unsigned,std::set<patchDof> > m_globalConnectionTable;

    /// \brief  lagrange multipl which couple a eliminated and free dof
    std::vector<unsigned> m_freeElimLagr;
    //------------------------------------------------------

    /// \brief  The S-orth Local Basis of the Primal variable Space
    std::vector<gsMatrix<T> > m_Phi;

    /// \brief  Local Lu factorization of [K C^T,C 0]
    std::vector<sparseLUfact* >m_LU_KC;

    /// \brief  Local Matrix representation of C
    std::vector<gsSparseMatrix<T> > m_C;

    /// \brief  Local LU factorization of Krr
    std::vector<sparseSPDfact* >m_LU_Krr;

    /// \brief  Global Schur complement wrt. the primal Variables
    sparseSPDfact* m_LU_Spp;

    /// \brief  Local Matrix Representation of Krp
    std::vector<gsSparseMatrix<T> > m_Krp;

    /// \brief  Local representations of Kpp
    std::vector<gsMatrix<T> > m_Kpp;

    // Vector rhs
    /// \brief  Primal components of the rhs
    gsMatrix<T> m_rhs_p;

    /// \brief  Remaining or dual components of the rhs
    std::vector<gsMatrix<T> > m_rhs_d;

    /// \brief  Interior components of the rhs
    std::vector<gsMatrix<T> > m_rhs_i;

    /// \brief  dirichlet dofs  (if available, e.g. if using nitsche)
    std::vector<gsMatrix<T> > m_dirDofs;

    /// \brief  rhs vector containing the dirichlet values for coupled  eliminated and free dofs (for gsIETIJumpOperator)
    gsMatrix<T> m_rhs_dir;

    /// \brief  Local LU factorization of Kii
    std::vector<sparseSPDfact* >m_LU_Kii;

    /// \brief  Local matrix representation of Kib
    std::vector<gsSparseMatrix<T> > m_Kib;

    /// \brief  Local matrix representation of Kbb
    std::vector<gsSparseMatrix<T> > m_Kbb;


protected:

    /// \brief  information of the edge number. (e.g. the 2. edge on the boundary left has edge number 8, -> orient(2,bounday::left.index() )
    gsMatrix<int> orient;

    std::vector<Eigen::IterScaling<gsSparseMatrix<T> > > m_scalingsKC;
    std::vector<Eigen::IterScaling<gsSparseMatrix<T> > > m_scalingsKii;
    std::vector<Eigen::IterScaling<gsSparseMatrix<T> > > m_scalingsKrr;
    Eigen::IterScaling<gsSparseMatrix<T> > m_scalingSpp;


    bool m_needRescaling;
    bool m_extraTimings;
};






} // namespace gismo

#ifndef GISMO_BUILD_LIB
#include GISMO_HPP_HEADER(gsIETIAssembler.hpp)
#endif
