/** @file gsPoissonAssembler.h

    @brief Provides assembler for the Poisson equation.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): S. Kleiss A. Mantzaflaris, J. Sogn
*/

#pragma once

#include <gsAssembler/gsAssemblerBase2.h>
#include <gsAssembler/gsAssemblerOptions.h>

#include <gsPde/gsPoissonPde.h>

//#include <gsAssembler/gsAssemblerUtils.h>

namespace gismo
{

/** @brief
    Implementation of an (multiple right-hand side) Poisson solver.

    The Poisson equation: \f$-\Delta\mathbf{u}=\mathbf{f} \f$

    It sets up an assembler and assembles the system patch wise and combines
    the patch-local stiffness matrices into a global system by various methods
    (see gismo::gsInterfaceStrategy). It can also enforce Dirichlet boundary
    conditions in various ways (see gismo::gsDirichletStrategy).
*/
template <class T>
class gsPoissonAssembler2 : public gsAssemblerBase2<T>
{
public:
    typedef gsAssemblerBase2<T> Base;

public:


    /* @brief
    Main Constructor of the assembler object.

    \param[in] pde A boundary value Poisson problem
    \param[in] bases a multi-basis that contains patch-wise bases
    \param[in] dirStrategy option for the treatment of Dirichlet boundary
    \param[in] intStrategy option for the treatment of patch interfaces

    gsPoissonAssembler2( gsPoissonPde<T> const         & pde,
                        gsMultiBasis<T> const         & bases,
                        gsDirichletStrategy           dirStrategy,
                        gsInterfaceStrategy           intStrategy)
//*/
protected:
    /// Default constructor
    gsPoissonAssembler2() {}

public:

    /** @brief
    Constructor of the assembler object.

    \param[in] patches is a gsMultiPatch object describing the geometry.
    \param[in] bases a multi-basis that contains patch-wise bases
    \param[in] bconditions is a gsBoundaryConditions object that holds all boundary conditions.
    \param[in] rhs is the right-hand side of the Poisson equation, \f$\mathbf{f}\f$.
    \param[in] dirStrategy option for the treatment of Dirichlet boundary
    \param[in] intStrategy option for the treatment of patch interfaces
*/
    gsPoissonAssembler2( gsMultiPatch<T> const         & patches,
                         gsMultiBasis<T> const         & bases,
                         gsBoundaryConditions<T> const & bconditions,
                         const gsFunction<T>           & rhs,
                         // gsAssemblerOptions options -- to do: update test files
                         dirichlet::strategy           dirStrategy,
                         iFace::strategy               intStrategy = iFace::glue)
        : m_rhsFun(&rhs)
    {
        //set the values, options,...
        m_options.dirStrategy = dirStrategy;
        m_options.intStrategy = intStrategy;
        m_options.dirValues = dirichlet::l2Projection;

        //set the values, options,...
        this->initialize(patches, bases,bconditions);
    }

    virtual gsPoissonAssembler2* minimalClone() const
    {
        gsPoissonAssembler2* c= new gsPoissonAssembler2<T>();
        c->m_rhsFun = m_rhsFun->clone();
        c->m_options = m_options;
        return c;
    }

    virtual void freeMinimalClone()
    {
        delete m_rhsFun;
    }

public:
    /// Sets the Poisson assembler options
    void setOptions(const gsAssemblerOptions& options);

    /// Look at gsAssemblerBases for documentation
    bool isSymmertric() const { return true; }

    /// Main assembly routine.
    void assemble();

    /// Reconstruct solution field from computed solution vector
    gsField<T> * constructSolution(const gsMatrix<T> & solVector) const;

    /// Reconstruct solution from computed solution vector
    void constructSolution(const gsMatrix<T>& solVector,
                           gsMultiPatch<T>& result) const;

    /// Penalty constant for patch \a k, used for Nitsche and
    /// Discontinuous Galerkin methods
    T penalty(int k) const
    {
        //return gsAssemblerUtils<T>::getMu(m_bases[0][k]);
        const int deg = m_bases[0][k].maxDegree();
        return (deg + m_bases[0][k].dim()) * (deg + 1) * T(2.0);
    }

    /// Returns an expression of the "full" assembled sparse
    /// matrix. Note that matrix() returns a lower diagonal matrix,
    /// since we exploit symmetry during assembly.
    Eigen::SparseSelfAdjointView< typename gsSparseMatrix<T>::Base, Lower> fullMatrix()
    {
        return m_matrix.template selfadjointView<Lower>();
    }

    /// Computes the Dirichlet DoF values
    void computeDirichletDofs();

protected:

    // initializes the pde specfic stuff
    void initializePdeSpecific()
    {
        applyOptions();

        m_dofs = m_dofMappers.front().freeSize();
    }

    /// Nitsche Dirichlet contributions
    virtual void assembleNitsche();

    /// Force Dirichlet boundary conditions by diagonal penalization
    void penalizeDirichlet();

    /// Neumann contributions
    void assembleNeumann();

    /// DG contributions
    virtual void assembleDg();

    /// Computes the Dirichlet DoF values by interpolation
    void computeDirichletDofsInptl();

    /// Computes the Dirichlet DoF values by L2 projection
    void computeDirichletDofsL2Proj();

    /// Applies the Poisson assembler options
    void applyOptions();

protected:

    /// Right hand side function
    const gsFunction<T> * m_rhsFun;


protected:

    // Members from gsAssemblerBase2
    using gsAssemblerBase2<T>::m_patches;
    using gsAssemblerBase2<T>::m_bases;
    using gsAssemblerBase2<T>::m_bConditions;
    using gsAssemblerBase2<T>::m_dofMappers;
    using gsAssemblerBase2<T>::m_ddof;
    using gsAssemblerBase2<T>::m_options;
    using gsAssemblerBase2<T>::m_matrix;
    using gsAssemblerBase2<T>::m_rhs;
    using gsAssemblerBase2<T>::m_dofs;
};


} // namespace gismo


//////////////////////////////////////////////////
//////////////////////////////////////////////////


#ifndef GISMO_BUILD_LIB
#include GISMO_HPP_HEADER(gsPoissonAssembler2.hpp)
#endif
