
/** @file gsPoissonHeterogeneousAssembler.h

    @brief Provides assembler for the Poisson equation.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): C. Hofer
*/

#pragma once

#include <gsAssembler/gsPoissonAssembler2.h>
#include <gsAssembler/gsVisitorPoissonHeterogeneous.h> // Stiffness volume integrals and load vector
#include <gsAssembler/gsVisitorDg2.h>
#include <gsAssembler/gsVisitorNitsche2.h>
//#include <gsAssembler/gsAssemblerUtils.h>


namespace gismo
{

/** @brief
    Implementation of an (multiple right-hand side) Poisson solver.

    The Poisson equation: \f$-\Delta\mathbf{u}=\mathbf{f} \f$

    It sets up an assembler and assembles the system patch wise and combines
    the patch-local stiffness matrices into a global system by various methods
    (see gismo::gsInterfaceStrategy). It can also enforce Dirichlet boundary
    conditions in various ways (see gismo::gsDirichletStrategy).
*/
template <class T>
class gsPoissonHeterogeneousAssembler : public gsPoissonAssembler2<T>
{
public:
    typedef gsPoissonAssembler2<T> Base;

protected:
    /// Default Constructor
    gsPoissonHeterogeneousAssembler() {}

public:
    /** @brief
    Constructor of the assembler object.

    \param[in] patches is a gsMultiPatch object describing the geometry.
    \param[in] bases a multi-basis that contains patch-wise bases
    \param[in] bconditions is a gsBoundaryConditions object that holds all boundary conditions.
    \param[in] rhs is the right-hand side of the Poisson equation, \f$\mathbf{f}\f$.
    \param[in] dirStrategy option for the treatment of Dirichlet boundary
    \param[in] intStrategy option for the treatment of patch interfaces
*/
    gsPoissonHeterogeneousAssembler( gsMultiPatch<T> const         & patches,
                         gsMultiBasis<T> const         & bases,
                         gsBoundaryConditions<T> const & bconditions,
                         const gsFunction<T>           & rhs,
                         const gsPiecewiseFunction<T>  & alpha,
                         dirichlet::strategy           dirStrategy,
                         iFace::strategy               intStrategy = iFace::glue)
        : Base(patches, bases, bconditions,rhs,dirStrategy, intStrategy), m_alpha(&alpha)
    {

    }


    virtual gsPoissonHeterogeneousAssembler* minimalClone() const
    {
        gsPoissonHeterogeneousAssembler* c= new gsPoissonHeterogeneousAssembler<T>();
        c->m_rhsFun = m_rhsFun->clone(); // gsFunctionExpr is not threadsave
        c->m_alpha = m_alpha->clone();   // gsPiecewiseFunction may not be threadsave

        c->m_options = m_options;
        return c;
    }

    virtual void freeMinimalClone()
    {
        delete m_rhsFun;
        delete m_alpha;
    }

public:
    /// Main assembly routine.
    void assemble()
    {
        this->computeDirichletDofs();

        if (m_dofs == 0 ) // Are there any interior dofs ?
        {
            gsWarn << " No internal DOFs. Computed Dirichlet boundary only.\n" <<"\n" ;
            return;
        }

        // Pre-allocate non-zero elements for each column of the
        // sparse matrix
        int nonZerosPerCol = 1;
        for (int i = 0; i < m_bases.front().dim(); ++i) // to do: improve
            nonZerosPerCol *= 3 * m_bases.front().maxDegree(i) + 1; // need more for DG !

        m_matrix = gsSparseMatrix<T>(m_dofs, m_dofs); // Clean matrices
        m_matrix.reserve( gsVector<int>::Constant(m_dofs, nonZerosPerCol) );

        // Resize the load vector
        m_rhs.setZero(m_dofs, m_rhsFun->targetDim() );

        // Assemble volume stiffness and load vector integrals
        gsVisitorPoissonHeterogeneous<T> poisson(*m_rhsFun,*m_alpha);
        for (unsigned np=0; np < m_patches.nPatches(); ++np )
        {
            //Assemble stiffness matrix and rhs for the local patch
            // with index np and add to m_matrix and m_rhs
            this->apply(poisson, np);
        }

        // If requested, force Dirichlet boundary conditions by Nitsche's method
        if ( m_options.dirStrategy == dirichlet::nitsche )
            this->assembleNitsche();

        // If requested, force Dirichlet boundary conditions by diagonal penalization
        if ( m_options.dirStrategy == dirichlet::penalize )
            Base::penalizeDirichlet();

        // Enforce Neumann boundary conditions
        Base::assembleNeumann();

        // If we are in in dg (Discontinuous Galerkin) mode: add
        // interface contributions
        if ( m_options.intStrategy == iFace::dg )
            assembleDg();

        // Assembly is done, compress the matrix
        m_matrix.makeCompressed();
    }

    T penalty(int k) const
    {
        const int deg = m_bases[0][k].maxDegree();
        return (deg + m_bases[0][k].dim()) * (deg + 1) * T(2.0);
    }

    void getAlphaValue(unsigned np, const gsMatrix<T>& x, gsMatrix<T>& res) const
    {
        m_alpha->eval_into(np,x,res);
    }

    T getAlphaValue(unsigned np)
    {
        gsMatrix<T> result;
        m_alpha->eval_into(np,gsMatrix<T>::Zero(m_patches[np].parDim(),1),result);
        return result(0,0);
    }

    /// @brief Returns the used-dG Visitor (should be an abstract class)
    gsVisitorDg2<T> visitorDg(const boundaryInterface& bi) {return gsVisitorDg2<T>(penalty(bi.first().patch),bi,m_alpha);}

protected:
    // DG contributions
    void assembleDg()
    {
        for ( typename gsMultiPatch<T>::iiterator
              it = m_patches.iBegin(); it != m_patches.iEnd(); ++it )
        {
            const boundaryInterface & iFace =
                    ( m_bases[0][it->first() .patch].numElements(it->first() .side() ) <
                    m_bases[0][it->second().patch].numElements(it->second().side() ) ?
                        it->getInverse() : *it );

            const T pp = penalty( iFace.first().patch );

            gsVisitorDg2<T> dg(pp, iFace,m_alpha,1);

            this->apply(dg, iFace);
        }
    }

    void assembleNitsche()
    {
        for ( typename gsBoundaryConditions<T>::const_iterator
              it = m_bConditions.dirichletBegin();
              it != m_bConditions.dirichletEnd(); ++it )
        {
            const T pp = this->penalty(it->patch());
            gsVisitorNitsche2<T> nitsche(*it->function(), pp, it->side(),m_alpha);

            // Note: it->unknown() == 0
            this->apply(nitsche, it->patch(), it->side() );
        }
    }

protected:

    const gsPiecewiseFunction<T>* m_alpha;


protected:

    // Members from gsAssemblerBase2
    using Base::m_patches;
    using Base::m_bases;
    using Base::m_bConditions;
    using Base::m_dofMappers;
    using Base::m_ddof;
    using Base::m_options;
    using Base::m_matrix;
    using Base::m_rhs;
    using Base::m_dofs;

    using Base::m_rhsFun;
};



//////////////////////////////////////////////////
//////////////////////////////////////////////////



} // namespace gismo
