/** @file gsSpaceTimeAssembler.h

    @brief Provides assembler for a homogenius Heat equation.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s):  S. Moore
*/


#pragma once

#include <gsAssembler/gsAssemblerBase2.h>
#include <gsAssembler/gsPoissonAssembler2.h>
#include <gsAssembler/gsVisitorSpaceTime.h>
#include <gsAssembler/gsVisitorNitscheSpaceTime.h>
#include <gsAssembler/gsVisitorNeumann.h> // Neumann boundary integrals
#include <gsPde/gsPoissonPde.h>

namespace gismo
{

/** @brief
    Implementation of a Heat Equation Assembler.

    It sets up an assembler and assembles the system patch wise and
    combines the patch-local stiffness matrices into a global system.
    Dirichlet boundary can also be imposed weakly (i.e Nitsche ) 
*/
template <class T>
class gsSpaceTimeAssembler: public gsPoissonAssembler2<T>
{
public:
    typedef gsPoissonAssembler2<T> Base;

public:
/** @brief
    Constructor of the assembler object.

    \param[in] patches is a gsMultiPatch object describing the geometry.
    \param[in] bases a multi-basis that contains patch-wise bases
    \param[in] bconditions  is a gsBoundaryConditions object that holds boundary conditions on the form:
    \f[ \text{Dirichlet: } u = g \text{ on } \Sigma_N, \text{ and Neumann: } \nabla_x u \cdot \mathbf{n} = g_N \text{ on } \Sigma_N\f]
    \f$ v \f$ is the test function and \f$ \Sigma_D \f$ is the boundary side.
    \param[in] rhs is the right-hand side of the Biharmonic equation, \f$\mathbf{f}\f$.
    \param[in] dirStrategy option for the treatment of Dirichlet boundary in the \em bconditions object.
    \param[in] intStrategy option for the treatment of patch interfaces
*/
    gsSpaceTimeAssembler( gsMultiPatch<T> const         & patches,
                          gsMultiBasis<T> const         & bases,
                          gsBoundaryConditions<T> const & bconditions,
                          const gsFunction<T>           & rhs,
                          dirichlet::strategy           dirStrategy,
                          iFace::strategy               intStrategy = iFace::glue)
    : Base( patches,   bases,  bconditions, rhs, dirStrategy, intStrategy) 
    {    }

    /// Main assembly routine
    void assemble()
    {
        Base::computeDirichletDofs();

        if (m_dofs == 0 ) // Are there any interior dofs ?
        {
            gsWarn << " No internal DOFs. Computed Dirichlet boundary only.\n" <<"\n" ;
            return;
        }
        
        // Pre-allocate non-zero elements for each column of the
        // sparse matrix
        int nonZerosPerCol = 1;
        for (int i = 0; i < m_bases.front().dim(); ++i) // to do: improve
            nonZerosPerCol *= 3 * m_bases.front().maxDegree(i) + 1;

        m_matrix = gsSparseMatrix<T>(m_dofs, m_dofs); // Clean matrices
        m_matrix.reserve( gsVector<int>::Constant(m_dofs, nonZerosPerCol) );

        // Resize the load vector
        m_rhs.setZero(m_dofs, m_rhsFun->targetDim() );

        // Assemble modified volume stiffness and load vector integrals
        gsVisitorSpaceTime<T> spacetime(*m_rhsFun);
        for (unsigned np=0; np < m_patches.nPatches(); ++np )
        {
            //Assemble stiffness matrix and rhs for the local patch
            // with index np and add to m_matrix and m_rhs
            this->apply(spacetime, np);
        }

        for ( typename gsBoundaryConditions<T>::const_iterator
                  it = m_bConditions.neumannBegin();
              it != m_bConditions.neumannEnd(); ++it )
        {
            gsVisitorNeumann<T> neumann(*it->function(), it->side());
            this->apply(neumann, it->patch(), it->side() );
        }

        // If requested, force Dirichlet boundary conditions by Nitsche's method
        if ( m_options.dirStrategy == dirichlet::nitsche )
        {
            for ( typename gsBoundaryConditions<T>::const_iterator
                      it = m_bConditions.dirichletBegin();
                  it != m_bConditions.dirichletEnd(); ++it )
            {
                const T pp = this->penalty(it->patch());
                gsVisitorNitscheSpaceTime<T> nitsche(*it->function(), pp , it->side());
            
                // Note: it->unknown() == 0
                this->apply(nitsche, it->patch(), it->side() );
            }
        }
        
        // Assembly is done, compress the matrix
        m_matrix.makeCompressed();
    }


protected:

    // Members from gsPoissonAssembler
    using Base::m_patches;
    using Base::m_bases;
    using Base::m_dofMappers;
    using Base::m_ddof;
    using Base::m_matrix;
    using Base::m_rhs;
    using Base::m_dofs;
    using Base::m_rhsFun;
    using Base::m_bConditions;
    using Base::m_options;

};


//////////////////////////////////////////////////
//////////////////////////////////////////////////


} // namespace gismo