/** @file gsVisitorSpaceTime.h

    @brief Heat equation element visitor.

    This file is part of the G+Smo library. 

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
    
    Author(s): S. Moore
*/

#pragma once

namespace gismo
{

/** \brief Visitor for the Heat equation.
 *
 * Assembles the bilinear terms
 * \f[ (\partial_t u , v)_Q + (\nabla_x u,\nabla_x v)_Q \text{ and } (f,v)_Q \f]
 * For \f[ u = g \quad on  \quad \Sigma \f],
 *     \f[ u(x,0) = u_0(x) \quad on \quad \Omega \times \{0\} \f],
 */

template <class T, bool paramCoef = false>
class gsVisitorSpaceTime
{
public:
    
    /** \brief Constructor for gsVisitorSpaceTime.
     *
     * \param[in] rhs Given right-hand-side function/source term that, for
     * \param[in] coeff_a Given coefficient
     */
    /// Constructor with the right hand side function of the Poisson equation
    gsVisitorSpaceTime(const gsFunction<T> & rhs) : 
    rhs_ptr(&rhs)
    { }

    gsVisitorSpaceTime() :     rhs_ptr(NULL) { }


    void initialize(const gsBasis<T> & basis, 
                    gsQuadRule<T>    & rule, 
                    unsigned         & evFlags )
    {
        gsVector<index_t> numQuadNodes( basis.dim() );
        for (int i = 0; i < basis.dim(); ++i) // to do: improve
            numQuadNodes[i] = basis.degree(i) + 1;
        
        // Setup Quadrature
        rule = gsGaussRule<T>(numQuadNodes);// harmless slicing occurs here

        // Set Geometry evaluation flags
        evFlags = NEED_VALUE | NEED_MEASURE | NEED_GRAD_TRANSFORM;
    }

    // Evaluate on element.
    inline void evaluate(gsBasis<T> const       & basis, // to do: more unknowns
                         gsGeometryEvaluator<T> & geoEval,
                         gsMatrix<T> const      & quNodes)
    {
        // Compute the active basis functions
        // Assumes actives are the same for all quadrature points on the elements
        basis.active_into(quNodes.col(0), actives);
        numActive = actives.rows();
        
        // Evaluate basis functions on element
        basis.evalAllDers_into( quNodes, 1, basisData );
        
        // Compute image of Gauss nodes under geometry mapping as well as Jacobians
        geoEval.evaluateAt(quNodes);// is this generic ??
        
        // Evaluate right-hand side at the geometry points paramCoef
        // specifies whether the right hand side function should be
        // evaluated in parametric(true) or physical (false)
        rhs_ptr->eval_into( (paramCoef ?  quNodes :  geoEval.values() ), rhsVals );
        
        // Initialize local matrix/rhs
        localMat.setZero(numActive, numActive      );
        localRhs.setZero(numActive, rhsVals.rows() );//multiple right-hand sides
    }
    
    inline void assemble(gsDomainIterator<T>    & element, 
                         gsGeometryEvaluator<T> & geoEval,
                         gsVector<T> const      & quWeights)
    {
        const unsigned d = element.dim();
        const gsMatrix<T> & basisVals  = basisData[0];
        const gsMatrix<T> & basisGrads = basisData[1];

        for (index_t k = 0; k < quWeights.rows(); ++k) // loop over quadrature nodes
        {
            // Multiply weight by the geometry measure
            const T weight = quWeights[k] * geoEval.measure(k);
            
            // Compute physical gradients at k as a Dim x NumActive matrix
            geoEval.transformGradients(k, basisGrads, physGrad);

            // Physical gradients in only (Dim)-coordinates at k 
            // is a (Dim-1) x NumActive matrix    
            // implemented for d = 2,3
            physGradSpace   = physGrad.topRows(d-1);
            physGradTime    = physGrad.bottomRows(1);
            
                        
            localRhs.noalias() += weight * ( basisVals.col(k) * rhsVals.col(k).transpose() ) ;
            localMat.noalias() += weight * ( physGradSpace.transpose() * physGradSpace 
                                           + basisVals.col(k) * physGradTime);
            
        }
    }
    
    inline void localToGlobal(const gsDofMapper     & mapper,
                              const gsMatrix<T>     & eliminatedDofs,
                              const int patchIndex,
                              gsSparseMatrix<T>     & sysMatrix,
                              gsMatrix<T>           & rhsMatrix )
    {
        //Assert eliminatedDofs.rows() == mapper.boundarySize()

        // Local DoFs to global DoFs
        mapper.localToGlobal(actives, patchIndex, actives);
        //const int numActive = actives.rows();
        
        for (index_t i=0; i < numActive; ++i)
        {
            const int ii = actives(i);
            if ( mapper.is_free_index(ii) )
            {
                rhsMatrix.row(ii) += localRhs.row(i);
                
                for (index_t j=0; j < numActive; ++j)
                {
                    const int jj = actives(j);
                    if ( mapper.is_free_index(jj) )
                    {
                        // Matrix is symmetric, so we could store only
                        // lower triangular part
                        //if ( jj <= ii ) 
                        sysMatrix.coeffRef(ii, jj) += localMat(i, j);
                    }
                    else // if ( mapper.is_boundary_index(jj) ) // Fixed DoF?
                    {
                        rhsMatrix.row(ii).noalias() -= localMat(i, j) * 
                            eliminatedDofs.row( mapper.global_to_bindex(jj) );
                    }
                }
            }
        }
    }

protected:
    // Right hand side
    const gsFunction<T> * rhs_ptr;

protected:
    // Basis values
    std::vector<gsMatrix<T> > basisData;
    gsMatrix<T>        physGrad;
    gsMatrix<unsigned> actives;
    index_t numActive;

    gsMatrix<T> physGradTime, physGradSpace;

protected:
    // Local values of the right hand side
    gsMatrix<T> rhsVals;

protected:
    // Local matrices
    gsMatrix<T> localMat, localMass;
    gsMatrix<T> localRhs;
};


} // namespace gismo
