
#include <gsCore/gsBasisEvaluator.h>
#include <gsAssembler/gsLocalToGlobal.h>
#include <gsAssembler/gsGaussRule.h>
#include <gsCore/gsGeometry.h>

#include <gsPde/gsPoissonPde.h>

#include <iostream>
#include <iomanip>

#pragma once

namespace gismo {

template <typename T>
struct SystemMatrices
{
    gsDofMapper                 dof_map;
    gsSparseMatrix<T>           sys;
    gsSparseMatrix<T>           rhs_mod;
    gsMatrix<T>                 rhs;
    gsSparseMatrix<T>           elim;
    gsMatrix<T>                 elim_rhs;
};

template<typename T>
std::ostream & operator<< ( std::ostream &os, const SystemMatrices<T> &matrices )
{
    os.width(5);
    os<< std::setprecision(5);
    const bool print = matrices.sys.cols()*matrices.sys.rows()<=225;

    os<< "system: (" <<  matrices.sys.rows() << "x" << matrices.sys.cols() << ")\n";
    if (print) os<<matrices.sys.toDense()<<"\n";
    os<< "rhs_mod: (" <<  matrices.rhs_mod.rows() << "x" << matrices.rhs_mod.cols() << ")\n";
    if (print) os<<matrices.rhs_mod.toDense()<<"\n";
    os<< "rhs: (" <<  matrices.rhs.rows() << "x" << matrices.rhs.cols() << ")\n";
    if (print) os<<matrices.rhs<<"\n";
    os<< "elim_sys: (" <<  matrices.elim.rows() << "x" << matrices.elim.cols() << ")\n";
    if (print) os<<matrices.elim.toDense()<<"\n";
    os<< "elim_rhs: (" <<  matrices.elim_rhs.rows() << "x" << matrices.elim_rhs.cols() << ")\n";
    if (print) os<<matrices.elim_rhs<<"\n";

    return os;
}


template <typename T>
class gsElementAssembler
{
public:
    virtual ~gsElementAssembler() {}
    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights) = 0;
};


template <typename T>
class gsScalarEASkeleton
{
protected:
    gsL2GMapped<gsSparseMatrix<T>,gsSparseMatrix<T> >
    m_sys;
    gsL2GMappedRhs<gsMatrix<T> >       m_rhs;
    gsBasisEvaluator<T>               *m_basis;
    gsGeometryEvaluator<T>            *m_geo;
public:
    gsScalarEASkeleton(SystemMatrices<T> &matrices,
                       gsBasis<T>        &basis,
                       const gsGeometry<T>     &geo,
                       index_t            patch_index
                       )
        : m_sys(matrices.sys,matrices.rhs_mod,matrices.dof_map,matrices.dof_map,patch_index),
          m_rhs(matrices.rhs,matrices.dof_map,patch_index)
    {
        m_geo   = geo.evaluator(0);
        m_basis = makeBasisEvaluator(basis);
    }

    virtual ~gsScalarEASkeleton()
    {
        delete m_geo;
        delete m_basis;
    }
};

template <typename T>
class gsScalarBoundaryEASkeleton
{
protected:
    typedef gsBoundaryWriter<gsSparseMatrix<T> > SW;
    typedef gsBoundaryWriter<gsMatrix<T> > SWR;

    typedef gsDOFMappedMapping<T>                    DOF;
    typedef gsL2GBase<DOF,DOF,SW>                    MassMapper;
    typedef gsL2GBase<DOF,gsIdentityMapping,SWR>     RhsMapper;

    MassMapper                m_sys;
    RhsMapper                 m_rhs;
    gsBasisEvaluator<T>      *m_basis;
    gsGeometryEvaluator<T>   *m_geo;
public:
    gsScalarBoundaryEASkeleton(SystemMatrices<T>  &matrices,
                               gsBasis<T>         &basis,
                               const gsGeometry<T>      &geo,
                               index_t             patch_id
                               )
        : m_sys(DOF(matrices.dof_map,patch_id),DOF(matrices.dof_map,patch_id),SW(matrices.elim,matrices.dof_map.freeSize(),matrices.dof_map.freeSize())),
          m_rhs(DOF(matrices.dof_map,patch_id),gsIdentityMapping() ,SWR(matrices.elim_rhs,matrices.dof_map.freeSize(),0))
    {
        m_geo   = geo.evaluator(0);
        m_basis = makeBasisEvaluator(basis);
    }
    virtual ~gsScalarBoundaryEASkeleton()
    {
        delete m_basis;
        delete m_geo;
    }
};

template <typename T>
class gsMassElementAssembler : public gsElementAssembler<T> , public gsScalarEASkeleton<T>
{
protected:
    using gsScalarEASkeleton<T>::m_sys;
    using gsScalarEASkeleton<T>::m_rhs;
    using gsScalarEASkeleton<T>::m_basis;
    using gsScalarEASkeleton<T>::m_geo;
public:
    gsMassElementAssembler (
            SystemMatrices<T> &matrices,
            gsBasis<T>        &basis,
            gsGeometry<T>     &geo,
            index_t            patch_index)
        : gsScalarEASkeleton<T>(matrices,basis,geo,patch_index )
    {
        m_basis->addFlags(NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_MEASURE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,m_geo);

        Eigen::DiagonalWrapper<gsVector<T> > weights(quad_weights.array()*m_geo->measures().array());
        gsMatrix<T> element_matrix = m_basis->values()*weights*m_basis->values().transpose();
        m_sys.store(m_basis->actives(),m_basis->actives(), element_matrix);
    }
};


template <typename T>
class gsMassBoundaryElementAssembler : public gsElementAssembler<T> , public gsScalarBoundaryEASkeleton<T>
{
protected:
    using gsScalarBoundaryEASkeleton<T>::m_sys;
    using gsScalarBoundaryEASkeleton<T>::m_rhs;
    using gsScalarBoundaryEASkeleton<T>::m_basis;
    using gsScalarBoundaryEASkeleton<T>::m_geo;
    patchSide     m_side;
public:
    gsMassBoundaryElementAssembler (
            SystemMatrices<T> &matrices,
            gsBasis<T>        &basis,
            gsGeometry<T>     &geo,
            patchSide         side)
        : gsScalarBoundaryEASkeleton<T>(matrices,basis,geo,side.patch), m_side(side)
    {
        m_basis->addFlags(NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_OUTER_NORMAL);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,m_geo);

        gsVector<T> measures(quad_nodes.cols());
        for (index_t i=0;i<quad_nodes.cols();++i)
        {
            gsVector<T> out;
            m_geo->outerNormal(i,m_side.side(), out);
            measures(i) = out.norm()*quad_weights(i);
        }
        Eigen::DiagonalWrapper<gsVector<T> > weights(measures);
        gsMatrix<T> element_matrix = m_basis->values()*weights*m_basis->values().transpose();
        m_sys.store(m_basis->actives(),m_basis->actives(), element_matrix);
    }
};





template <typename T>
class gsStiffnessElementAssembler : public gsScalarEASkeleton<T>
{
protected:
    using gsScalarBoundaryEASkeleton<T>::m_sys;
    using gsScalarBoundaryEASkeleton<T>::m_rhs;
    using gsScalarBoundaryEASkeleton<T>::m_basis;
    using gsScalarBoundaryEASkeleton<T>::m_geo;
public:
    gsStiffnessElementAssembler (
            SystemMatrices<T>  &matrices,
            gsBasis<T>         &basis,
            gsGeometry<T>      &geo,
            index_t             patch)
        : gsScalarBoundaryEASkeleton<T>(matrices,basis,geo,patch)
    {
        m_basis->addFlags(NEED_GRAD);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_MEASURE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,*m_geo);
        const index_t grad_size=m_basis->getParDim();

        gsVector<T> weight_vec = quad_weights.array()*m_geo->measures().array();
        gsMatrix<T> element_matrix;
        element_matrix.setZero(m_basis->actives().rows(),m_basis->actives().rows());
        for (index_t col=0; col< m_basis->actives().rows(); ++col)
        {
            for (index_t row=0; row< col; ++row)
            {
                gsMatrix<T> tempM = (m_basis->derivs().middleRows(row*grad_size,grad_size ).array() * m_basis->derivs().middleRows(col*grad_size,grad_size).array());
                T  temp = tempM.colwise().sum().dot(weight_vec);
                element_matrix(row,col) = temp;
                element_matrix(col,row) = temp;
            }
            T temp =weight_vec.dot(m_basis->derivs().middleRows(col*grad_size,grad_size ).colwise().squaredNorm());
            element_matrix(col,col) = temp;
        }
        m_sys.store(m_basis->actives(),m_basis->actives(),element_matrix);
    }
};

template <typename T>
class gsScalarDirichletElementAssembler : public gsElementAssembler<T> , public gsScalarBoundaryEASkeleton<T>
{
protected:
    using gsScalarBoundaryEASkeleton<T>::m_sys;
    using gsScalarBoundaryEASkeleton<T>::m_rhs;
    using gsScalarBoundaryEASkeleton<T>::m_basis;
    using gsScalarBoundaryEASkeleton<T>::m_geo;
    gsFunction<T> *m_rhs_fun;
    patchSide     m_side;
public:
    gsScalarDirichletElementAssembler (
            SystemMatrices<T> &matrices,
            gsBasis<T>        &basis,
            const gsGeometry<T>     &geo,
            gsFunction<T>     *rhs_fun,
            patchSide         side)
        : gsScalarBoundaryEASkeleton<T>(matrices,basis,geo,side.patch), m_rhs_fun(rhs_fun), m_side(side)
    {
        m_basis->addFlags(NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_OUTER_NORMAL);
        if (rhs_fun) m_geo->addFlags(NEED_VALUE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,*m_geo);

        gsVector<T> measures(quad_nodes.cols());
        for (index_t i=0;i<quad_nodes.cols();++i)
        {
            gsVector<T> out;
            m_geo->outerNormal(i,m_side.side(), out);
            measures(i) = out.norm()*quad_weights(i);
        }
        Eigen::DiagonalWrapper<Eigen::Matrix<T,Dynamic,1> > weights(measures);
        gsMatrix<T> element_matrix = m_basis->values()*weights*m_basis->values().transpose();
        m_sys.store(m_basis->actives(),m_basis->actives(),element_matrix);

        if (m_rhs_fun)
        {
            gsMatrix<T> fun_values;
            m_rhs_fun->eval_into(m_geo->values(),fun_values);
            gsMatrix<T> element_rhs = m_basis->values()*weights*fun_values.transpose();
            m_rhs.store(m_basis->actives(),m_basis->actives(),element_rhs);
        }
    }
};


template <typename T>
class gsScalarNeumannElementAssembler : public gsElementAssembler<T> , public gsScalarEASkeleton<T>
{
protected:
    using gsScalarEASkeleton<T>::m_sys;
    using gsScalarEASkeleton<T>::m_rhs;
    using gsScalarEASkeleton<T>::m_basis;
    using gsScalarEASkeleton<T>::m_geo;
    gsFunction<T> *m_rhs_fun;
    patchSide     m_side;
public:
    gsScalarNeumannElementAssembler (
            SystemMatrices<T> &matrices,
            gsBasis<T>        &basis,
            const gsGeometry<T>     &geo,
            gsFunction<T>     *rhs_fun,
            patchSide         side)
        : gsScalarEASkeleton<T>(matrices,basis,geo,side.patch), m_rhs_fun(rhs_fun), m_side(side)
    {
        m_basis->addFlags(NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags()| NEED_OUTER_NORMAL);
        if (rhs_fun) m_geo->addFlags(NEED_VALUE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        if (m_rhs_fun)
        {
            m_geo->evaluateAt(quad_nodes);
            m_basis->evaluateAt(quad_nodes ,*m_geo);

            gsVector<T> measures(quad_nodes.cols());
            for (index_t i=0;i<quad_nodes.cols();++i)
            {
                gsVector<T> out;
                m_geo->outerNormal(i,m_side.side(), out);
                measures(i) = out.norm()*quad_weights(i);
            }

            Eigen::DiagonalWrapper<Eigen::Matrix<T,Dynamic,1> > weights( measures);

            gsMatrix<T> fun_values;
            m_rhs_fun->eval_into(m_geo->values(),fun_values);
            gsMatrix<T> element_rhs= m_basis->values()*weights*fun_values.transpose();
            m_rhs.store(m_basis->actives(),m_basis->actives(), element_rhs);
        }
    }
};

}
