/** @file gsGenericAssembler.h

    @brief DESCRIPTION

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan
**/

#include <gsPde/gsPde.h>
#include <gsAssembler3/gsElementAssembler.h>
#include <gsAssembler/gsAssemblerOptions.h>
#include <gsCore/gsMultiBasis.h>
#include <gsCore/gsDofMapper.h>

#pragma once

namespace gismo {


template <typename T>
class gsPdeAssembler
{
protected:
    const gsPde<T>                   &m_pde;
    SystemMatrices<T>           m_matrices;
public:
    gsPdeAssembler(const gsPde<T> &pde)
        :m_pde(pde)
     {}
    virtual ~gsPdeAssembler() {}

    // functions to be implemented by subclasses
    virtual gsElementAssembler<T> *getElementAssembler             (index_t patch)  = 0;
    virtual gsElementAssembler<T> *getElementAssemblerForBoundary  (const boundary_condition<T> &bc)  = 0;
    virtual gsElementAssembler<T> *getElementAssemblerForInterface (const boundaryInterface     &bi)  = 0;

    virtual void getIntegrationRule             ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, index_t patch)   = 0;
    virtual void getIntegrationRuleForBoundary  ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, patchSide side)  = 0;
    virtual void getIntegrationRuleForInterface ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, patchSide side1, patchSide side2)  = 0;

    virtual void initSystemMatrices () = 0;

    // static functions
    void assemble()
    {
        this->initSystemMatrices();
        assemble_volumes    ();
        assemble_boundaries ();
        assemble_interfaces ();
    }
    SystemMatrices<T> &getSystemMatrices () { return  m_matrices; }

    void assemble_boundaries()
    {
        // assemble boundary conditions
        std::vector<boundary_condition<T> > bc = m_pde.boundaryConditions().allConditions();
        typename gsBoundaryConditions<T>::const_iterator bc_it;
        for (bc_it=bc.begin(); bc_it!=bc.end(); ++bc_it)
        {
            gsElementAssembler<T> * element_assembler = this->getElementAssemblerForBoundary(*bc_it);
            if (element_assembler)
            {
                gsDomainIterator<T> *element;
                gsGaussRule<T>       *quadrature;
                this->getIntegrationRuleForBoundary(&element , &quadrature, bc_it->ps );
                quadrature_assemble(*element,*quadrature,*element_assembler);
                delete quadrature;
                delete element;
                delete element_assembler;
            }
        }
    }
    void assemble_interfaces()
    {
        // assemble interface integrals
        // need to rethink the interfaces structures
        typename gsMultiPatch<T>::const_iiterator in_it;
        for (in_it = m_pde.domain().iBegin(); in_it != m_pde.domain().iEnd(); ++in_it)
        {
            gsElementAssembler<T> * element_assembler = this->getElementAssemblerForInterface(*in_it);
            if (element_assembler)
            {
                gsDomainIterator<T> *element;
                gsGaussRule<T>       *quadrature;
                this->getIntegrationRuleForInterface(&element , &quadrature,in_it->first(),in_it->second());
                quadrature_assemble(*element,*quadrature,*element_assembler);
                delete quadrature;
                delete element;
                delete element_assembler;
            }
        }
    }
    void assemble_volumes ()
    {
        // assemble volume integrals
        for (index_t patch=0; patch<m_pde.domain().size(); ++patch)
        {
            gsElementAssembler<T> * element_assembler = this->getElementAssembler(patch);
            if (element_assembler)
            {
                gsDomainIterator<T> *element;
                gsGaussRule<T>       *quadrature;
                this->getIntegrationRule(&element , &quadrature,patch);
                quadrature_assemble(*element,*quadrature,*element_assembler);
                delete quadrature;
                delete element;
                delete element_assembler;
            }
        }
    }
protected:
    inline void quadrature_assemble(gsDomainIterator<T> & element, gsGaussRule<T>  & quadrature, gsElementAssembler<T> &worker)
    {
        gsMatrix<T> quad_nodes;
        gsVector<T> quad_weight;
        for (element.reset(); element.good(); element.next())
        {
            quadrature.mapTo( element.lowerCorner(), element.upperCorner(), quad_nodes, quad_weight );
            worker(element, quad_nodes, quad_weight );
        }
    }
};



template <typename T>
class gsScalarPdeAssembler : public gsPdeAssembler<T>
{
protected:
    using gsPdeAssembler<T>::m_matrices;
    using gsPdeAssembler<T>::m_pde;

    gsMultiBasis<T>      m_basis;
    dirichlet::strategy  m_dir_strategy;
    iFace::strategy      m_interface_strategy;
public:
    gsScalarPdeAssembler (const gsPde<T> &pde, const gsMultiBasis<T> &basis, dirichlet::strategy  dir_strategy, iFace::strategy  interface_strategy)
        :
        gsPdeAssembler<T>(pde),
        m_basis(basis),
        m_dir_strategy(dir_strategy),
        m_interface_strategy(interface_strategy)
    {}

    gsField<T> *constructSolution(const gsMatrix<T> & solCoeff, const gsMatrix<T> & dirValues)
    {
    const gsMultiPatch<T> &m_geo = m_pde.domain();
    gsDofMapper     &m_dof_mapper = m_matrices.dof_map;
    std::vector<gsFunction<T>*> sols ;
    for (size_t p=0; p < m_geo.nPatches(); ++p )
    {
        const int sz  = m_basis[p].size();
        const int dim = m_pde.numRhs();

        gsMatrix<T> coeffs(sz, dim);
        for (index_t i = 0; i < sz; ++i)
        {
            if ( m_dof_mapper.is_free(i, p) ) // internal or interface
                coeffs.row(i) = solCoeff.row(m_dof_mapper.index(i, p));
            else // eliminated Dof: fill with Dirichlet data
                coeffs.row(i).array() = dirValues.row( m_dof_mapper.bindex(i, p) );
        }
        sols.push_back( m_basis[p].makeGeometry(coeffs));
    }
    return new gsField<T>( m_geo, sols, true );
    }

    virtual void getIntegrationRule             ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, index_t patch)
    {
        gsVector<index_t> numQuadNodes;
        numQuadNodes.setConstant(m_pde.domain().dim(), m_basis[patch].maxDegree()+1);
        *quadrature = new gsGaussRule<T>(numQuadNodes);
        *elements   = m_basis[patch].makeDomainIterator().release();
    }

    virtual void getIntegrationRuleForBoundary  ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, patchSide side)
    {
        gsVector<index_t> numQuadNodes;
        numQuadNodes.setConstant(m_pde.domain().dim(), m_basis[side.patch].maxDegree()+1);
        numQuadNodes[side.direction()] = 1;
        *quadrature = new gsGaussRule<T>(numQuadNodes);
        *elements   = m_basis[side.patch].makeDomainIterator(side.side()).release();
    }

    virtual void getIntegrationRuleForInterface ( gsDomainIterator<T> **elements, gsGaussRule<T> **quadrature, patchSide side1, patchSide side2)
    {
        gsVector<index_t> numQuadNodes;
        unsigned degMax = std::max(m_basis[side1.patch].maxDegree(),m_basis[side1.patch].maxDegree());
        numQuadNodes.setConstant(m_pde.domain().dim(), degMax+1);
        numQuadNodes[side1.direction()] = 1;
        *quadrature = new gsGaussRule<T>(numQuadNodes);
        /// TODO this is assuming that the side1 has a finer mesh then
        /// side2, in principle we should take the coarsest refinement of both
        *elements   = m_basis[side1.patch].makeDomainIterator(side1.side()).release();
    }

    virtual void initSystemMatrices ()
    {
        // init dof mapper
        m_matrices.dof_map.init(m_basis);
        if ( m_interface_strategy == iFace::glue)
        {
            for (typename gsMultiPatch<T>::const_iiterator
                 it = m_pde.domain().iBegin() ; it != m_pde.domain().iEnd(); ++it )
            {
                m_basis.matchInterface(*it, m_matrices.dof_map);
            }
        }
        if ( m_dir_strategy == dirichlet::elimination )
        {
            for (typename gsBoundaryConditions<T>::const_iterator
                 it = m_pde.boundaryConditions().dirichletBegin() ; it != m_pde.boundaryConditions().dirichletEnd(); ++it )
            {
                gsMatrix<unsigned> * bnd = m_basis[it->ps.patch].boundary( it->ps.side() );
                m_matrices.dof_map.markBoundary(it->ps.patch, *bnd);
                delete bnd;
            }
        }
        m_matrices.dof_map.finalize();
       // set matrices to 0
        m_matrices.rhs.setZero(m_matrices.dof_map.freeSize(),m_pde.numRhs());
        m_matrices.elim_rhs.setZero(m_matrices.dof_map.boundarySize(),m_pde.numRhs());
        // resize sparse matrices
        m_matrices.sys.resize(m_matrices.dof_map.freeSize(),m_matrices.dof_map.freeSize());
        m_matrices.rhs_mod.resize(m_matrices.dof_map.freeSize(),m_matrices.dof_map.boundarySize());
        m_matrices.elim.resize(m_matrices.dof_map.boundarySize(),m_matrices.dof_map.boundarySize());
        // preallocate memory for sparse matrices
        int nonZerosPerCol=1;
        for (int dir = 0; dir < m_basis.dim(); ++dir)
        {
            nonZerosPerCol *= 2 * m_basis.maxDegree(dir) + 1;
        }
        if (m_matrices.sys.rows()*m_matrices.sys.cols())
            m_matrices.sys.reserve    ( gsVector<int>::Constant(m_matrices.sys.cols(),     nonZerosPerCol) );
        if (m_matrices.rhs_mod.rows()*m_matrices.rhs_mod.cols())
            m_matrices.rhs_mod.reserve( gsVector<int>::Constant(m_matrices.rhs_mod.cols(), nonZerosPerCol) );
        if (m_matrices.elim.rows()*m_matrices.elim.cols())
            m_matrices.elim.reserve   ( gsVector<int>::Constant(m_matrices.elim.cols(),    nonZerosPerCol) );
    }
};


}
