#include <gsAssembler3/gsElementAssembler.h>
#include <gsAssembler3/gsGenericAssembler.h>

#pragma once

namespace gismo {


template <typename T>
class gsPoissonElementAssembler : public gsElementAssembler<T> ,public gsScalarEASkeleton<T>
{
protected:
    using gsScalarEASkeleton<T>::m_sys;
    using gsScalarEASkeleton<T>::m_rhs;
    using gsScalarEASkeleton<T>::m_basis;
    using gsScalarEASkeleton<T>::m_geo;
    const gsFunction<T>              *m_rhs_fun;
public:
    gsPoissonElementAssembler (
            SystemMatrices<T> &matrices,
            gsBasis<T>        &basis,
            const gsGeometry<T>     &geo,
            const gsFunction<T>     *source,
            index_t            patch)
        : gsScalarEASkeleton<T>(matrices,basis,geo,patch), m_rhs_fun(source)
    {
        m_basis->addFlags(NEED_GRAD | NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_MEASURE);
        if (m_rhs_fun) m_geo->addFlags(NEED_VALUE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,*m_geo);
        const index_t grad_size=m_basis->derivs().rows()/m_basis->actives().rows();

        gsVector<T> weight_vec = quad_weights.array()*m_geo->measures().array();
        gsMatrix<T> element_matrix;
        element_matrix.setZero(m_basis->actives().rows(),m_basis->actives().rows());
        for (index_t col=0; col< m_basis->actives().rows(); ++col)
        {
            for (index_t row=0; row< col; ++row)
            {
                gsMatrix<T> tempM = (m_basis->derivs().middleRows(row*grad_size,grad_size ).array() * m_basis->derivs().middleRows(col*grad_size,grad_size).array());
                T  temp = tempM.colwise().sum().dot(weight_vec);
                element_matrix(row,col) = temp;
                element_matrix(col,row) = temp;
            }
            T temp =weight_vec.dot(m_basis->derivs().middleRows(col*grad_size,grad_size ).colwise().squaredNorm());
            element_matrix(col,col) = temp;
        }
        m_sys.store(m_basis->actives(),m_basis->actives(),element_matrix);

        Eigen::DiagonalWrapper<Eigen::Matrix<T,Dynamic,1> > weights(weight_vec);
        if (m_rhs_fun)
        {
            gsMatrix<T> fun_values;
            m_rhs_fun->eval_into(m_geo->values(),fun_values);
            gsMatrix<T> element_rhs = m_basis->values()*weights*fun_values.transpose();
            m_rhs.store(m_basis->actives(),m_basis->actives(),element_rhs);
        }
    }
};


template <typename T>
class gsPoissonNitscheBoundaryAssembler : public gsElementAssembler<T> , public gsScalarEASkeleton<T>
{
protected:
    using gsScalarEASkeleton<T>::m_sys;
    using gsScalarEASkeleton<T>::m_rhs;
    using gsScalarEASkeleton<T>::m_basis;
    using gsScalarEASkeleton<T>::m_geo;
    gsFunction<T>  *m_dir_fun;
    patchSide      m_side;
    T               m_nitsche;
    T               m_hparam;
public:
    gsPoissonNitscheBoundaryAssembler (
            SystemMatrices<T>  &matrices,
            gsBasis<T>         &basis,
            const gsGeometry<T>      &geo,
            gsFunction<T>      *data,
            T                   nitsche,
            T                   hparam,
            patchSide          side
            )
        : gsScalarEASkeleton<T>(matrices,basis,geo,side.patch), m_dir_fun(data)
        , m_side(side), m_nitsche(nitsche), m_hparam(hparam)
    {
        m_basis->addFlags(NEED_GRAD | NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_OUTER_NORMAL );
        if (m_dir_fun) m_geo->addFlags(NEED_VALUE);
    }

    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,*m_geo);

        const index_t grad_size=m_basis->derivs().rows()/m_basis->actives().rows();

        gsVector<T> measures(quad_nodes.cols());
        gsMatrix<T> outer_normals;
        gsVector<T> out;
        m_geo->outerNormal(0,m_side.side(), out);
        outer_normals.resize(out.rows(),quad_nodes.cols());
        outer_normals.col(0)=out;
        measures(0)=out.norm()*quad_weights(0);
        for (index_t i=1;i<quad_nodes.cols();++i)
        {
            m_geo->outerNormal(i,m_side.side(), out);
            outer_normals.col(i)=out;
            measures(i) = out.norm()*quad_weights(i);
        }
        Eigen::DiagonalWrapper<Eigen::Matrix<T,Dynamic,1> > weights(measures);

        gsMatrix<T> mass_matrix = (m_nitsche/m_hparam)*m_basis->values()*weights*m_basis->values().transpose();
        m_sys.store(m_basis->actives(),m_basis->actives(),mass_matrix);

        // normal derivative contains area factor so we use different weights
        gsMatrix<T> N_matrix;
        N_matrix.setZero(m_basis->actives().cols(),m_basis->actives().cols());
        for (index_t col=0; col< m_basis->actives().cols(); ++col)
        {
            for (index_t row=0; row< m_basis->actives().cols(); ++row)
            {
                gsVector<T> normalDer = (m_basis->derivs().middleRows(row*grad_size,grad_size ).array()*outer_normals.array()).colwise().sum();
                N_matrix(row,col) = -(quad_weights.array()*normalDer.array()*m_basis->values().row(col).transpose().array()).sum();
            }
        }
        m_sys.store(m_basis->actives(),m_basis->actives(),N_matrix);
        m_sys.store(m_basis->actives(),m_basis->actives(),N_matrix.transpose());


        if (m_dir_fun)
        {
            gsMatrix<T> fun_values;
            m_dir_fun->eval_into(m_geo->values(),fun_values);
            gsMatrix<T> element_rhs =
                    (m_nitsche/m_hparam)*m_basis->values()*weights*fun_values.transpose();
            for (index_t row=0; row< m_basis->actives().cols(); ++row)
            {
                gsMatrix<T> r1 = (m_basis->derivs().middleRows(row*grad_size,grad_size ).array()*outer_normals.array()).colwise().sum();
                gsMatrix<T> r2 = (weights*fun_values.transpose());
                gsMatrix<T> rowValues = r1*r2;
                element_rhs.row(row) -= rowValues.row(0);
            }
            m_rhs.store(m_basis->actives(),m_basis->actives(),element_rhs);
        }
    }
};


template <typename T>
class gsPoissonDGInterfaceAssembler : public gsElementAssembler<T> , public gsScalarEASkeleton<T>
{
protected:
    using gsScalarEASkeleton<T>::m_sys;
    using gsScalarEASkeleton<T>::m_rhs;
    using gsScalarEASkeleton<T>::m_basis;
    using gsScalarEASkeleton<T>::m_geo;
    gsBasisEvaluator<T>        *m_basis2;
    gsGeometryEvaluator<T>     *m_geo2;
    boundaryInterface m_int;
    gsFunction<T>    *m_map;
    T                 m_penalty;
public:
    gsPoissonDGInterfaceAssembler (
            SystemMatrices<T>  &matrices,
            gsBasis<T>          &basis1,
            gsBasis<T>          &basis2,
            const gsGeometry<T> &geo1,
            const gsGeometry<T> &geo2,
            const boundaryInterface &interface,
            T                    penalty
            )
        : gsScalarEASkeleton<T>(matrices,basis1,geo1,interface.first().patch)
           , m_int(interface), m_penalty(penalty)
    {
        m_geo2   = geo2.evaluator(0);
        m_basis2 = makeBasisEvaluator(basis2);

        m_basis->addFlags(NEED_GRAD | NEED_VALUE);
        m_geo->addFlags(m_basis->getGeoFlags() | NEED_OUTER_NORMAL );

        m_basis2->addFlags(m_basis->getFlags());
        m_geo2->addFlags(m_basis2->getGeoFlags());

//        m_map = makeInterfaceMap(m_int.first().side(), m_int.second().side(),m_int.orient, geo1.support(), geo2.support());  // first() finer than second()
        //m_map = makeInterfaceMap(m_int.second(), m_int.first(), geo2.support(), geo1.support()); // second() finer than first()
    }
    ~gsPoissonDGInterfaceAssembler()
    {
        delete m_basis2;
        delete m_geo2;
        delete m_map;
    }


    virtual void operator() (const gsDomainIterator<T> &element, const gsMatrix<T> &quad_nodes, const gsVector<T> &quad_weights)
    {
        gsMatrix<T> pointsInOtherPatch;
        m_map->eval_into(quad_nodes,pointsInOtherPatch);

        m_geo->evaluateAt(quad_nodes);
        m_basis->evaluateAt(quad_nodes ,*m_geo);
        m_geo2->evaluateAt(pointsInOtherPatch);
        m_basis2->evaluateAt(pointsInOtherPatch ,*m_geo2);

  //      const index_t grad_size=m_basis->derivs().rows()/m_basis->actives().rows();

        gsVector<T> measures(quad_nodes.cols());
        gsMatrix<T> outer_normals;

        const index_t numActive1 = m_basis->actives().rows();
        const index_t numActive2 = m_basis2->actives().rows();

        gsVector<T> out;
        m_geo->outerNormal(0,m_int.first().side(), out);
        outer_normals.resize(out.rows(),quad_nodes.cols());
        outer_normals.col(0)=out;
        measures(0)=out.norm()*quad_weights(0);

        for (index_t i=1;i<quad_nodes.cols();++i)
        {
            m_geo->outerNormal(i,m_int.first().side(), out);
            outer_normals.col(i)=out/out.norm();
            measures(i) = out.norm()*quad_weights(i);
        }
        Eigen::DiagonalWrapper<Eigen::Matrix<T,Dynamic,1> > weights(measures);

        gsMatrix<T> P1, P2, P12;

        gsMatrix<T> B11, B12, E11, E12, N1, B22, B21, E22, E21, N2;

    ///

        B11.setZero(numActive1, numActive1); B12.setZero(numActive1, numActive2);
        E11.setZero(numActive1, numActive1); E12.setZero(numActive1, numActive2);
        B22.setZero(numActive2, numActive2); B21.setZero(numActive2, numActive1);
        E22.setZero(numActive2, numActive2); E21.setZero(numActive2, numActive1);

        for (index_t k = 0; k < quad_weights.rows(); ++k) // loop over quadrature nodes
        {

            // Compute element matrices
            const T c1     = measures(k) * T(0.5);

            N1.noalias()   = outer_normals.col(k).transpose() * m_basis->derivs();
            N2.noalias()   = outer_normals.col(k).transpose() * m_basis2->derivs();
            B11.noalias() += c1 * ( m_basis->values() * N1 );
            B12.noalias() += c1 * ( m_basis->values() * N2 );
            B22.noalias() -= c1 * ( m_basis2->values() * N2 );
            B21.noalias() -= c1 * ( m_basis2->values() * N1 );

            const T c2     = measures(k) * m_penalty / element.getCellSize();
            E11.noalias() += c2 * ( m_basis->values() * m_basis->values().transpose() );
            E12.noalias() += c2 * ( m_basis->values() * m_basis2->values().transpose() );
            E22.noalias() += c2 * ( m_basis2->values() * m_basis2->values().transpose() );
            E21.noalias() += c2 * ( m_basis2->values() * m_basis->values().transpose() );
        }


/*
        // Local Dofs to global dofs
        gsMatrix<unsigned> actives1, actives2;
        mapper.localToGlobal(m_basis->actives(), bi->first().patch, actives1);
        mapper.localToGlobal(m_basis2->actives(), bi->first().patch, actives2);
        // Push element contributions 1-2 to the global matrix and load vector
        for (index_t j=0; j!=numActive1; ++j)
        {
            const index_t jj1 = actives1(j); // N1_j

            for (index_t i=0; i!=numActive1; ++i)
            {
                const index_t  ii1 = actives1(i); // N1_i

                if ( jj1 <= ii1 )
                    sysMatrix( ii1, jj1 ) -=  B11(i,j) + B11(j,i) - E11(i,j);
            }

            for (index_t i=0; i!=numActive2; ++i)
            {
                const index_t  ii2 = actives2(i); // N2_i

                if ( jj1 <= ii2 )
                    sysMatrix( ii2, jj1)  -=  B21(i,j) + B12(j,i) + E21(i,j);
            }
        }

        // Push element contributions 2-1 to the global matrix and load vector
        for (index_t j=0; j!=numActive2; ++j)
        {
            const index_t jj2 = actives2(j); // N2_j

            for (index_t i=0; i!=numActive2; ++i)
            {
                const index_t  ii2 = actives2(i); // N2_i

                if ( jj2 <= ii2 )
                    sysMatrix( ii2, jj2 ) -=  B22(i,j) + B22(j,i) - E22(i,j);
            }

            for (index_t i=0; i!=numActive1; ++i)
            {
                const index_t  ii1 = actives1(i); // N1_i

                if ( jj2 <= ii1 )
                    sysMatrix( ii1, jj2)  -=  B12(i,j) + B21(j,i) + E12(i,j);
            }
        }
        */
    }
};



template <typename T>
class gsPoissonAssembler3 : public gsScalarPdeAssembler<T>
{
protected:
    using gsPdeAssembler<T>::m_matrices;
    using gsPdeAssembler<T>::m_pde;

    using gsScalarPdeAssembler<T>::m_dir_strategy;
    using gsScalarPdeAssembler<T>::m_interface_strategy;
    using gsScalarPdeAssembler<T>::m_basis;
public:
    gsPoissonAssembler3(
            const gsPoissonPde<T> &pde,
            const gsMultiBasis<T> &basis,
            dirichlet::strategy    dirichlet_strategy = dirichlet::nitsche,
            iFace::strategy        interface_strategy = iFace::glue
            )
        : gsScalarPdeAssembler<T>(pde,basis,dirichlet_strategy, interface_strategy)
    {}

    virtual gsElementAssembler<T> *getElementAssembler             (index_t patch)
    {
        return new gsPoissonElementAssembler<T>(m_matrices, m_basis[patch], m_pde.domain()[patch], static_cast<const gsPoissonPde<T> *>(&m_pde)->rhs(), patch);
    }
    virtual gsElementAssembler<T> *getElementAssemblerForBoundary  (const boundary_condition<T> &bc)
    {
        switch (bc.type())
        {
        case condition_type::dirichlet:
            switch (m_dir_strategy)
            {
            case dirichlet::elimination:
                return new gsScalarDirichletElementAssembler<T>(m_matrices, m_basis[bc.ps.patch], m_pde.domain()[bc.ps.patch], bc.function().get(), bc.ps);
                break;
            case dirichlet::nitsche:
                {
                gsBasis<T> &b=m_basis[bc.ps.patch];
                T nitscheC=( (b.degree(0)+b.dim())* (b.degree(0)+1) * 2.0 );
                T hparam = math::pow( (T) b.size(), -1.0 / b.dim() );
                return new gsPoissonNitscheBoundaryAssembler<T>(m_matrices, m_basis[bc.ps.patch], m_pde.domain()[bc.ps.patch], bc.function().get(), nitscheC, hparam, bc.ps);
                }
                break;
            case dirichlet::none:
                return NULL;
                break;
            default:
                GISMO_ERROR("Dirichlet strategy is not supported");
            }
            break;
        case condition_type::neumann:
            return new gsScalarNeumannElementAssembler<T>(m_matrices, m_basis[bc.ps.patch], m_pde.domain()[bc.ps.patch], bc.function().get(), bc.ps);
            break;
        case condition_type::robin:
        default:
            GISMO_ERROR("This boundary condition is not supported");
        }
    }
    virtual gsElementAssembler<T> *getElementAssemblerForInterface (const boundaryInterface  &bi)
    {
        switch (m_interface_strategy)
        {
            case iFace::glue:
                return NULL;
                break;
            case iFace::dg:
                gsWarn<<"Assuming mesh on patch 1 of interface is sufficiently refined to integrate on interface";
                return new gsPoissonDGInterfaceAssembler<T>( m_matrices,m_basis[bi.first().patch],m_basis[bi.second().patch],
                    m_pde.domain()[bi.first().patch],m_pde.domain()[bi.second().patch],bi, 2);
                break;
            default:
                GISMO_ERROR("Interfaces need to be extended to really support DG gluing");

        }
    }
};

}
