/** @file gsFunctionSet.h

    @brief This is the interface of all objects that computes functions
    on points like gsBasis, gsGeometry and gsFunctions.

    Related to this is the gsFuncData object that is a cache of computed
    values.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan, A. Mantzaflaris
*/

#pragma once

#include <gsCore/gsForwardDeclarations.h>
#include <gsCore/gsLinearAlgebra.h>

#include <gsCore/gsTransform.h>

#include <gsCore/gsFuncData.h>

namespace gismo {

/**
   \brief Interface for the evaluation of functions.
   All G+Smo objects that evaluate function[s] derive from this object.
   Examples are gsMatrix, gsFunction and gsGeometry.

   The available evaluation procedures are:

   Name of procedure            | Evaluate what
   -----------------------------|-------------------------------
   \c eval   (points)           | value
   \c deriv  (points)           | first derivative(s)
   \c deriv2 (points)           | second derivative(s)
   \c div    (points)           | divergence
   \c curl   (points)           | curl / rotor
   \c laplacian (points)        | laplacian

   There are two classes of functions for evaluation that differ in the arguments and in
   the returned values. The argument can either be

   1) a matrix;
   2) a gsTransform object representing the parametrization of the domain.

   In the first case the the matrix specifies a set of points (one point per column) and it
   is the caller responsibility to distinguish between functions defined on the physical and
   on the parametric domain. In the second case the evaluator will fetch the appropriate points
   depending on whether it is defined on the parametric or in the physical space.
   In complex situations: in particular when the physical domain of a function is the parametric
   domain of another the autamatic mechanism is not guaranteed to work.

   The result is a matrix in which the i-th column contains the requested values
   or derivatives at the i-th point, i.e. the point whose coordinate are the i-th column
   of the input matrix or of the matrix supplied by the gsTransform object.
   On each column the data is grouped in blocks corresponding to different functions,
   so that that if the requested evaluation contains s values
   \f$ v_1,\ldots,v_s \f$ for each pair \f$ (f_i,p_j) \f$ of (function,point) the output matrix looks like

   \f[
   \left[
   \begin{array}{ccccc}
   v_1(f_1,p_1) & v_1(f_1,p_2) & \ldots & v_1(f_1,p_n)\\
   v_2(f_1,p_1) & v_2(f_1,p_2) & \ldots & v_2(f_1,p_n)\\
   \vdots       & \vdots       &        & \vdots\\
   v_s(f_1,p_1) & v_s(f_1,p_2) & \ldots & v_s(f_1,p_n)\\
   v_1(f_2,p_1) & v_1(f_2,p_2) & \ldots & v_1(f_2,p_n)\\
   \vdots       & \vdots       &        & \vdots\\
   v_s(f_m,p_1) & v_s(f_m,p_2) & \ldots & v_s(f_m,p_n)
   \end{array}
   \right]
   \f]

   Any implementation of the gsFunctionSet interface must:

   1) overload the virtual function int size() that returns the
   number of functions in the set;

   2) fill in the gsFuncInfo data in a meaningful way.

   3) overload the neede evaluation functions with gsMatrix
   argument: eval_into, deriv_into, deriv2_into; the one that
   are not implemented will fail at runtime printing which
   function need implementing to the console

   and possibly:

   4) write optimized versions of div, curl, laplacian
   ... evaluation. By default they are computed by evaluating
   the derivatives and applying the definition.

   \tparam T type for real numbers

   \ingroup  function
*/
template <typename T>
class gsFunctionSet
{
public:
    typedef const gsFuncInfo Info;

public:

    /// @brief Clone this basis, making a deep copy.
    virtual gsFunctionSet * clone() const {GISMO_NO_IMPLEMENTATION;}

public:
    /**
     * @brief active_into returns the active (non-zero) functions
     * for each point, the columns are sorted in increasing order,
     * if on a point there are less active then the number of rows
     * in the result matrix (some other point has more actives)
     * then the rest of the column is filled with 0s.  @param
     * u @param result
     */
    virtual void active_into (const gsMatrix<T>  & u, gsMatrix<unsigned> &result) const;
    
public:
    /**
       @brief eval
       For scalar valued functions \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}\f$ format is:
       \f[
       \left[
       \begin{array}{ccccc}
       f_1(p_1) & f_1(p_2) & \ldots & f_1(p_n)\\
       f_2(p_1) & f_2(p_2) & \ldots & f_2(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       f_m(p_1) & f_m(p_2) & \ldots & f_m(p_n)
       \end{array}
       \right]
       \f]
       For vector valued functions function \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}^s\f$ the format is:
       \f[
       \left[
       \begin{array}{ccccc}
       f_1^1(p_1) & f_1^1(p_2) & \ldots & f_1^1(p_n)\\
       f_1^2(p_1) & f_1^2(p_2) & \ldots & f_1^2(p_n)\\
       \vdots     & \vdots     &        & \vdots\\
       f_1^s(p_1) & f_1^s(p_2) & \ldots & f_1^s(p_n)\\
       f_2^1(p_1) & f_2^1(p_2) & \ldots & f_2^1(p_n)\\
       \vdots     & \vdots     &        & \vdots\\
       f_m^s(p_1) & f_m^s(p_2) & \ldots & f_m^s(p_n)
       \end{array}
       \right]
       \f]
       where \f$f^i_j\f$ is the i-th component of function \f$f_j\f$.
       @param u
       @param result
    */
    virtual void eval_into      (const gsMatrix<T>  & u, gsMatrix<T> &result) const;

    /**
       @brief deriv
       For scalar valued functions \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}\f$ format is:
       \f[
       \left[
       \begin{array}{ccccc}
       \partial_{1}f_1(p_1) & \partial_{1}f_1(p_2) & \ldots & \partial_{1}f_1(p_n)\\
       \partial_{2}f_1(p_1) & \partial_{2}f_1(p_2) & \ldots & \partial_{2}f_1(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       \partial_{k}f_1(p_1) & \partial_{k}f_1(p_2) & \ldots & \partial_{k}f_1(p_n)\\
       \partial_{1}f_2(p_1) & \partial_{1}f_2(p_2) & \ldots & \partial_{1}f_2(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       \partial_{k}f_m(p_1) & \partial_{k}f_m(p_2) & \ldots & \partial_{k}f_m(p_n)\\
       \end{array}
       \right]
       \f]
       For vector valued functions function \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}^s\f$ the format is:
       \f[
       \left[
       \begin{array}{ccccc}
       \partial_{1}f_1^1(p_1) & \partial_{1}f_1^1(p_2) & \ldots & \partial_{1}f_1^1(p_n)\\
       \partial_{2}f_1^1(p_1) & \partial_{1}f_2^1(p_2) & \ldots & \partial_{1}f_2^1(p_n)\\
       \vdots     & \vdots     &        & \vdots\\
       \partial_{k}f_1^1(p_1) & \partial_{k}f_1^1(p_2) & \ldots & \partial_{k}f_1^1(p_n)\\
       \partial_{1}f_1^2(p_1) & \partial_{1}f_1^2(p_2) & \ldots & \partial_{1}f_1^2(p_n)\\
       \vdots     & \vdots     &        & \vdots\\
       \partial_{k}f_1^2(p_1) & \partial_{k}f_1^2(p_2) & \ldots & \partial_{k}f_1^2(p_n)\\
       \partial_{1}f_2^1(p_1) & \partial_{1}f_2^1(p_2) & \ldots & \partial_{1}f_2^1(p_n)\\
       \vdots     & \vdots     &        & \vdots\\
       \partial_{k}f_m^s(p_1) & \partial_{k}f_m^s(p_2) & \ldots & \partial_{k}f_m^s(p_n)
       \end{array}
       \right]
       \f]
       where \f$f^i_j\f$ is the i-th component of function \f$f_j\f$.
       @param u
       @param result
    */
    virtual void deriv_into     (const gsMatrix<T>  &  u, gsMatrix<T> &result) const;

    /**
     * @brief deriv2
     For scalar valued functions \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}\f$ format is:
     \f[
     \left[
     \begin{array}{ccccc}
     \partial_{1}\partial_{1}f_1(p_1) & \partial_{1}\partial_{1}f_1(p_2) & \ldots & \partial_{1}\partial_{1}f_1(p_n)\\
     \partial_{2}\partial_{2}f_1(p_1) & \partial_{2}\partial_{2}f_1(p_2) & \ldots & \partial_{2}\partial_{2}f_1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k}\partial_{k}f_1(p_1) & \partial_{k}\partial_{k}f_1(p_2) & \ldots & \partial_{k}\partial_{k}f_1(p_n)\\
     \partial_{1}\partial_{2}f_1(p_1) & \partial_{1}\partial_{2}f_1(p_2) & \ldots & \partial_{1}\partial_{2}f_1(p_n)\\
     \partial_{1}\partial_{3}f_1(p_1) & \partial_{1}\partial_{3}f_1(p_2) & \ldots & \partial_{1}\partial_{3}f_1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{1}\partial_{k}f_1(p_1) & \partial_{1}\partial_{k}f_1(p_2) & \ldots & \partial_{1}\partial_{k}f_1(p_n)\\
     \partial_{2}\partial_{3}f_1(p_1) & \partial_{2}\partial_{3}f_1(p_2) & \ldots & \partial_{2}\partial_{3}f_1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{2}\partial_{k}f_1(p_1) & \partial_{2}\partial_{k}f_1(p_2) & \ldots & \partial_{2}\partial_{k}f_1(p_n)\\
     \partial_{3}\partial_{4}f_1(p_1) & \partial_{3}\partial_{4}f_1(p_2) & \ldots & \partial_{3}\partial_{4}f_1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k-1}\partial_{k}f_1(p_1) & \partial_{k-1}\partial_{k}f_1(p_2) & \ldots & \partial_{k-1}\partial_{k}f_1(p_n)\\
     \partial_{1}\partial_{1}f_2(p_1) & \partial_{1}\partial_{1}f_2(p_2) & \ldots & \partial_{1}\partial_{1}f_2(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k-1}\partial_{k}f_m(p_1) & \partial_{k-1}\partial_{k}f_m(p_2) & \ldots & \partial_{k-1}\partial_{k}f_m(p_n)\\
     \end{array}
     \right]
     \f]
     For vector valued functions function \f$f_1, \ldots, f_m\f$ from \f$\mathbb{R}^k\rightarrow\mathbb{R}^s\f$ the format is:
     \f[
     \left[
     \begin{array}{ccccc}
     \partial_{1}\partial_{1}f_1^1(p_1) & \partial_{1}\partial_{1}f_1^1(p_2) & \ldots & \partial_{1}\partial_{1}f_1^1(p_n)\\
     \partial_{2}\partial_{2}f_1^1(p_1) & \partial_{2}\partial_{2}f_1^1(p_2) & \ldots & \partial_{2}\partial_{2}f_1^1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k}\partial_{k}f_1^1(p_1) & \partial_{k}\partial_{k}f_1^1(p_2) & \ldots & \partial_{k}\partial_{k}f_1^1(p_n)\\
     \partial_{1}\partial_{2}f_1^1(p_1) & \partial_{1}\partial_{2}f_1^1(p_2) & \ldots & \partial_{1}\partial_{2}f_1^1(p_n)\\
     \partial_{1}\partial_{3}f_1^1(p_1) & \partial_{1}\partial_{3}f_1^1(p_2) & \ldots & \partial_{1}\partial_{3}f_1^1(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k-1}\partial_{k}f_1^1(p_1) & \partial_{k-1}\partial_{k}f_1^1(p_2) & \ldots & \partial_{k-1}\partial_{k}f_1^1(p_n)\\
     \partial_{1}\partial_{1}f_1^2(p_1) & \partial_{1}\partial_{1}f_1^2(p_2) & \ldots & \partial_{1}\partial_{1}f_1^2(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k-1}\partial_{k}f_1^s(p_1) & \partial_{k-1}\partial_{k}f_1^s(p_2) & \ldots & \partial_{k-1}\partial_{k}f_1^s(p_n)\\
     \vdots       & \vdots       &        & \vdots\\
     \partial_{k-1}\partial_{k}f_m^s(p_1) & \partial_{k-1}\partial_{k}f_m^s(p_2) & \ldots & \partial_{k-1}\partial_{k}f_m^s(p_n)\\
     \end{array}
     \right]
     \f]
     where \f$ f^i_j\f$ is the i-th component of function \f$ f_j\f$.
     @param u
     @param result
    */
    virtual void deriv2_into    (const gsMatrix<T>  &  u, gsMatrix<T> &result) const;

    /// @brief Evaluate the nonzero functions and their derivatives up
    /// to order \a n
    virtual void evalAllDers_into(const gsMatrix<T> & u,    int n,
                                  std::vector<gsMatrix<T> >& result) const;

    /**
       @brief div
       This makes sense only for vector valued functions whose target dimension is a multiple of the domain dimension
       \f[\left[
       \begin{array}{ccccc}
       \text{div} f_1(p_1) & \text{div} f_1(p_2) & \ldots & \text{div} f_1(p_n)\\
       \text{div} f_2(p_1) & \text{div} f_2(p_2) & \ldots & \text{div} f_2(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       \text{div} f_m(p_1) & \text{div} f_m(p_2) & \ldots & \text{div} f_m(p_n)
       \end{array}
       \right]
       \f]
       If the target dimension is an integer multiple of the domain dimension \f$\text{div} f(p)\f$ contains
       (target dim)/(domain dim) rows, each row is the divergence of an (domain dim)-tuple of entries of f.
       Equivalently f is interpreted as a matrix valued function in column maior standard and the divergence corresponding vector valued divergence is returned.
       If the target dimension is not an integer multiple of the domain dimension calling this function is an error.
       @param u
       @param result
    */
    virtual void div_into       (const gsMatrix<T>  & u, gsMatrix<T> &result) const;

    /**
       @brief curl
       This makes sense only for vector valued functions whose target dimension is a multiple of the domain dimension
       \f[\left[
       \begin{array}{ccccc}
       \text{curl} f_1(p_1) & \text{curl} f_1(p_2) & \ldots & \text{curl} f_1(p_n)\\
       \text{curl} f_2(p_1) & \text{curl} f_2(p_2) & \ldots & \text{curl} f_2(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       \text{curl} f_m(p_1) & \text{curl} f_m(p_2) & \ldots & \text{curl} f_m(p_n)
       \end{array}
       \right]
       \f]
       If the target dimension is an integer multiple of the domain dimension \f$\text{curl} f(p)\f$ contains
       (target dim)/(domain dim) rows, each row is the rotor of an (domain dim)-tuple of entries of f.
       Equivalently \f$ f\f$ is interpreted as a matrix valued function in column maior standard and the rotor corresponding vector valued rotor is returned.
       If the target dimension is not an integer multiple of the domain dimension calling this function is an error.
       @param u
       @param result
    */
    virtual void curl_into      (const gsMatrix<T> &  u, gsMatrix<T> &result) const;

    /**
       @brief laplacian
       \f[\left[
       \begin{array}{ccccc}
       \Delta f_1(p_1) & \Delta f_1(p_2) & \ldots & \Delta f_1(p_n)\\
       \Delta f_2(p_1) & \Delta f_2(p_2) & \ldots & \Delta f_2(p_n)\\
       \vdots       & \vdots       &        & \vdots\\
       \Delta f_m(p_1) & \Delta f_m(p_2) & \ldots & \Delta f_m(p_n)
       \end{array}
       \right]
       \f]
       If the target dimension is different from 1 then each \f$\Delta f_i(p_j)\f$ is a block of target dimension rows containing the component wise Laplacian.
       @param u
       @param result
    */
    virtual void laplacian_into (const gsMatrix<T>  &  u, gsMatrix<T> &result) const;


    virtual void compute(const gsMatrix<T>  &u,   gsFuncData<T> & result) const;
    virtual void compute(const gsMapData<T> &geometry, gsFuncData<T> & result) const;

// move to gsFunction
     virtual void compute(const gsMatrix<T>  &u,   gsMapData<T>  & result) const;
public:
    /**
       @brief domainDim
       @return the dimension of the domain: for \f$f:\mathbb{R}^n\rightarrow\mathbb{R}^m\f$ returns \f$n\f$
     */
    virtual int domainDim () const = 0;

    /**
       @brief targetDim
       @return the dimension of the domain: for \f$f:\mathbb{R}^n\rightarrow\mathbb{R}^m\f$ returns \f$m\f$
     */
    virtual int targetDim () const = 0;

    /**
       @brief info
       @return a gsFuncInfo struct containing domainDim and targetDim
     */
    gsFuncInfo info() const {return gsFuncInfo(domainDim(),targetDim());}

    /**
     * @brief size
     *
     * \warning Beware of bugs: gsGeometry has size() == 1
     *
     * @return the size of the function set: the number of functions
     *
     */
    virtual index_t size() const = 0;

    ////
    virtual unsigned getGeoFlags (unsigned flags) const {return 0;}
};



/**
    @brief helper class for subclasses that only implement the compute interface,
    other calls will be redirected to compute.

    If you derive from this class you only need to implement

    virtual void    compute  (const gsMatrix <T> &points,  gsFuncData<T> &result) const = 0;
    virtual void    compute  (const gsMapData<T> &geoData, gsFuncData<T> &result) const = 0;
    virtual index_t size     () const = 0;

   and to initialize the m_info correctly
**/
template <typename T>
class gsFuncSetComputeOnly : public gsFunctionSet<T>
{
public:
    virtual void active_into    (const gsMatrix<T> &points, gsMatrix<unsigned> &result) const
    {
        gsFuncData<T> output    (NEED_ACTIVE);
        compute(points,output);
        result=output.actives;
    }

    virtual void eval_into      (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_VALUE);
        compute(points,output);
        std::swap(result,output.values[0]);
    }
    virtual void deriv_into     (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_DERIV);
        compute(points,output);
        std::swap(result,output.values[1]);
    }
    virtual void deriv2_into    (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_DERIV2);
        compute(points,output);
        std::swap(result,output.values[2]);
    }
    virtual void div_into       (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_DIV);
        compute(points,output);
        std::swap(result,output.divs);
    }
    virtual void curl_into      (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_CURL);
        compute(points,output);
        std::swap(result,output.curls);
    }
    virtual void laplacian_into (const gsMatrix<T> &points, gsMatrix<T> &result) const
    {
        gsFuncData<T> output    (NEED_LAPLACIAN);
        compute(points,output);
        std::swap(result,output.laplacians);
    }

    virtual void compute        (const gsMatrix <T> &points,  gsFuncData<T> &result) const = 0;
    virtual void compute        (const gsMapData<T> &geoData, gsFuncData<T> &result) const = 0;
};



#ifndef GISMO_BUILD_LIB
#include GISMO_HPP_HEADER(gsFunctionSet.hpp)
#endif

} // namespace gismo

