/** @file gsFunctionWithDerivatives.h

    @brief A function with explicitly set derivatives

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan
**/

#include <gsCore/gsFunction.h>


#pragma once

namespace gismo {

template <typename T=real_t>
class gsFunctionWithDerivatives : public gsFunction<T>
{
protected:
    gsFunction<T> *m_values;
    gsFunction<T> *m_derivs;
    gsFunction<T> *m_deriv2;
public:
    gsFunctionWithDerivatives(const gsFunction<T> &func)
        :
          m_values(func.clone()),
          m_derivs(NULL),
          m_deriv2(NULL)
    {}


    gsFunctionWithDerivatives(const gsFunction<T> &func, const gsFunction<T> &deriv)
        :
          m_values(func.clone()),
          m_derivs(deriv.clone()),
          m_deriv2(NULL)
    {
        checkDimensions();
    }


    gsFunctionWithDerivatives(const gsFunction<T> &func, const gsFunction<T> &deriv, const gsFunction<T> &deriv2)
        :
          m_values(func.clone()),
          m_derivs(deriv.clone()),
          m_deriv2(deriv2.clone())
    {
        checkDimensions();
    }

    ~gsFunctionWithDerivatives()
    {
        delete m_values;
        delete m_derivs;
        delete m_deriv2;
    }

    void eval_into(const gsMatrix<T> &u, gsMatrix<T> &result) const
    {
        m_values->eval_into(u,result);
    }
    void deriv_into(const gsMatrix<T> &u, gsMatrix<T> &result) const
    {
        GISMO_ASSERT(m_derivs,"CANNOT COMPUTE DERIVATIVES");
        m_derivs->eval_into(u,result);
    }
    void deriv2_into(const gsMatrix<T> &u, gsMatrix<T> &result) const
    {
        GISMO_ASSERT(m_deriv2,"CANNOT COMPUTE SECOND DERIVATIVES");
        m_deriv2->eval_into(u,result);
    }

    gsMatrix<T>* laplacian(const gsMatrix<T> &u) const
    {
        gsMatrix<T> *result=new gsMatrix<T>();
        gsMatrix<T>  secDer;
        deriv2_into(u,secDer);
        *result = secDer.topRows(m_values->domainDim()).colwise().sum();
        return result;
    }

    const gsFunction<T>* getDeriv () const
    {
        return m_derivs;
    }
    const gsFunction<T>* getDeriv2 () const
    {
        return m_deriv2;
    }

    gsFunction<T>* clone() const
    {
        if(m_deriv2)
            return new gsFunctionWithDerivatives(*m_values, *m_derivs,*m_deriv2);
        else if(m_derivs)
            return new gsFunctionWithDerivatives(*m_values, *m_derivs);
        else
            return new gsFunctionWithDerivatives(*m_values);
    }

    int targetDim () const
    {
        return m_values->targetDim();
    }
    int domainDim () const
    {
        return m_values->domainDim();
    }
private:
    void checkDimensions()
    {
        bool ok=true;

        const int parDim=m_values->domainDim();
        const int tarDim=m_values->targetDim();

        if (m_derivs)
        {
            ok = ok &&  m_derivs->domainDim() == parDim;
            ok = ok &&  m_derivs->targetDim() == parDim*tarDim;
        }
        if (m_deriv2)
        {
            ok = ok &&  m_deriv2->domainDim() == parDim;
            ok = ok &&  m_deriv2->targetDim() == (parDim-1)*parDim*tarDim/2;
        }
        GISMO_ASSERT(ok, "INCOMPATIBLE FUNCTION FOR VALUES AND DERIVATIVES" );
    }
};


}
