/**
 * @file gsTransformPure.hpp
 *
 * @brief Implementation of the mapping of functions between the parametric
 *   and the physical domain.
 *
 *   This file is part of the G+Smo library.
 *
 *   This Source Code Form is subject to the terms of the Mozilla Public
 *   License, v. 2.0. If a copy of the MPL was not distributed with this
 *   file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 *   Author(s): A. Bressan, A. Manzaflaris
**/

#include <gsCore/gsFuncData.h>
#include <gsCore/gsEvalUtils.hpp>

namespace gismo {


/*
 * COMMON SCHEMA OF A PURE TRANSFORMATION OBJECT
template <typename T>
class gsTransformPure // No transformation
{
public:
    static unsigned getFuncFlags  (unsigned flags) {return flags;}
    static unsigned getGeoFlags (unsigned flags)   {return 0;}

    static void transform   (const gsMapData<T> &geoData,
                                    const gsFuncData<T> &parmFunc,
                                    gsFuncData<T> &phisFunc)
    {
        ...
    }

    static void transformInPlace (const gsMapData<T> &geoData, gsFuncData<T> &data)
    {
        gsFuncData<T> temp;
        std::swap(data,temp);
        transform(geoData,temp,data);
    }
};
*/


template <typename T>
class gsTransformPureGradConforming
{
public:
    static unsigned getFuncFlags (unsigned flags)
    {
        unsigned result=flags& ~(NEED_CURL|NEED_DIV|NEED_LAPLACIAN);
        result |= flags&NEED_CURL         ? NEED_DERIV  : 0;
        result |= flags&NEED_DIV          ? NEED_DERIV  : 0;
        result |= flags&NEED_LAPLACIAN    ? NEED_DERIV2 : 0;
        return result;
    }
    static unsigned getGeoFlags  (unsigned flags)
    {
        unsigned result=0;
        result |= flags&NEED_VALUE     ? 0 : 0;
        result |= flags&NEED_DERIV     ? NEED_GRAD_TRANSFORM : 0;
        result |= flags&NEED_DERIV2    ? NEED_GRAD_TRANSFORM|NEED_DERIV2 : 0;
        result |= flags&NEED_DIV       ? NEED_GRAD_TRANSFORM : 0;
        result |= flags&NEED_CURL      ? NEED_GRAD_TRANSFORM : 0;
        result |= flags&NEED_LAPLACIAN ? NEED_GRAD_TRANSFORM : 0;
        return result;
    }

    static void transform   (const gsMapData<T>  &geoData,
                             const gsFuncData<T> &parmFunc,
                                   gsFuncData<T> &phisFunc)
    {
        const unsigned flags=phisFunc.flags;
        phisFunc.info.domainDim=geoData.info.targetDim;
        phisFunc.info.targetDim=parmFunc.info.targetDim;

        phisFunc.values.resize(parmFunc.values.size());
        if (flags&NEED_VALUE)
            phisFunc.values[0]=parmFunc.values[0];
        if (flags&NEED_DERIV)
        {
            transformDeriv(geoData,parmFunc,phisFunc);
            if (flags&NEED_DIV)
                convertValue<T>::derivToDiv(phisFunc.values[1],phisFunc.divs,phisFunc.info);
            if (flags&NEED_CURL)
                convertValue<T>::derivToCurl(phisFunc.values[1],phisFunc.curls,phisFunc.info);
        }
        if (flags&NEED_DERIV2)
        {
            transformDeriv2(geoData,parmFunc,phisFunc);
            if (flags&NEED_LAPLACIAN)
                convertValue<T>::deriv2ToLaplacian(phisFunc.values[2],phisFunc.laplacians,phisFunc.info);
        }
        else if (flags&NEED_LAPLACIAN)
        {
            // optimized transformation for Laplacian that does not need the full
            // transformation of the deriv2
            transformLaplacian(geoData,parmFunc,phisFunc);
        }
    }

    static inline void transformDeriv(const gsMapData<T>  &geoData,
                                      const gsFuncData<T> &parmFunc,
                                            gsFuncData<T> &phisFunc)
    {
        const index_t numPoints=geoData.points.cols();
        const index_t paramSize=parmFunc.info.domainDim;
        const index_t physiSize=phisFunc.info.domainDim;

        const index_t derivNum =parmFunc.values[1].rows()/paramSize;

        phisFunc.values[1].resize(physiSize*derivNum,numPoints);
        for (index_t p=0; p<numPoints; ++p)
        {
            phisFunc.values[1].reshapeCol(p,physiSize,derivNum)=
                    geoData.fundForm(p)*parmFunc.values[1].reshapeCol(p,paramSize,derivNum);
        }
    }

    static inline void transformDeriv2(const gsMapData<T>  &geoData,
                                       const gsFuncData<T> &prmFunc,
                                             gsFuncData<T> &phsFunc)
    {
        // UNTESTED
        const index_t numPoints=geoData.points.cols();
        const index_t prmDomDim =  prmFunc.info.domainDim;
        const index_t phsDomDim =  phsFunc.info.domainDim;

        const index_t prmSecDirSize = prmDomDim*(prmDomDim+1)/2;
        const index_t phsSecDirSize = phsDomDim*(phsDomDim+1)/2;

        const index_t numBlocks = prmFunc.values[2].rows()/prmSecDirSize;

        phsFunc.values[2].resize(phsSecDirSize*numBlocks,numPoints);

        gsMatrix<T> cmpHessian(prmDomDim,prmDomDim); // hessian of a component of the function
        gsMatrix<T> geoHessianMod(phsSecDirSize,numBlocks); // hessian of a component of the parametrization

        for (index_t p=0; p<numPoints; ++p)
        {
            // First part: J^-T H J^-1
            for (index_t b=0; b<numBlocks;++b)
            {
                convertValue<T>::deriv2ToHessianSingle(
                            prmFunc.values[2].block(b*prmSecDirSize,p,prmSecDirSize,1),
                            cmpHessian,prmDomDim);
                convertValue<T>::hessianToDeriv2Single(
                            geoData.fundForm(p)*cmpHessian*geoData.fundForm(p).transpose(),
                            phsFunc.values[2].block(b*phsSecDirSize,p,phsSecDirSize,1), phsDomDim);
            }
            // Second part: J^-T G J^-1 DDG J^-1
            for (index_t c=0;c<phsDomDim; ++c)
            {
                convertValue<T>::deriv2ToHessianSingle(
                            geoData.values[2].block(c*prmSecDirSize,p,prmSecDirSize,1),
                            cmpHessian,prmDomDim);
                convertValue<T>::hessianToDeriv2Single(
                            geoData.fundForm(p)*cmpHessian*geoData.fundForm(p).transpose(),
                            geoHessianMod.col(p),
                            phsDomDim);
            }

            phsFunc.values[2].reshapeCol(p,phsSecDirSize,numBlocks).noalias() -=
                    geoHessianMod*geoData.fundForm(p)
                    *phsFunc.values[1].reshapeCol(p,prmDomDim,numBlocks);
        }
    }

    static inline void transformLaplacian(const gsMapData<T>  &geoData,
                                          const gsFuncData<T> &parmFunc,
                                                gsFuncData<T> &phisFunc)
    {
        // to be taken from gsGeometryEvaluator.hpp
        GISMO_ERROR("transformLaplacian NOT IMPLEMENTED");

    }

    static void transformInPlace (const gsMapData<T> &geoData, gsFuncData<T> &data)
    {
        gsFuncData<T> temp;
        std::swap(data,temp);
        transform(geoData,temp,data);
    }
};


template <typename T>
class gsTransformPureDivConforming
{
public:
    static unsigned getFuncFlags (unsigned flags)
    {
        unsigned result=0;
        result |= flags&NEED_VALUE  ? NEED_VALUE : 0;

        result |= flags&NEED_CURL   ? NEED_DERIV : 0;
        result |= flags&NEED_DERIV  ? NEED_DERIV : 0;
        result |= ((flags&NEED_DIV) && !(result&NEED_DERIV) ) ?  NEED_DIV : 0;

        if (flags&NEED_DERIV2 || flags&NEED_LAPLACIAN) GISMO_ERROR("cannot computer 3rd derivative");
        return result;
    }
    static unsigned getGeoFlags  (unsigned flags)
    {
        unsigned result=0;
        result |= flags&NEED_VALUE  ? NEED_GRAD_TRANSFORM|NEED_MEASURE : 0;

        flags  |= flags&NEED_CURL   ? NEED_DERIV : 0;
        result |= flags&NEED_DERIV  ? NEED_GRAD_TRANSFORM|NEED_DERIV2 : 0;

        if (flags&NEED_DERIV2 || flags&NEED_LAPLACIAN) GISMO_ERROR("cannot computer 3rd derivative");
        return result;
    }

    static void transform   (const gsMapData<T>  &geoData,
                             const gsFuncData<T> &parmFunc,
                                   gsFuncData<T> &phisFunc)
    {
        const unsigned flags=phisFunc.flags;
        phisFunc.info.domainDim=geoData.info.targetDim;
        phisFunc.info.targetDim=parmFunc.info.targetDim;

        phisFunc.values.resize(parmFunc.values.size());
        if (flags&NEED_VALUE)
            transformValues(geoData,parmFunc,phisFunc);
        if (flags&NEED_DERIV)
        {
            transformDeriv(geoData,parmFunc,phisFunc);
            if (flags&NEED_DIV)
                convertValue<T>::derivToDiv(phisFunc.values[1],phisFunc.divs,phisFunc.info);
            if (flags&NEED_CURL)
                convertValue<T>::derivToCurl(phisFunc.values[1],phisFunc.curls,phisFunc.info);
        }
        else if (flags&NEED_DIV)
            transformDiv(geoData,parmFunc,phisFunc); // optimized without computing the full deriv transformation

        if (flags&NEED_DERIV2)
        {
            transformDeriv2(geoData,parmFunc,phisFunc);
            if (flags&NEED_LAPLACIAN)
                convertValue<T>::deriv2ToLaplacian(phisFunc.values[2],phisFunc.laplacians,phisFunc.info);
        }
    }

    static inline void transformValues(const gsMapData<T>  &geoData,
                                       const gsFuncData<T> &parmFunc,
                                             gsFuncData<T> &phisFunc)
    {
        const index_t paramSize=parmFunc.info.domainDim;
        const index_t physiSize=phisFunc.info.domainDim;
        const index_t numValues = parmFunc.values[0].rows()/paramSize;
        const index_t numPoints = geoData.points.cols();

        phisFunc.values[0].resize(numValues*physiSize,numPoints);
        for (index_t p=0; p< numPoints;++p)
        {
            phisFunc.values[0].reshapeCol(p,physiSize,numValues)=
                    geoData.fundForm(p)*parmFunc.values[0].reshapeCol(p,paramSize,numValues)/geoData.measures(0,p);
        }
    }

    static inline void transformDeriv(const gsMapData<T>  &geoData,
                                      const gsFuncData<T> &parmFunc,
                                            gsFuncData<T> &phisFunc)
    {
        // to be taken from gsGeometryEvaluator.hpp
        GISMO_ERROR("transformDeriv NOT IMPLEMENTED");
    }

    static inline void transformDeriv2(const gsMapData<T>  &geoData,
                                       const gsFuncData<T> &parmFunc,
                                             gsFuncData<T> &phisFunc)
    {
        // to be taken from gsGeometryEvaluator.hpp
        GISMO_ERROR("transformDeriv2 NOT IMPLEMENTED");
    }
    static inline void transformDiv(const gsMapData<T>  &geoData,
                                    const gsFuncData<T> &parmFunc,
                                          gsFuncData<T> &phisFunc)
    {
        // to be taken from gsGeometryEvaluator.hpp
        GISMO_ERROR("transformDiv NOT IMPLEMENTED");
    }

    static void transformInPlace (const gsMapData<T> &geoData, gsFuncData<T> &data)
    {
        gsFuncData<T> temp;
        std::swap(data,temp);
        transform(geoData,temp,data);
    }
};


/**
 * To restrict a function defined on \f$\mathbb{R}^n\f$ to a lower dimensional
 * manifold parametrized by the geometry.
 *
 * This transformation project the derivatives on the tangent plane so that
 * the orthogonal component of the derivatives is NULL.
 *
 * The use is to compute norms on surfaces.
 */
template <typename T>
class gsTransformPureRestriction
{
public:
    static unsigned getFuncFlags (unsigned flags)
    {
        unsigned result=flags& ~(NEED_CURL|NEED_DIV|NEED_LAPLACIAN);
        result |= flags&NEED_CURL         ? NEED_DERIV  : 0;
        result |= flags&NEED_DERIV        ? NEED_DERIV  : 0;
        result |= flags&NEED_LAPLACIAN    ? NEED_DERIV2 : 0;
        return result;
    }
    static unsigned getGeoFlags  (unsigned flags)
    {
        unsigned result=0;
        result = (flags& (NEED_DIV|NEED_CURL|NEED_LAPLACIAN|NEED_DERIV|NEED_DERIV2)) ? NEED_DERIV|NEED_GRAD_TRANSFORM : 0;
        return result;
    }

    static void transform   (const gsMapData<T>  &geoData,
                             const gsFuncData<T> &original,
                             gsFuncData<T>       &restricted)
    {
        const unsigned flags=restricted.flags;
        restricted.info.domainDim=geoData.info.targetDim;
        restricted.info.targetDim=original.info.targetDim;

        restricted.values.resize(original.values.size());
        if (flags&NEED_VALUE)
            restricted.values[0]=original.values[0];
        if (flags&NEED_DERIV)
        {
            transformDeriv(geoData,original,restricted);
            if (flags&NEED_DIV)
                convertValue<T>::derivToDiv(restricted.values[1],restricted.divs,restricted.info);
            if (flags&NEED_CURL)
                convertValue<T>::derivToCurl(restricted.values[1],restricted.curls,restricted.info);
        }
        else if (flags&NEED_DIV)
            transformDiv(geoData,original,restricted); // optimized without computing the full deriv transformation

        if (flags&NEED_DERIV2)
        {
            transformDeriv2(geoData,original,restricted);
            if (flags&NEED_LAPLACIAN)
                convertValue<T>::deriv2ToLaplacian(restricted.values[2],restricted.laplacians,restricted.info);
        }
    }

    static inline void transformDeriv(const gsMapData<T>  &geoData,
                                      const gsFuncData<T> &original,
                                      gsFuncData<T>       &restricted)
    {
        const index_t numPoints=geoData.points.cols();
        const index_t derivSize=original.info.domainDim;
        const index_t derivNum =original.values[1].rows()/derivSize;

        restricted.values[1].resizeLike(original.values[1]);
        for (index_t p=0; p<numPoints; ++p)
        {
            restricted.values[1].reshapeCol(p,derivSize,derivNum)=
                    geoData.fundForm(p)*geoData.jacobian(p)*original.values[1].reshapeCol(p,derivSize,derivNum);
        }
    }

    static inline void transformDeriv2(const gsMapData<T>  &geoData,
                                       const gsFuncData<T> &original,
                                       gsFuncData<T>       &restricted)
    {

        const index_t numPoints=geoData.points.cols();
        const index_t domainDim=original.info.domainDim;
        const index_t derivSize=original.info.domainDim;
        const index_t numBlocks=original.values[2].rows()/derivSize;

        restricted.values[1].resizeLike(original.values[2]);

        gsMatrix<T> temp;
        gsMatrix<T> cmpHessian;

        for (index_t p=0; p<numPoints; ++p)
        {
            temp=geoData.fundForm(p)*geoData.jacobian(p); // projection matrix
            for (index_t b=0; b<numBlocks;++b)
            {
                convertValue<T>::deriv2ToHessianSingle(
                            original.values[2].block(b*derivSize,p,derivSize,1),
                            cmpHessian,domainDim);
                convertValue<T>::hessianToDeriv2Single(
                            temp.transpose()*cmpHessian*temp(),
                            restricted.values[2].block(b*derivSize,p,derivSize,1), domainDim);
            }
        }
    }

    static void transformInPlace (const gsMapData<T> &geoData, gsFuncData<T> &data)
    {
        gsFuncData<T> temp;
        std::swap(data,temp);
        transform(geoData,temp,data);
    }
};



} // namespace gismo
