/**
 * @file gsPhysicalEvaluator.h
 *
 * \brief The physical evaluator is an object the wraps a parametric evaluator and
 *        and automatically computes the transformed values on the physical domain.
 *
 *        At the moment only 2 physical evaluators are provided:
 *          1) grad conforming for which the space on the physical domain is defined by
 *             \f[ \phi(x) = \hat\phi(\G^{-1}(x) ) \f], where \f[\phi\f] is the
 *             mapped function and  \f[G\f] is the parametrization of the physical domain.
 *             This is the most common mapping in IGA.
 *          2) div conforming for which the (vector valued) space on the physical domain is defined by
 *             \f[\phi(x) =  detJ^{-1}(x) J(\G^{-1}(x))\hat\phi(\G^{-1}(x) ) \f], where \f[\phi\f] is the
 *             mapped function, \f[G\f] is the parametrization of the physical domain and \f[J\f] the
 *             Jacobian of the parametrization.
 *
 *   This file is part of the G+Smo library.
 *
 *   This Source Code Form is subject to the terms of the Mozilla Public
 *   License, v. 2.0. If a copy of the MPL was not distributed with this
 *   file, You can obtain one at http://mozilla.org/MPL/2.0/.
 *
 *   Author(s): A. Bressan
**/

#pragma once

#include <gsCore/gsTransformPure.h>
#include <gsCore/gsFunctionSet.h>
#include <gsCore/gsFuncData.h>

namespace gismo
{

template <typename T, typename transform=gsTransformPureGradConforming<T> >
class gsTransformedFuncSet : public gsFuncSetComputeOnly<T>
{
protected:
    const gsFunctionSet<T>  &m_funValue;
    const gsFunctionSet<T>  &m_geoValue;
public:

    gsTransformedFuncSet (const gsFunctionSet<T> &func, const gsFunctionSet<T> &geo)
        : m_funValue(func), m_geoValue(geo)
    {
        GISMO_ASSERT(geo.size()==1,"it is impossible to transform a function set with a multi map");
    }

public:
    virtual unsigned getGeoFlags (unsigned flags) const {return transform::getGeoFlags(flags);}

    virtual index_t size () const {return m_funValue.size();}

    virtual int domainDim () const {return m_funValue.domainDim();}
    virtual int targetDim () const {return m_funValue.targetDim();}

    virtual void compute        (const gsMatrix <T> &points, gsFuncData<T> &result) const
    {
        gsFuncData<T> paramData (transform::getFuncFlags(result.flags));
        gsMapData<T>  geoData   (transform::getGeoFlags (result.flags));
        m_funValue.compute(points,paramData);
        m_geoValue.compute(points,geoData);
        transform::transform(geoData,paramData,result);
    }
    virtual void compute        (const gsMapData<T> &geoData, gsFuncData<T> &result) const
    {
        gsFuncData<T> paramData (transform::getFuncFlags(result.flags));
        m_funValue.compute(geoData,paramData);
        transform::transform(geoData,paramData,result);
    }
};


typedef gsTransformedFuncSet<real_t, gsTransformPureGradConforming<real_t> > gsGradConformingTFS;
typedef gsTransformedFuncSet<real_t, gsTransformPureDivConforming<real_t> >  gsDivConformingTFS;
typedef gsTransformedFuncSet<real_t, gsTransformPureRestriction<real_t>  >   gsRestrictTFS;

} // namespace
