/** @file gsVectorValuedFunctionSet.h

    @brief Construct a vector valued function set out of the components

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan, A. Mantzaflaris
*/

#include <gsCore/gsFunctionSet.h>

namespace gismo {

// gsVBasis<int _Tr,T>
// components of targetdim= 1
// inherit from gsBasis for convenience
template <typename T>
class gsVectorValuedFunctionSet : public gsFuncSetComputeOnly<T>
{
protected:
    std::vector<gsFunctionSet<T>*> m_comp;// gsBasis<T>*
    std::vector<size_t>            m_map;
    index_t                        m_size; // sum(m_comp[i].size())
    int                            m_domDim;// m_comp.front().targetDim()
    int                            m_tarDim;// m_comp.size()

    void initSize()
    {
        m_size=0;
        for (int i=0; i<m_tarDim;++i)
            m_size+=m_comp[m_map[i]]->size();
    }
public:
    gsVectorValuedFunctionSet(gsFunctionSet<T>& func, int copies)
        : m_comp(1, &func),
          m_map(copies,0),
          m_domDim(func.domainDim()),
          m_tarDim(copies)
    {
        GISMO_ASSERT(func.info().targetDim==1, "Cannot create vector from vector valued functions");
        initSize();
    }

    virtual int domainDim() const {return m_domDim;}
    virtual int targetDim() const {return m_tarDim;}

    gsVectorValuedFunctionSet(std::vector<gsFunctionSet<T>*> basis)
        : m_map(basis.size()),
          m_tarDim(basis.size())
    {
        // build a map to avoid to recompute the same basis more then once
        typename std::vector<gsFunctionSet<T>*>::const_iterator beg=basis.begin();
        typename std::vector<gsFunctionSet<T>*>::const_iterator end=basis.end();
        typename std::vector<gsFunctionSet<T>*>::const_iterator cur=basis.begin();

        if (m_tarDim)
            m_domDim=basis[0]->domainDim();
        else GISMO_ERROR("vector functions must have at least a component");

        for (; cur!=end; ++cur)
        {
            typename  std::vector<gsFunctionSet<T>*>::const_iterator tmp=std::find(beg,cur,*cur);
            if (tmp!=cur)
                m_map[cur-beg]=tmp-beg;
            else
            {
                m_map[cur-beg]=m_comp.size();
                m_comp.push_back(*cur);
            }
        }
        initSize();
    }

public:
    virtual unsigned getGeoFlags (unsigned flags) const {return 0;}
    virtual index_t size () const {return m_size;}

    static unsigned maskFlags (unsigned flag)
    {
        const unsigned toZero=NEED_CURL|NEED_DIV;
        if (flag & toZero)
            flag |= NEED_DERIV;
        return flag & ~toZero;
    }

    virtual void compute        (const gsMatrix <T> &points, gsFuncData<T> &result) const
    {
        std::vector<gsFuncData<T> > computed(m_map.size());
        for (size_t c=0; c<computed.size(); ++c)
        {
            computed[c].flags=maskFlags(result.flags);
            computed[c].patchId=result.patchId;
            m_comp[c]->compute(points,computed[c]);
        }
        construct(computed,result);
    }

    virtual void compute        (const gsMapData<T> &geoData, gsFuncData<T> &result) const
    {
        std::vector<gsFuncData<T> > computed(m_map.size()); // mutable cache ?
        for (size_t c=0; c<computed.size(); ++c)
        {
            computed[c].flags=maskFlags(result.flags);
            computed[c].patchId=result.patchId;
            m_comp[c]->compute(geoData,computed[c]);
        }
        construct(computed,result);
    }

    void construct (const std::vector<gsFuncData<T> > &computed, gsFuncData<T> &result) const
    {
        initResult(computed,result);
        index_t written=0;
        for (int c=0; c<m_tarDim;++c)
        {
            written+=writeComponent(c,computed[m_map[c]],written, result);
        }
        computeDerived(result);
    }

    void initResult(const std::vector<gsFuncData<T> > &computed,gsFuncData<T> &result) const
    {
        const unsigned flags = result.flags;
        gsFuncInfo info(m_domDim,m_tarDim);
        gsFuncInfo infoComp(m_domDim,1);
        result.info = info;

        index_t activeNum=0;
        index_t activeC=computed[0].actives.cols();

        if (flags & NEED_ACTIVE )
            for (int c=0; c<m_tarDim;++c)
                activeNum+=computed[c].values[0].rows();
        else if (flags & NEED_VALUE )
            for (int c=0; c<m_tarDim;++c)
                activeNum+=computed[c].values[0].rows();
        else if (flags & NEED_DERIV )
            for (int c=0; c<m_tarDim;++c)
                activeNum+=computed[c].values[1].rows()/infoComp.derivSize();
        else if (flags & NEED_DERIV2 )
            for (int c=0; c<m_tarDim;++c)
                activeNum+=computed[c].values[2].rows()/infoComp.deriv2Size();

        if (flags & NEED_ACTIVE )
            result.actives.resize(activeNum,activeC);

        if (flags & NEED_DERIV2 )
            result.values.resize(3);
        else if ( flags & NEED_DERIV )
            result.values.resize(2);
        else if (flags & NEED_VALUE)
            result.values.resize(1);
        else result.values.resize(0);

        if (flags & NEED_VALUE)
            result.values[0].resize(activeNum*info.targetDim,computed[0].values[0].cols());
        if ( flags & NEED_DERIV )
            result.values[1].resize(activeNum*info.derivSize(),computed[0].values[1].cols());
        if ( flags & NEED_DERIV2 )
            result.values[2].resize(activeNum*info.deriv2Size(),computed[0].values[2].cols());

        // assume first derivative is set because is needed to compute the values
        if (flags & NEED_DIV)
            result.divs.resize(activeNum*info.divSize(),result.values[1].cols() );
        if (flags & NEED_CURL)
            result.curls.resize(activeNum*info.derivSize(),result.values[1].cols() );

        // does not assume anything as laplacian is always component wise
        if (flags & NEED_LAPLACIAN)
            result.laplacians.resize(activeNum*info.targetDim, computed[0].laplacians.cols());
    }

    index_t writeComponent (int comp, const gsFuncData<T>& data, index_t written, gsFuncData<T> &result) const
    {
        const unsigned flags = result.flags;
        index_t activeNum=0;
        gsFuncInfo info=gsFunctionSet<T>::info();

        if (flags & NEED_ACTIVE )
            activeNum=writeBlock(1,comp,written,data.actives,result.actives);
        if (flags & NEED_VALUE)
            activeNum=writeBlock(1,    comp,written,data.values[0],result.values[0]);
        if ( flags & NEED_DERIV )
            activeNum=writeBlock(info.domainDim,  comp,written,data.values[1],result.values[1]);
        if ( flags & NEED_DERIV2 )
            activeNum=writeBlock(info.deriv2Size()/info.targetDim, comp,written,data.values[2],result.values[2]);
        if (flags & NEED_LAPLACIAN)
            activeNum=writeBlock(1,    comp,written,data.laplacians,result.laplacians);
        return activeNum;
    }

    // needed for both unsigned and real_t
    template<typename TT>
    index_t writeBlock (index_t dataUnitSize, index_t comp, index_t startingPos, const gsMatrix<TT> &origin, gsMatrix<TT> &destination) const
    {
        gsFuncInfo info=gsFunctionSet<T>::info();
        index_t blocksNum = origin.rows()/dataUnitSize;
        index_t startRow  = startingPos*dataUnitSize*info.targetDim;

        index_t blockTempSize=info.targetDim*dataUnitSize;

        gsMatrix<T> blockTemp;
        blockTemp.setZero(blockTempSize, origin.cols());
        for (index_t b=0; b<blocksNum;++b)
        {
            blockTemp.middleRows(comp*dataUnitSize,dataUnitSize)=origin.middleRows(b*dataUnitSize,dataUnitSize);
            destination.middleRows(startRow+b*blockTempSize,blockTempSize)=blockTemp;
        }
        return blocksNum;
    }

    void computeDerived (gsFuncData<T> &result) const
    {
        gsFuncInfo info=gsFunctionSet<T>::info();
        const unsigned flags=result.flags;
        if ( flags & NEED_DIV )
            convertValue<T>::derivToDiv(result.values[1], result.divs, info);
        if ( flags & NEED_CURL )
            convertValue<T>::derivToCurl(result.values[1], result.curls, info);
    }
};


} // namespace gismo
