
#pragma once

#include <gsPde/gsPde.h>

namespace gismo
{


template <class T> class gsFunction;

/** @brief
    A perverted Stokes PDE.

    This class describes a perverted Stokes PDE, with an arbitrary right-hand side
    function and optionally a known solution.

    \ingroup Pde
    \ingroup pdeclass
 */

template<class T>
class gsPervertedStokesPde : public gsPde<T>
{
protected:
    gsPervertedStokesPde( ) { }
    using gsPde<T>::m_domain;
    using gsPde<T>::m_unknownDim;
    using gsPde<T>::m_solution;

public:
    gsPervertedStokesPde(
        const gsMultiPatch<T>          &domain,
        const gsBoundaryConditions<T>  &bc,
         gsFunction<T>       *force,
         gsFunction<T>       *velStateSol = NULL,
         gsFunction<T>       *preStateSol = NULL,
         gsFunction<T>       *velMultiSol = NULL,
         gsFunction<T>       *preMultiSol = NULL,
         gsFunction<T>       *source = NULL,
        const T                    viscosity = 1
        )
        :
            gsPde<T>(domain,bc),    m_viscosity(viscosity)

    {
        m_force  = force  ? force->clone()  : NULL;
        m_source = source ? source->clone() : NULL;
        m_solution.resize(4);
        m_solution[0] = velStateSol ? velStateSol->clone() : NULL;
        m_solution[1] = preStateSol ? preStateSol->clone() : NULL;
        m_solution[2] = velMultiSol ? velMultiSol->clone() : NULL;
        m_solution[3] = preMultiSol ? preMultiSol->clone() : NULL;

        m_unknownDim.push_back(m_domain.dim());
        m_unknownDim.push_back(1);
        m_unknownDim.push_back(m_domain.dim());
        m_unknownDim.push_back(1);
    }

    ~gsPervertedStokesPde( )
    {
        delete m_force;
        delete m_source;
    }

    const gsFunction<T>* rhs() const
    { return m_force; }
    const gsFunction<T>* force() const
    { return m_force; }
    const gsFunction<T>* source() const
    { return m_source; }

    const gsFunction<T>* velocityStateSolution() const
    { return m_solution[0]; }
    const gsFunction<T>* pressureStateSolution() const
    { return m_solution[1]; }
    const gsFunction<T>* velocityMultiplierSolution() const
    { return m_solution[2]; }
    const gsFunction<T>* pressureMultiplierSolution() const
    { return m_solution[3]; }
    T viscocity() const                 { return m_viscosity; }


    /// Prints the object as a string.
    std::ostream &print(std::ostream &os) const
    {
        os<<"Perverted Stokes's :\n"
          <<"ask Jarle Sogn!\n"
          <<"with:\n";
        if ( m_force )
        os<<"Force  function f= "<< *m_force <<".\n";
        if ( m_source )
        os<<"Source function g= "<< *m_source <<".\n";
        if ( m_solution[0] )
            os<<"Exact solution state u = "<< * m_solution[0] <<".\n";
        if ( m_solution[1] )
            os<<"Exact solution state p = "<< * m_solution[1] <<".\n";
        if ( m_solution[2] )
            os<<"Exact solution multiplier v = "<< * m_solution[2] <<".\n";
        if ( m_solution[3] )
            os<<"Exact solution multiplier q = "<< * m_solution[3] <<".\n";
        return os;
    }
    /// Consistency check
    bool check()
    {
        return true;
    }
protected:
    const gsFunction<T> * m_force;
    const gsFunction<T> * m_source;

    T m_viscosity;
}; // class gsPervertedStokesPde

} // namespace gismo
