#pragma once

#include <gsCore/gsLinearAlgebra.h>
#include <gsCore/gsGeometryEvaluator.h>
#include <gsCore/gsBasisEvaluator.h>

namespace gismo {


template <typename T>
class gsPointBasisEvaluator
{
private:
    const gsBasisEvaluator<T> * m_basis;
          unsigned              m_k;
public:

    gsPointBasisEvaluator() : m_basis(NULL), m_k(0) { }

    gsPointBasisEvaluator(const gsBasisEvaluator<T> &basis, unsigned point_idx)
        : m_basis(&basis), m_k(point_idx)
    {}
                   unsigned                  getParDim()  const {return m_basis->getParDim();}
                   unsigned                  getTarDim()  const {return m_basis->getTarDim();}
                   unsigned                  getSpaceDim()const {return m_basis->getSpaceDim();}

    const          gsMatrix<unsigned>       &actives ()   const {return m_basis->actives();}
    const typename gsMatrix<T>::constColumn  value ()     const {return m_basis->value(m_k);}
    const typename gsMatrix<T>::constColumn  deriv ()     const {return m_basis->deriv(m_k);}
    const typename gsMatrix<T>::constColumns jacobian ()  const {return m_basis->jacobian(m_k);}
    const typename gsMatrix<T>::constColumn  deriv2 ()    const {return m_basis->deriv2(m_k);}
    const typename gsMatrix<T>::constColumn  div  ()      const {return m_basis->div(m_k);}
    const typename gsMatrix<T>::constColumn  curl ()      const {return m_basis->curl(m_k);}
    const typename gsMatrix<T>::constColumn  laplacian () const {return m_basis->laplacian(m_k);}
};


template <typename T>
class gsPointGeometryEvaluator
{
private:
    const gsGeometryEvaluator<T> &m_geo;
    const unsigned                m_k;
public:
    gsPointGeometryEvaluator(const gsGeometryEvaluator<T> &geo, unsigned point_idx)
        : m_geo(geo), m_k(point_idx)
    {}
    const typename gsMatrix<T>::constColumn  value ()     const {return m_geo.value(m_k);}
    const typename gsMatrix<T>::constColumns jacobian ()  const {return m_geo.jacobian(m_k);}
    const typename gsMatrix<T>::constColumn  deriv2 ()    const {return m_geo.deriv2(m_k);}
    //               T                         div  ()      const {return m_geo.div(m_k);}
    //const typename gsMatrix<T>::constColumn  curl ()      const {return m_geo.(m_k);}
    //               T                         laplacian () const {return m_geo.laplacian(m_k);}
                   T                         measure ()   const {return m_geo.measure(m_k);}
    //const typename gsMatrix<T>::constColumn  normal ()    const {return m_geo.normal(m_k);}
    virtual void                             outerNormal (boxSide s, gsVector<T> & result) const
        { m_geo.outerNormal(m_k, s, result);}

    void transformLaplaceHgrad(const typename gsMatrix<T>::constColumn & grad,const typename gsMatrix<T>::constColumn & deriv2,gsMatrix<T> & output)const;
    real_t elementArea () const {return m_geo.measures().sum();}
};

template <typename T>
class gsOperator2
{
public:
    virtual ~gsOperator2()
    {}

    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const = 0;
    virtual unsigned    tBasisOptions() const  = 0;
    virtual unsigned    uBasisOptions() const  = 0;
    virtual unsigned    geoFlags()    const  = 0;
    virtual void        outputSize (unsigned &r, unsigned &c) const {}
};

//Works only for scalar function spaces and test and trail are assumed the same.
template <typename T>
class gsBiharmonicOp : public gsOperator2<T>
{
public:
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;

    virtual unsigned  tBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_2ND_DER;}
    virtual unsigned  uBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_2ND_DER;}
    virtual unsigned  geoFlags()    const {return  NEED_VALUE | NEED_MEASURE | NEED_GRAD_TRANSFORM | NEED_2ND_DER;}
};


//JS2: (\Delta u,  v) not verified
template<typename T>
class gsLaplaceOp : public gsOperator2<T>
{
public:
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;
    //BUG! I Need to pass flags here. This should not be necessary for the test space
    virtual unsigned  tBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_2ND_DER | NEED_LAPLACIAN;}
    virtual unsigned  uBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_2ND_DER | NEED_LAPLACIAN;}
    virtual unsigned  geoFlags()    const {return  NEED_VALUE | NEED_MEASURE | NEED_GRAD_TRANSFORM | NEED_2ND_DER;}
};


template <typename T>
class gsGradGradOp : public gsOperator2<T>
{
public:
    virtual gsMatrix<T>  pointEval (
                    const gsPointBasisEvaluator<T>     &tBasisEval,
                    const gsPointBasisEvaluator<T>     &uBasisEval,
                    const gsPointGeometryEvaluator<T>  &geoEval
                    ) const;
    virtual unsigned  tBasisOptions() const {return NEED_GRAD;}
    virtual unsigned  uBasisOptions() const {return NEED_GRAD;}
    virtual unsigned  geoFlags()    const {return NEED_GRAD_TRANSFORM|NEED_MEASURE;}
};



template <typename T>
class gsL2ScalarOp : public gsOperator2<T>
{
public:
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;

    virtual unsigned  tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned  uBasisOptions() const {return NEED_VALUE;}
    virtual unsigned  geoFlags()    const {return NEED_MEASURE;}
};


template <typename T>
class gsL2TestOp : public gsOperator2<T>
{
private:
    const gsFunction<T>    *testFunc;
public:
    gsL2TestOp(const gsFunction<T> &test)
        : testFunc(&test)
    {}
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;
    virtual unsigned  tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned  uBasisOptions() const {return 0;}
    virtual unsigned  geoFlags()    const   {return NEED_VALUE|NEED_MEASURE;}
    virtual void      outputSize (unsigned &r, unsigned &c) const
    { c=testFunc->targetDim(); }
};


template <typename T>
class gsL2TestVecOp : public gsOperator2<T>
{
private:
    const gsFunction<T>    *testFunc;
public:
    gsL2TestVecOp(const gsFunction<T> &test)
        : testFunc(&test)
    {}
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;
    virtual unsigned  tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned  uBasisOptions() const {return 0;}
    virtual unsigned  geoFlags()    const   {return NEED_VALUE|NEED_MEASURE;}
    virtual void      outputSize (unsigned &r, unsigned &c) const { c=1; }
};


template <typename T>
class gsDivergenceOp : public gsOperator2<T>
{
public:
    gsDivergenceOp() {}
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const
    { return tBasisEval.value()*uBasisEval.div().transpose()*geoEval.measure (); }
    virtual unsigned    tBasisOptions() const {return NEED_VALUE ;}
    virtual unsigned    uBasisOptions() const {return NEED_VALUE | NEED_DIV;}
    virtual unsigned    geoFlags()    const  {return NEED_MEASURE;}
};

//JS2: (grad(u), v) not tested
template <typename T>
class gsGradientOp : public gsOperator2<T>
{
public:
    gsGradientOp() {}
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_VALUE | NEED_GRAD;}
    virtual unsigned    uBasisOptions() const {return NEED_VALUE | NEED_GRAD;}
    virtual unsigned    geoFlags()    const  {return NEED_GRAD_TRANSFORM|NEED_MEASURE;}
};


// ///////////////////// boundary

template <typename T>
class gsBoundaryL2TestOp :
        public gsOperator2<T>
{
private:
            gsFunction<T>    *testFunc;
    mutable gsMatrix<T>       temp;
            boxSide    side;
public:
    gsBoundaryL2TestOp(gsFunction<T>    &f, boxSide a=boundary::none)
        : testFunc(&f), side(a)
    {}
    void setSide (boxSide a) { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned    uBasisOptions() const {return 0;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
    virtual void        outputSize (unsigned &r, unsigned &c) const
    { c=testFunc->targetDim(); }
};


template <typename T>
class gsBoundaryL2TestVecOp :
        public gsOperator2<T>
{
private:
            gsFunction<T>    *testFunc;
    mutable gsMatrix<T>       temp;
            boxSide    side;
public:
    gsBoundaryL2TestVecOp(gsFunction<T>    &f, boxSide a=boundary::none)
        : testFunc(&f), side(a)
    {}
    void setSide (boxSide a) { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned    uBasisOptions() const {return 0;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
    virtual void        outputSize (unsigned &r, unsigned &c) const { c=1; }
};

template <typename T>
class gsBoundaryL2ScalarOp :
        public gsOperator2<T>
{
private:
            boxSide    side;
public:
    gsBoundaryL2ScalarOp( boxSide a=boundary::none)
        : side(a)
    {}
    void setSide (boxSide a) { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_VALUE;}
    virtual unsigned    uBasisOptions() const {return NEED_VALUE;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
};

template <typename T>
class gsBoundaryNormalDerValueOp :
        public gsOperator2<T>
{
private:
            boxSide    side;
public:
    gsBoundaryNormalDerValueOp( boxSide a=boundary::none)
        : side(a)
    {}
    void setSide (boxSide a) { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_VALUE | NEED_JACOBIAN;}
    virtual unsigned    uBasisOptions() const {return NEED_VALUE;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
};

template <typename T>
class gsBoundaryNormalDerTestOp :
        public gsOperator2<T>
{
private:
            gsFunction<T>    *testFunc;
    mutable gsMatrix<T>       temp;
            boxSide    side;
public:
    gsBoundaryNormalDerTestOp(gsFunction<T>    &f, boxSide a=boundary::none)
        : testFunc(&f), side(a)
    {}
    void setSide (boxSide a)
    { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_JACOBIAN;}
    virtual unsigned    uBasisOptions() const {return 0;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
    virtual void        outputSize (unsigned &r, unsigned &c) const
    { c=testFunc->targetDim(); }
};


template <typename T>
class gsBoundaryNormalDerTestVecOp :
        public gsOperator2<T>
{
private:
            gsFunction<T>    *testFunc;
    mutable gsMatrix<T>       temp;
            boxSide    side;
public:
    gsBoundaryNormalDerTestVecOp(gsFunction<T>    &f, boxSide a=boundary::none)
        : testFunc(&f), side(a)
    {}
    void setSide (boxSide a) { side=a; }
    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval // not constant because transform gradients is not a constant function and trying to make it so result in problems compiling the library
                ) const;
    virtual unsigned    tBasisOptions() const {return NEED_JACOBIAN;}
    virtual unsigned    uBasisOptions() const {return 0;}
    virtual unsigned    geoFlags()    const  {return NEED_VALUE | NEED_OUTER_NORMAL;}
    virtual void        outputSize (unsigned &r, unsigned &c) const
    { c=1; }
};


template <typename T>
class gsLinElastOp : public gsOperator2<T>
{
public:

    gsLinElastOp( T YoungsModulus, T PoissonsRatio )
    {
        // maybe superfluous to keep all parameters,
        // but these two extra variables won't hurt.
        m_YoungsModulus=YoungsModulus;
        m_PoissonsRatio=PoissonsRatio;

        m_lambda = m_YoungsModulus * m_PoissonsRatio / ( (1.+m_PoissonsRatio)*(1.-2.*m_PoissonsRatio)) ;
        m_mu     = m_YoungsModulus / (2.*(1.+m_PoissonsRatio)) ;
    }

    virtual gsMatrix<T>  pointEval (
                const gsPointBasisEvaluator<T>     &tBasisEval,
                const gsPointBasisEvaluator<T>     &uBasisEval,
                const gsPointGeometryEvaluator<T>  &geoEval
                ) const;

    virtual unsigned  tBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_DIV;}
    virtual unsigned  uBasisOptions() const {return NEED_VALUE | NEED_GRAD | NEED_DIV;}
    virtual unsigned  geoFlags()    const {return  NEED_VALUE | NEED_MEASURE | NEED_GRAD_TRANSFORM;}

    T m_YoungsModulus;
    T m_PoissonsRatio;
    T m_lambda;
    T m_mu;
};


} // namespace gismo
