/** @file gsRecipe.h

    @brief The recipe assembler is set of tools to easily assemble
    system matrices associated to PDEs.

    The main tool is the function gsRecipeAssemble that takes as
    argument a gsRecipe, a list of gsBasisEvaluators and a numeric
    integration specified by a domain_iterator and a quadrature rule.

    A gsRecipe is a list of gsRecipeIngredient. Each of those specifies
    the construction of a matrix \em A whose coefficients are
       f$ a_{i,j} = \int_{\Omega} F(\psi_i, \phi_j) f$
    where f$ F f$ is an operator, the se of the f$\psi f$ is basis of the
    test space and that of the f$\phi f$ is a basis of the unknown space.

    The bases are specified as indexes in a list gsBasisEvaluators
    The PDE operator f$F f$ is modelled by a subclass of gsOperator2
    (see gsRecipeAssemler/gsPDEOperators.h).
    Writing to the destination matrix is done through a
    gsLocalToGlobalMapper (see gsSolver/gsLocalToGlobal.h) that also
    handles the transformation between the active function indexes
    and the matrix entry position.
    More gsRecipeIngredient can write to the same matrix.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan
    Created on: 2014-10-28
*/

#include <gsCore/gsConfig.h>

#pragma once
namespace gismo {

// forward declarations
template <typename T=real_t> class gsPointBasisEvaluator;
template <typename T=real_t> class gsPointGeometryEvaluator;
template <typename T=real_t> class gsRecipe;
template <typename T=real_t> class gsOperator2;
template <typename T=real_t> class gsL2TestOp;
template <typename T=real_t> class gsGradGradOp;
template <typename T=real_t> class gsL2ScalarOp;

}

#include <gsCore/gsGeometryEvaluator.h>
#include <gsCore/gsBasisEvaluator.h>
#include <gsAssembler/gsLocalToGlobal.h>
#include <gsRecipeAssembler/gsPDEOperators.h>
#include <vector>


namespace gismo {

/**
    \brief group together an operator, a pair of indexes of discrete
    spaces and a gsLocalToGlobalMapper that specify the destination matrix.

        The model is to construct a matrix \em A whose coefficients are
       f$ a_{i,j} = \int_{\Omega} F(\psi_i, \phi_j) f$
    where the f$\psi f$ are a basis of the test space and the f$\phi f$
    are a basis of the unknown space.


    gsRecipeIngredient assumes to \em own the operator and the storage rule
    it is provided and will thus free them automatically.
**/

template<typename T>
class gsRecipeIngredient
{
public:
    typedef memory::shared_ptr<gsOperator2<T>            >   opHandle  ;
    typedef memory::shared_ptr< gsLocalToGlobalMapper<T> >   ruleHandle;

public:

    const opHandle   & getOp  () const { return m_op  ; }
    const ruleHandle & getRule() const { return m_rule; }

    unsigned tSpaceId() const { return m_tSpace; }
    unsigned uSpaceId() const { return m_uSpace; }
    
    void operator= (gsRecipeIngredient other);

private:
    opHandle                    m_op;
    ruleHandle                  m_rule;
    unsigned                    m_tSpace;
    unsigned                    m_uSpace;
public:
    gsRecipeIngredient();

    /**
     * \brief setOperator changes the operator associated to
     *        this recipe entry.
     *
     * Since gsRecipeIngredient will own the given operator and
     * delete it automatically we recommend to construct the
     * proper operator on the spot like in:
     *
     * setOperator(new gsGradGradOp<>() );
     *
     * \param op a pointer to a gsOperator2 object
     */
    void setOperator (gsOperator2<T> *op);

    /**
     * \brief setOperator changes the operator associated to
     *        this recipe entry.
     *
     * \param op a shared_ptr pointer to a gsOperator2 object
     */
    void setOperator (const opHandle & op);

    /**
     * \brief setOperator changes the operator associated to
     *        this recipe entry.
     *
     * Since gsRecipeIngredient will own the given operator and
     * delete it automatically we recommend to construct the
     * proper operator on the spot like in:
     *
     * setRule(new gsL2GActives<>(sys_matrix) );
     *
     * \param rule a pointer to a gsLocalToGlobalMapper object
     */
    void setRule (gsLocalToGlobalMapper<T>   *rule);

    /**
     * \brief setOperator changes the operator associated to
     *        this recipe entry.
     *
     * \param rule a shared_ptr pointer to a gsLocalToGlobalMapper object
     */
    void setRule ( const ruleHandle &         rule);

    /**
     * \brief set the test space for this ingredient
     *        (the space varying with rows)
     * \param tSpace
     */
    void setTestSpace (unsigned tSpace);

    /**
     * \brief set the unknown space for this ingredient
     *        (the space varying with columns)
     * \param uSpace
     */
    void setUnknownSpace (unsigned uSpace);
};



/**
    \brief A recipe describes a way to assemble a system matrix and/or right-hand side

    Internally, it is a container for a list of gsRecipeIngredients
**/
template <typename T>
class gsRecipe
{
public:
    typedef typename gsRecipeIngredient<T>::opHandle   opHandle  ;
    typedef typename gsRecipeIngredient<T>::ruleHandle ruleHandle;

public:
    gsRecipe();
    gsRecipe(size_t size);

public:// to do: document

    gsRecipeIngredient<T>& operator[](size_t i);

    const gsRecipeIngredient<T>& operator[](size_t i) const;

    size_t size() const;

    void operator= (gsRecipe<T> other);

    void append (const gsRecipe & other);

    void add(const gsRecipeIngredient<T> & ingredient);

    void resize (size_t new_size);

    /**
        \brief assemble the matrices according to the recipe.

        \param quadRule is the quadrature rule used in numerical integration
        \param domain is a domain iterator that provides the integration
               elements: each of these is assumed to be fully contained
               in a Bezies element for the discrete spaces
        \param geometry is the parametrization of the domain, multipatch
               geometries must be handled manually outside of this function
        \param spaces is a list of gsBasisEvaluators that are used to
               evaluate the discrete spaces on the quadrature points

        This function asks each operator in the recipe to know what to
        compute on each node for each space and for the geometry.
        The gsBasisEvaluators provide a caching mechanism that allows
        to evaluate each space once per quadrature node.

        Note that this implementation can not compete with hand tuned
        assembler for specific equations as the interfaces can introduce
        some additional copying.
    **/
    void assemble( const gsQuadRule<T>               & quadRule,
                   gsDomainIterator<T>               & domain  ,
                   const gsGeometry<T>               & geometry,
                   std::vector<gsBasisEvaluator<T>*>   spaces  ) const;

private:
    std::vector< gsRecipeIngredient<T> > m_data;
};

} // namespace gismo

#ifndef GISMO_BUILD_LIB
#include <gsRecipeAssembler/gsRecipe.hpp>
#endif
