/** @file gsRemappedBasis.h

    @brief abstract implementation of a basis that delegates evaluation
    to other basis locally

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan
**/

#pragma once

#include <gsNurbs/gsTensorBSplineBasis.h>
#include <gsCore/gsFuncData.h>
#include <gsCore/gsFunctionSet.h>
#include <gsMapUtils/gsMapper.h>
#include <gsRemappedBasis/gsSelector.h>

namespace gismo {

template <typename baseBasis>
class myWrappedBasis : public gsFunctionSet<real_t>, public baseBasis
{
public:
    typedef typename baseBasis::domainIter domainIter;

public:
    myWrappedBasis(const baseBasis &basis)
        : baseBasis(basis)
    {}

    virtual ~myWrappedBasis() {}

    virtual int domainDim () const{ return baseBasis::dim(); }

    virtual int targetDim () const{ return 1; }

    virtual void active_into    (const gsMatrix<real_t>  &points, gsMatrix<unsigned> &result) const
    {
        baseBasis::active_into(points,result);
    }
    virtual void eval_into      (const gsMatrix<real_t>  &points, gsMatrix<real_t> &result) const
    {
        baseBasis::eval_into(points,result);
    }
    virtual void deriv_into     (const gsMatrix<real_t>  &points, gsMatrix<real_t> &result) const
    {
        baseBasis::deriv_into(points,result);
    }
    virtual void deriv2_into    (const gsMatrix<real_t>  &points, gsMatrix<real_t> &result) const
    {
        baseBasis::deriv2_into(points,result);
    }
    index_t size() const
    {
        return baseBasis::size();
    }
    unsigned getGeoFlags(unsigned flags) const
    {
        return 0;
    }
    void compute(const gsMapData<real_t> &geometry, gsFuncData<real_t> & result) const
    {
        gsFunctionSet<real_t>::compute(geometry.points, result);
    }
};



class GISMO_EXPORT gsRemappedBasis : public gsFuncSetComputeOnly<real_t>
{
protected:
    gsMapper   m_repr; // matrix representing global functions as linear combinations
    gsSelector m_sele; // choose which basis to use based on point
    std::vector<gsFunctionSet<real_t>*> m_basis;
    std::vector<unsigned>               m_shift; // {Bi} set of local basis :  0, dim B1, dim B1+dim B2 ..., sum_i dim Bi

    gsFuncInfo m_info; //

    // Empty constructor is protected for use by derived classes.
    gsRemappedBasis()
    {}
public:
    // Constructor providing everything is public.
    gsRemappedBasis( const gsMapper   &repr,
                     const gsSelector &sele,
                     const std::vector<gsFunctionSet<real_t>* > &basis ):
        m_repr ( repr ),
        m_sele ( sele ),
        m_basis( basis )
    {
        GISMO_ASSERT (m_basis.size(), "cannot construct a gsRemappedBasis without local basis\n");
        m_info.domainDim=m_basis[0]->info().domainDim;
        m_info.targetDim=m_basis[0]->info().targetDim;

        for (size_t b=0; b<m_basis.size();++b)
            if ( !(m_basis[b]->info() == m_info) )
                GISMO_ERROR("all basis must have same domain and target dimension");
    }

    virtual index_t size() const
    {
        return m_repr.getNrOfTargets();
    }

    virtual void compute(const gsMatrix<real_t>  &points,
                               gsFuncData<real_t>  & result) const;

    virtual void compute(const gsMapData<real_t> &geometry, gsFuncData<real_t> & result) const
    {
        compute(geometry.points,result);
    }
private:
    template <typename T>
    void copyAndComplete (gsMatrix<T> gsFuncData<real_t>::* inPtr, const std::vector<gsFuncData<real_t> >&data, gsFuncData<real_t> &result) const;
    void copyAndComplete (int index, const std::vector<gsFuncData<real_t> >&data, gsFuncData<real_t> &result) const;

    template <typename inType, typename outType> // can be matrix or column
    void remapValues (
            const gsMatrix<real_t>  &matrix,
            const inType            &input,
            outType                 &output
            ) const;

    template <typename actType> // can be a matrix or a column
    void getMapperSubmatrix(
            const gsMatrix<unsigned>   &rowsId, // local actives
            const actType              &colsId, // global actives
            gsMatrix<real_t>           &submatrix
            ) const;

    void getActivesFromRows(
            const gsMatrix<unsigned>    &rowsId, // input Rows to search
                  std::vector<unsigned> &colsId  // columns with non zero coefficients in at least one of the rows in rowsId
            ) const;

    void getActivesFromRows(
            const gsMatrix<unsigned>  &rowsId, // input Rows to search
                  gsMatrix<unsigned>  &colsId  // columns with non zero coefficients in at least one of the rows in rowsId
            ) const;

    // should be moved to utils
    static void mergeSortUnique (std::vector<std::vector<unsigned> >::iterator inputBeg,
                                 std::vector<std::vector<unsigned> >::iterator inputEnd,
                                 std::vector<unsigned> &output);

    static void mergeSortUnique(
            std::vector<unsigned> &v1,
            std::vector<unsigned> &v2,
            std::vector<unsigned> output
            );
};


}
