/** @file gsRemappedTHB.h

    @brief New implementation of THB-splines using gsRemappedBasis.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan, D. Mokris

    The authors would like to thank F. Buchegger for his help ;).
**/

#pragma once

#include <vector>
#include <gsNurbs/gsKnotVector.h>
#include <gsRemappedBasis/gsSpecializedRemappedBasis.h>
#include <gsUtils/gsMultiIndexIterators.h>

namespace gismo
{

class GISMO_EXPORT gsRemappedTHB : public gsSpecializedRemappedBasis
{
private:
    /** Constructs a sparse matrix with nRows rows, nCols cols, value value on
     * the diagonal and zeros elsewhere and returns it as the result. */
    void diagMatrix(int nRows,
                    int nCols,
                    real_t value,
                    gsSparseMatrix<real_t, Eigen::RowMajor> &result);

    /** Returns the knots that are in the basis number startLvl
     * but not in the basis number endLvl
     * and in the direction dir. */
    std::vector< real_t > knotDiff1D( gsKnotVector<real_t> smallerKV,
                                      gsKnotVector<real_t> biggerKV );

    /** Returns the knots (coordinate-wise)
     * that are in the basis number endLvl but not in
     * the basis number startLvl. */
    template <int dim>
    std::vector< std::vector< real_t > > knotDiffHelper( unsigned startLvl,
                                                         unsigned endLvl ) // maybe size_t? // even better we can typedef it somewhere
    {
        gsTensorBSplineBasis<dim,real_t> *b1
                = static_cast<gsTensorBSplineBasis<dim,real_t>* >(
                    dynamic_cast<myWrappedBasis<gsTensorBSplineBasis<dim,real_t> >*>(
                        m_basis[startLvl]) );

        gsTensorBSplineBasis<dim,real_t> *b2
                = static_cast<gsTensorBSplineBasis<dim,real_t>* >(
                    dynamic_cast<myWrappedBasis<gsTensorBSplineBasis<dim,real_t> >*>(
                        m_basis[endLvl]) );

        std::vector< std::vector< real_t > > result;

        for( size_t i = 0; i < dim; ++i )
            result.push_back( knotDiff1D( b1->knots( static_cast< int >( i )),
                                          b2->knots( static_cast< int >( i ))));

        return result;
    }

    /** Switch between versions of knotDiffDim for different dimensions. */
    std::vector< std::vector< real_t > > knotDiff( unsigned startLvl,
                                                   unsigned endLvl );
    template <int dim>
    gsSparseMatrix<real_t, RowMajor> getTransitionHelper( unsigned startLvl,
                                                          unsigned endLvl )
    {
        gsSparseMatrix<real_t, RowMajor> result;
        if( startLvl > endLvl )
        {
            diagMatrix( (m_basis[startLvl])->size(),
                        (m_basis[endLvl])->size(), 0.0, result );
        }
        else if ( startLvl == endLvl )
        {
            diagMatrix( m_basis[startLvl]->size(), m_basis[endLvl]->size(), 1.0, result );
        }
        else
        {
            gsTensorBSplineBasis<dim,real_t> *clonedBasis
                    = static_cast<gsTensorBSplineBasis<dim,real_t>* >(
                        dynamic_cast<myWrappedBasis<gsTensorBSplineBasis<dim,real_t> >*>
                        ( m_basis[startLvl]->clone() ));
            std::vector< std::vector<real_t> > knotDifference = knotDiff( startLvl, endLvl );
            clonedBasis->refine_withTransfer( result, knotDifference );
            delete( clonedBasis );
            // Lesson for life: forgotten ">" gives the "template argument 1 invalid" error.
        }
        return result;
    }
public:
    gsRemappedTHB() : gsSpecializedRemappedBasis(){}
    ~gsRemappedTHB(){}

    virtual int domainDim() const {return m_info.domainDim;}
    virtual int targetDim() const {return m_info.targetDim;}

protected:
    void computeSelectedAndInfluencing()
    {
        switch (m_info.domainDim)
        {
            case 1:
                return computeSelectedAndInfluencingDim<1> ();
            case 2:
                return computeSelectedAndInfluencingDim<2> ();
            case 3:
                return computeSelectedAndInfluencingDim<3> ();
            case 4:
                return computeSelectedAndInfluencingDim<4> ();
        default:
            GISMO_ERROR("Not implemented for domains that are 5 or more dimensional or less then 1");
        }
    }


    virtual std::vector<index_t> getSelected   ( unsigned lvl );
    virtual std::vector<index_t> getInfluencing( unsigned lvl );

    virtual gsSparseMatrix<real_t, RowMajor> getTransition( unsigned startLvl,
                                                            unsigned endLvl)
    {
        switch( this->info().domainDim )
        {
        case 1:
            return getTransitionHelper<1> ( startLvl,endLvl );
        case 2:
            return getTransitionHelper<2> ( startLvl,endLvl );
        case 3:
            return getTransitionHelper<3> ( startLvl,endLvl );
        case 4:
            return getTransitionHelper<4> ( startLvl,endLvl );
        default:
            gsWarn << "Sorry, getTransition has not been implemented for the dimension you request.";
            gsSparseMatrix<real_t, RowMajor> result;
            return result;
        }

    }

    template<int dim>
    void computeSelectedAndInfluencingDim()
    {
        std::vector<std::vector<unsigned> > influencing(m_basis.size());
        std::vector<std::vector<unsigned> > selected(m_basis.size());
        std::vector<std::vector<unsigned> > selectedCandidates    (m_basis.size());
        std::vector<std::vector<unsigned> > selectedCandidatesMult(m_basis.size());

        // find influencing, sure selected and selected candidates
        gsBoxList bList = m_sele.domainMap.asBoxList(m_boundingBox);
        for (size_t b=0; b< bList.size(); ++b)
        {
            gsVector<index_t,dim> min,max, deg;
            gsTensorGridIterator<index_t,dim> *it = getInfluencingForBox<dim>(bList.box(b),bList.basisId(b));
            min=it->multiIndex();
            it->last();
            max=it->multiIndex();
            for (int c=0; c<dim; ++c)
                deg(c)=dynamic_cast<gsTensorBSplineBasis<dim,real_t>*>(m_basis[bList.basisId(b)])->degree(c);
            for (it->first();it->good();it->next())
            {
                addUnique( it->flatIndex(), influencing[bList.basisId(b)]);
                if ( wellContained( it->multiIndex(), min,max,deg))
                    addUnique ( it->flatIndex(), selected[bList.basisId(b)]);
                else
                {
                    size_t pos=addUnique( it->flatIndex(), selectedCandidates[bList.basisId(b)]);
                    if (pos< selectedCandidatesMult[bList.basisId(b)].size())
                        ++selectedCandidatesMult[bList.basisId(b)].at(pos);
                    else
                        selectedCandidatesMult[bList.basisId(b)].insert(selectedCandidatesMult[bList.basisId(b)].begin()+pos, 1);
                }
            delete it;
            }
        }

        // choose from selectedCandidates

        // write code
    }


private:
    template<int dim>
    bool wellContained(
        const gsVector<index_t,dim> &pos,
        const gsVector<index_t,dim> &min,
        const gsVector<index_t,dim> &max,
        const gsVector<index_t,dim> &dst
         )
    {
        return ( pos.array()>=min.array()+dst.array() && pos.array()<max.array()-dst.array()).all();
    }
    // returns the position if already present
    static inline size_t addUnique(unsigned val, std::vector<unsigned> &vec)
    {
        std::vector<unsigned>::iterator it = std::lower_bound(vec.begin(),vec.end(),val);
        if (*it==val)
            return it-vec.begin();
        vec.insert(it,val);
        return vec.size();
    }
protected:
    template<int dim>
    gsTensorGridIterator<index_t, dim>* getInfluencingForBox ( const gsMatrix<real_t> &box, unsigned lvl )
    {
        gsTensorBSplineBasis<dim,real_t,gsKnotVector<real_t> > *tensorBasis
                = dynamic_cast<gsTensorBSplineBasis<dim,real_t>*>(m_basis[lvl]);

        gsVector<index_t, dim> maxIter;
        gsVector<index_t, dim> m, M; // As in the makeSubGridIterator below.
        for( unsigned i = 0; i != static_cast<unsigned>(dim); ++i )
        {
            gsKnotVector<real_t> knotVector = tensorBasis->knots(i);
            maxIter[i] = knotVector.size() - tensorBasis->degree(i) - 1;

            // This is a bug. We probably need to use findIndex instead, as this handles the multiplicities not the way we want them here.
            m(i) = knotVector.findspan(box(i,0)) - tensorBasis->degree(i);
            M(i) = knotVector.findspan(box(i,1));// we need +1 if we are in the middle and no change if we are aligned with the boxes (the maximum is not part of the iteration set)
        }
        gsTensorGridIterator<index_t,dim> it(maxIter);
        return it.makeSubGridIterator( m, M );
    }

};

// Terminology:
// Active function: see Basis.h
// influencing of basis(basisId): functions from basis(basisId) which are used for evaluation.
// I.e., such that their corresponding line in the representation matrix has at least one non-zero entry.
// selected of basis(basisId): selected by Kraft's procedure (only for hierarchial remapped).

} // namespace gismo
