/** @file gsSelector

    @brief choose based on coordinates which lower level basis to call

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): A. Bressan
**/

#pragma once

#include <gsCore/gsExport.h>
#include <gsCore/gsLinearAlgebra.h>
#include <gsRemappedBasis/tree.h>
#include <gsCore/gsFunctionSet.h>

namespace gismo
{


struct ForkNodeData
{
    int     dir;
    real_t  par;
};

union NodeData
{
    ForkNodeData split;
    size_t       space;
};


struct domainMapDefaultPolicy
{
    size_t solve (const std::vector<size_t> &set)
    {
        if (set.size()==1)
            return set.back();
        else
            GISMO_ERROR("badly specified domain partition: boxes overlap");
    }
};

struct domainMapMaxPolicy
{
    size_t solve (const std::vector<size_t> &set)
    {
        if (set.size()==0)
            return 0;
        else
            return *std::max_element(set.begin(),set.end());
    }
};


struct domainMapMinPolicy
{
    size_t solve (const std::vector<size_t> &set)
    {
        if (set.size()==0)
            return 0;
        else
            return *std::min_element(set.begin(),set.end());
    }
};



struct GISMO_EXPORT gsBoxList
{
protected:
    std::vector<size_t> m_basisId;
    std::vector<real_t> m_boxes;
public:
    int                     dim;
    gsAsMatrix<real_t>      box     (size_t p)       {return gsAsMatrix<>(&m_boxes[dim*2*p],dim,2);}
    gsAsConstMatrix<real_t> box     (size_t p) const {return gsAsConstMatrix<>(&m_boxes[dim*2*p],dim,2);}
    size_t                  basisId (size_t p) const {return m_basisId[p];}
    size_t&                 basisId (size_t p)       {return m_basisId[p];}

    void                    append(const gsMatrix<real_t> &box, size_t basisId )
    {
        GISMO_ASSERT(box.rows()==dim, "dimension mismatch");
        m_basisId.push_back(basisId);
        m_boxes.resize(dim*2+m_boxes.size());
        gsAsMatrix<>(&m_boxes[(size()-1)*2*dim],dim,2)=box;
    }

    size_t             size () const {return m_basisId.size();}

    /**
     * @brief Returns the basisId of the basis capable of depicting all the basis functions
     * (from various basisId) active at the @a point.
     * @param point is a column of coordinates.
     * @param resolver is a strategy to decide in case of overlapping boxes
     * (it depicts the partial ordering of the basisId's).
     * @return basisId associated to @a point.
    */
    template <typename policy>
    size_t getBasisAt(const gsMatrix<real_t> point, policy resolver) const
    {
        std::vector<size_t> overlap;
        for (size_t b=0; b<size();++b)
        {
            if (contains(box(b),point))
                overlap.push_back(basisId(b));
        }
        return resolver.solve(overlap);
    }

    /**
     * @brief Checks whether @a m_boxes is a valid list.
     * @return always true; the check is through GISMO_ASSERT.
     */
    bool check() const
    {
         GISMO_ASSERT(2*dim*size()==m_boxes.size(),"corrupted boxes list");
         return true;
    }

private:
    /**
     * @brief Checks whether @a point is inside @a box.
     * Note that the boxes are closed from left and open from right.
     * @param box is given as a matrix with d rows and 1 column
     * @param point is also given this way
     * @return true iff @a point is inside @a box
     */
    static inline bool contains ( const gsMatrix<real_t> &box,
                                  const gsMatrix<real_t> &point)
    {
        for (index_t r=0; r<box.rows(); ++r)
            if (box(r,0)>point(r,0) || box(r,1)<=point(r,0))
                return false;
        return true;
    }
};

GISMO_EXPORT std::ostream& operator<<  (std::ostream &out, const gsBoxList &boxes);


class gsDomainMap : public BinaryTree<NodeData>
{
public:
    gsDomainMap()
    {}

    /**
     * Constructs the underlying binary tree from a list of boxes and a bounding box.
     * Template parameter @a policy tells what to do in case of a tie (i.e.,
     * describes the partial ordering of the basisId's.
     * @param boxes list of boxes
     * @param boundingBox bounding box
     */
    template <typename policy>
    void initFromBoxes( const gsBoxList &boxes,
                        const gsMatrix<> &boundingBox)
    {
        policy resolver;
        initFromBoxes(boxes,boundingBox,resolver);
    }

    /**
     * Constructs the underlying binary tree from a list of boxes and a bounding box.
     * @param boxes list of boxes
     * @param boundingBox bounding box
     * @param resolver tells what to do in case of a tie (i.e.,
     * describes the partial ordering of the basisId's.
     */
    template <typename policy>
    void initFromBoxes( const gsBoxList &boxes,
                        const gsMatrix<> &boundingBox,
                        policy resolver)
    {
        GISMO_ASSERT(boundingBox.rows()==boxes.dim,"dimension mismatch");
        boxes.check();
        generateSplitsFromBoxes(getRoot(),boxes,boundingBox,resolver);
    }

    /**
     * @brief Returns the bounding box of the boxes in the subtree starting in
     * @param node
     */
    gsMatrix<real_t> getBoundingBox(NodeId node, const gsMatrix<real_t> &domain)
    {
        gsMatrix<real_t> result;
        result.setConstant(domain.rows(),2,NAN);
        node_iterator_NFS pre(*this, node);
        node_iterator_NFS cur(*this, pre->parent);

        while(cur!=node_iterator_NFS(*this,no_node))
        {
            GISMO_ASSERT(cur->data.split.dir<domain.rows(), "dimension mismatch");
            if(cur->first==pre)
            {
                if (math::isnan(result(cur->data.split.dir,1 )))
                    result( cur->data.split.dir,1 )=cur->data.split.par;
                else
                    result( cur->data.split.dir,1 )=math::min(result(cur->data.split.dir,1 ),cur->data.split.par);
            }
            else // (cur->second==pre)
            {
                if (math::isnan(result(cur->data.split.dir,0 )))
                    result(cur->data.split.dir,0 )=cur->data.split.par;
                else
                    result(cur->data.split.dir,0 )=math::max(result(cur->data.split.dir,0 ),cur->data.split.par);
            }
            pre=cur;
            cur=node_iterator_NFS(*this,cur->parent);
        }

        return result;
    }

    /**
     * @brief Returns the domains in the tree restricted to @a boundingBox
     * as a list of boxes.
     */
    gsBoxList asBoxList(const gsMatrix<> &boundingBox)
    {
        gsBoxList result;
        result.dim=boundingBox.rows();

        leaf_iterator_FS it=begin<leaf_FS>();

        for ( ;it!=end<leaf_FS>();++it)
        {
            gsMatrix<real_t> box=getBoundingBox(it,boundingBox);
            for (int c=0; c<result.dim;++c)
            {
                if (math::isnan(box(c,0)) ) box(c,0)=boundingBox(c,0);
                if (math::isnan(box(c,1)) ) box(c,1)=boundingBox(c,1);
            }
            if (notEmpty(box))
                result.append(box,it->data.space);
        }
        return result;
    }
protected:

    struct EstimateData
    {
        real_t split;
        size_t before;
        size_t after;
        EstimateData()
            : split(NAN), before(0), after(0)
        {}
        EstimateData(real_t pos)
            : split(pos), before(0), after(0)
        {}

        bool operator< (const real_t pos) const {return split<pos;}
    };

    class EstimateComp : public std::vector<EstimateData>
    {
    public:
        EstimateComp() {}

        using std::vector<EstimateData>::insert;
        void addBreak (real_t pos)
        {
            iterator p = std::lower_bound(begin(),end(), pos);
            if ( p==end() || !(p->split==pos) )
                insert(p,EstimateData(pos));
        }
    };

    void estimateSplitFromBoxes (const gsBoxList &boxes, const gsMatrix<> &boundingBox, std::vector<EstimateComp> &estimate)
    {
        const int dim = boundingBox.rows();
        estimate.clear();
        estimate.resize(dim);
        // first loop fill the positions
        for (size_t b=0; b<boxes.size();++b)
        {
            if (!intersect(boxes.box(b), boundingBox))
                continue;
            for (int c=0; c<dim;++c)
            {
                if (boxes.box(b)(c,0)>boundingBox(c,0))
                    estimate[c].addBreak(boxes.box(b)(c,0));
                if (boxes.box(b)(c,1)<boundingBox(c,1))
                    estimate[c].addBreak(boxes.box(b)(c,1));
            }
        }
        // second loop to estimate
        for (size_t b=0; b<boxes.size();++b)
        {
            if (!intersect(boxes.box(b), boundingBox))
                continue;
            for (int c=0; c<dim;++c)
            {
                for (size_t s=0; s<estimate[c].size();++s)
                {
                    if (estimate[c][s].split>boxes.box(b)(c,0))
                        ++estimate[c][s].before;
                    if (estimate[c][s].split<boxes.box(b)(c,1))
                        ++estimate[c][s].after;
                }
            }
        }
    }

    template <typename policy>
    void generateSplitsFromBoxes (NodeId node, const gsBoxList &boxes, const gsMatrix<> &boundingBox)
    {
        policy pp;
        generateSplitsFromBoxes (node,  boxes,  boundingBox, pp);
    }

    template <typename policy>
    void generateSplitsFromBoxes (NodeId node, const gsBoxList &boxes, const gsMatrix<> &boundingBox,policy resolver)
    {
        std::vector<EstimateComp> splitEstimate;
        estimateSplitFromBoxes(boxes,boundingBox,splitEstimate);
        int    dir = -1;
        real_t par = NAN;
        chooseBestSplit(splitEstimate,dir,par);

        if(dir != -1)
        {
            (*this)[node].data.split.dir=dir;
            (*this)[node].data.split.par=par;

            NodeId first  = this->append(node,true);
            gsMatrix<real_t> bBox=boundingBox;
            bBox(dir,1)=par;
            generateSplitsFromBoxes(first,boxes,bBox, resolver);

            NodeId second  = this->append(node,false);
            bBox(dir,1)=boundingBox(dir,1);
            bBox(dir,0)=par;
            generateSplitsFromBoxes(second,boxes,bBox, resolver);
        }
        else
            (*this)[node].data.space=boxes.getBasisAt((boundingBox.col(0)+boundingBox.col(1))/2, resolver);
    }

    void chooseBestSplit (const std::vector<EstimateComp> &splitEstimate, int &dir, real_t &par)
    {
        const int dim=splitEstimate.size();
        bool mustSplit=false;
        for (int c=0;c<dim;++c)
            if (splitEstimate[c].size())
            {
                mustSplit=true;
                break;
            }
        if (!mustSplit)
        {
            dir=-1;
            return;
        }
        // choose the split closest to 50 / 50 to minimize the tree depth
        real_t maxDist=1;
        for (int c=0; c<dim;++c)
        {
            real_t lastCurDist=1;
            for (size_t can=0; can<splitEstimate[c].size();++can)
            {
                real_t curDist=1;
                if (splitEstimate[c][can].after>splitEstimate[c][can].before)
                    curDist=1-real_t(splitEstimate[c][can].before)/real_t(splitEstimate[c][can].after);
                else if (splitEstimate[c][can].after<splitEstimate[c][can].before)
                    curDist=1-real_t(splitEstimate[c][can].after)/real_t(splitEstimate[c][can].before);
                else if (splitEstimate[c][can].after==splitEstimate[c][can].before)
                    curDist=0;
                if(curDist > lastCurDist)
                    break;
                lastCurDist=curDist;
                if(curDist <= maxDist)
                {
                    dir=c;
                    par=splitEstimate[c][can].split;
                    maxDist=curDist;
                }
            }
        }
    }



private:
    static inline bool notEmpty (const gsMatrix<real_t> &box)
    {
        for (index_t r=0; r<box.rows(); ++r)
            if (box(r,0)>=box(r,1))
                return false;
        return true;
    }
    static inline bool intersect (const gsMatrix<real_t> &box1, const gsMatrix<real_t> &box2)
    {
        for (index_t r=0; r<box1.rows(); ++r)
            if (box1(r,0)>=box2(r,1) || box2(r,0)>=box1(r,1))
                return false;
        return true;
    }

};




class gsSelector
{
public:
    gsDomainMap domainMap;
    typedef gsDomainMap::NodeId NodeId;
private:
    /**
     * @brief Makes a restriction of the gsSelector to @a box
     * (cutting the unnecessary outer branches away).
     * @param box to which to restrict;
     * @param curNode current node;
     * @param result where the returned new gsSelector is written.
     */
    void restrictNode( const std::vector<real_t> &box,
                       const NodeId &curNode,
                       gsSelector &result )
    {
        if( result.domainMap[curNode].isLeaf() )
            return; // End of recursion.

        const NodeData &data = domainMap[curNode];
        if( data.split.par <= box[2 * (data.split.dir)])
        {
            NodeId second = result.domainMap[curNode].second;
            result.domainMap.cutBranchAway( curNode, true );
            restrictNode( box, second, result );
            return;
        }
        else if( data.split.par >= box[2 * (data.split.dir) + 1])
        {
            NodeId first = result.domainMap[curNode].first;
            result.domainMap.cutBranchAway( curNode, false );
            restrictNode( box, first, result );
            return;
        }
        else
        {
            restrictNode( box, result.domainMap[curNode].first, result );
            restrictNode( box, result.domainMap[curNode].second, result );
            return;
        }
    }
public:

    /**
     * Returns the basisId of the basis capable of depicting all the basis functions
     * active in the point @a p.
     */
    template <typename PointType>
    size_t getBasisAt (const PointType &p) const
    {
        NodeId curNode=domainMap.getRoot();
        while (domainMap[curNode].isFork())
        {
            const NodeData &data=domainMap[curNode];
            if( data.split.par>p(data.split.dir,0) )
                curNode=domainMap[curNode].first;
            else
                curNode=domainMap[curNode].second;
        }
        const NodeData &data=domainMap[curNode];
        return data.space;
    }

    /**
     * Returns the basisId of the basis capable of depicting all the basis functions
     * active in the given points.
     * @param points , each given as a column of the matrix;
     * @param out returned basisId's (one per point).
     */
    void getBasisAt (const gsMatrix<real_t> &points, std::vector<size_t> &out) const
    {
        out.resize(points.cols());
        for (index_t p=0; p<points.cols();++p)
            out[p]=getBasisAt(points.col(p));
    }

    /**
     * @brief Makes a new gsSelector by restricting (non-destructively) the existing
     * one to @a box and cutting the unnecessary outer branches away.
     * @param box (bounding);
     * @param result new gsSelector returned.
     */
    void restrictTree( const std::vector<real_t> &box, gsSelector &result )
    {
        // The box is: begin in dir 0, end in dir 0, begin in dir 1, end in dir 1 and so on.
        result = *this;
        restrictNode(box, result.domainMap.root(), result );
        //std::cout << result.domainMap << std::endl;
    }
};

GISMO_EXPORT  std::ostream& operator<<  (std::ostream &out, const NodeData &node);





} // namespace gismo
