/** @file gsCompositeAssemblerUtils.h

    @brief Provides utils for assemblers working with gsCompositeBasis or Geometry.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): F. Buchegger
*/

#pragma once

#include <gsSmoothPatches/gsCompositeBasis.h>
#include <gsSmoothPatches/gsCompositeHBasis.h>
#include <gsSmoothPatches/gsCompositeGeom.h>
#include <gsSmoothPatches/gsCompositeSingleBasis.h>

#include <gsRecipeAssembler/gsRecipeAssemblerQMOpt.h>
#include <gsRecipeAssembler/gsRecipeAssemblerFitting.h>
#include <gsRecipeAssembler/gsRecipeAssemblerPoisson.h>
#include <gsRecipeAssembler/gsRecipeAssemblerBiharmonic.h>

#include <gsAssembler/gsNormL2.h>
#include <gsAssembler/gsSeminormH1.h>
#include <gsAssembler/gsSeminormH2.h>

#include <gsCore/gsFieldCreator.h>
#include <gsIO/gsWriteParaview.h>
#include <gsUtils/gsStopwatch.h>

namespace gismo
{

//============================================== NORMS ===============================================//

real_t compositeL2Error(const gsField<real_t>& computedSol,const gsFunction<real_t>& exactSol,bool exactSolIsParam=false)
{
    gsNormL2<real_t> l2DistNorm(computedSol,exactSol,exactSolIsParam);
    l2DistNorm.compute();
    gsNormL2<real_t> l2Norm(computedSol);
    l2Norm.compute();
    return l2DistNorm.value()/l2Norm.value();
}

real_t compositeH1Error(const gsField<real_t>& computedSol,const gsFunction<real_t>& exactSol,const gsFunction<real_t>& dExactSol,bool exactSolIsParam=false)
{
    gsSeminormH1<real_t> h1DistNorm(computedSol,exactSol,dExactSol,exactSolIsParam);
    h1DistNorm.compute();
    gsSeminormH1<real_t> h1Norm(computedSol);
    h1Norm.compute();
    gsNormL2<real_t> l2DistNorm(computedSol,exactSol,exactSolIsParam);
    l2DistNorm.compute();
    gsNormL2<real_t> l2Norm(computedSol);
    l2Norm.compute();
    return sqrt(l2DistNorm.value()*l2DistNorm.value() + h1DistNorm.value()*h1DistNorm.value())/
            sqrt(l2Norm.value()*l2Norm.value() + h1Norm.value()*h1Norm.value());
}

real_t compositeH2Error(const gsField<real_t>& computedSol,const gsFunction<real_t>& exactSol,const gsFunction<real_t>& dExactSol,const gsFunction<real_t>& ddExactSol,bool exactSolIsParam=false)
{
    gsSeminormH2<real_t> h2DistNorm(computedSol,exactSol,ddExactSol,exactSolIsParam);
    h2DistNorm.compute();
    gsSeminormH2<real_t> h2Norm(computedSol);
    h2Norm.compute();
    gsSeminormH1<real_t> h1DistNorm(computedSol,exactSol,dExactSol,exactSolIsParam);
    h1DistNorm.compute();
    gsSeminormH1<real_t> h1Norm(computedSol);
    h1Norm.compute();
    gsNormL2<real_t> l2DistNorm(computedSol,exactSol,exactSolIsParam);
    l2DistNorm.compute();
    gsNormL2<real_t> l2Norm(computedSol);
    l2Norm.compute();
    return sqrt(l2DistNorm.value()*l2DistNorm.value() + h1DistNorm.value()*h1DistNorm.value() + h2DistNorm.value()*h2DistNorm.value())/
            sqrt(l2Norm.value()*l2Norm.value() + h1Norm.value()*h1Norm.value() + h2Norm.value()*h2Norm.value());
}

//============================================== UTILS ===============================================//

void addAllDirichletBoundaries(const gsMultiPatch<>& mp,gsFunction<real_t>& g,gsBoundaryConditions<real_t> & bcInfo)
{
    std::vector<patchSide> boundaries = mp.boundaries();
    for(unsigned i = 0;i<boundaries.size();++i)
        bcInfo.addCondition(boundaries[i].patch, boundaries[i].side(),  condition_type::dirichlet, &g);
}

bool solveSystemCG(const gsSparseMatrix<>& matrix,const gsMatrix<> rhs,gsMatrix<>& solVector,real_t tolerance = 1e-10,unsigned iterations=100000000)
{
    gsSparseSolver<>::CGDiagonal solver;
    solver.setMaxIterations(iterations);
    solver.setTolerance(tolerance);
    solver.compute(matrix);
    solVector = solver.solve( rhs );
    if(solver.error()>=tolerance)
        std::cout << "Solver error: " << solver.error() << "\n";
    return solver.error()<tolerance;
}

template<typename T>
bool solveSystemBiCGSTAB(const gsSparseMatrix<>& matrix,const gsMatrix<> rhs,gsMatrix<>& solVector,real_t tolerance = 1e-10,unsigned iterations=100000000)
{
    gsSparseSolver<>::BiCGSTABILUT solver(matrix);
    solver.setMaxIterations(iterations);
    solver.setTolerance(tolerance);
    if (solver.preconditioner().info() != Eigen::Success)
    {
        gsWarn << "The preconditioner failed. Aborting...\n";
        return false;
    }
    solVector = solver.solve(rhs);
    if(solver.error()>=tolerance)
        std::cout << "Solver error: " << solver.error() << "\n";
    return solver.error()<tolerance;
}

template <typename T>
gsMatrix<T> reconstructFullCoefs( const gsMatrix<T>& dirichletVal,const gsMatrix<T>& solVal )
{
    const int sz  = dirichletVal.rows()+solVal.rows();
    const int dim = 1;

    gsMatrix<T> coeffs(sz, dim);
    for (index_t i = 0; i < sz; ++i)
    {
        if ( i<solVal.rows() ) // internal or interface
            coeffs.row(i) = solVal.row(i);
        else // eliminated Dof: fill with Dirichlet data
            coeffs.row(i).array() = dirichletVal.row( i-solVal.rows() );
    }
    return coeffs;
}

template <unsigned d,class T>
void gsRefineMarkedElements(gsCompositeHBasis<d,T> & basis,
                            const std::vector<bool> & elMarked,
                            int refExtension = 0)
{
    const int dim = basis.dim();

    // numMarked: Number of marked cells on current patch, also currently marked cell
    // poffset  : offset index for the first element on a patch
    // globalCount: counter for the current global element index
    int numMarked, poffset = 0, globalCount = 0;

    // refBoxes: contains marked boxes on a given patch
    gsMatrix<T> refBoxes;

    for (unsigned pn=0; pn < basis.nPatches(); ++pn )// for all patches
    {
        // Get number of elements to be refined on this patch
        const int numEl = basis.getBase(pn).numElements();
        numMarked = std::count_if(elMarked.begin() + poffset,
                                  elMarked.begin() + poffset + numEl,
                                  std::bind2nd(std::equal_to<bool>(), true) );
        poffset += numEl;
        refBoxes.resize(dim, 2*numMarked);
        //gsDebugVar(numMarked);
        numMarked = 0;// counting current patch element to be refined

        // for all elements in patch pn
        typename gsBasis<T>::domainIter domIt = basis.basis(pn).makeDomainIterator();
        for (; domIt->good(); domIt->next())
        {
            if( elMarked[ globalCount++ ] ) // refine this element ?
            {
                const gsVector<T> & ctr = domIt->centerPoint();

                // Construct degenerate box by setting both
                // corners equal to the center
                refBoxes.col(2*numMarked  ) =
                refBoxes.col(2*numMarked+1) = ctr;

                // Advance marked cells counter
                numMarked++;
            }
        }
        // Refine all of the found refBoxes in this patch
        basis.refineWithExtension( pn, refBoxes, refExtension,false );
    }
}

void getConvergenceRates(const std::vector<real_t>& errors,std::vector<real_t>& rates)
{
    GISMO_ASSERT(errors.size()>1,"size too small for a rate");
    for(unsigned i = 0;i<errors.size()-1;++i)
        rates.push_back(errors[i]/errors[i+1]);
}


//========================================= POISSON EQUATION =========================================//

void solvePoisson(gsCompositeBasis<2,real_t> * compBasis,const gsBoundaryConditions<real_t> & bcInfo,
                  gsMultiPatch<real_t> * domain,
                  const gsFunctionExpr<real_t>& f,
                  real_t& time,gsMatrix<real_t>& reconstructedSol)
{
    // recipe based version

    gsMatrix<>        rhs;
    gsMatrix<>        dirichlet_values;

    gsStopwatch stopwatch;

    gsPoissonPde<real_t>poissonPde(*domain,bcInfo,&const_cast<gsFunctionExpr<real_t>&>(f));
    gsRecipeAssemblerPoisson assembler(poissonPde);
    assembler.setDirichletStrategy(dirichlet::elimination);
    std::vector<const gsBasis<real_t>*> bases;
    for(unsigned i =0;i<compBasis->nPatches();++i)
        bases.push_back(&compBasis->getBase(i));
    std::vector<gsPhysicalSpace*> physSpaces;
    physSpaces.push_back(new gsPhysicalSpaceScalar(bases,*domain,INVERSE_COMPOSITION,compBasis->getMapper()));
    assembler.setSpace(physSpaces);
    assembler.setZeroAverage(false);
    assembler.assemble();

    rhs=assembler.getSystemRhs();
    if ( assembler.getElimSize()>0 )
    {
        gsSparseSolver<>::CGDiagonal solver;
        dirichlet_values = solver.compute(assembler.getEliminatedMatrix()).solve(assembler.getEliminatedRhs());
        rhs-=  assembler.getRhsModMatrix()*dirichlet_values;
    }

    gsMatrix<> solVector;
    bool successfulSolve = solveSystemCG(assembler.getSystemMatrix(),rhs,solVector);
    if(!successfulSolve)
        std::cout<<"The solver was not able to reach the desired tolerance.\n";
    GISMO_ASSERT(successfulSolve,"The solver was not able to reach the desired tolerance.");

    reconstructedSol = assembler.reconstructSolution(0,solVector,dirichlet_values);

    time = stopwatch.stop();
}

void solvePoissonCollectResults_oneStep(
        gsCompositeBasis<2,real_t> * compBasis,
        const gsBoundaryConditions<real_t> & bcInfo,
        gsMultiPatch<real_t> * domain,
        const gsFunctionExpr<real_t>& g,const gsFunctionExpr<real_t>& dg,const gsFunctionExpr<real_t>& f,
        int iteration,
        real_t& h,real_t& l2error,real_t& h1error,
        unsigned& dof,real_t& time,int& plotNr,bool plot=false)
{
    gsMatrix<real_t> reconstructedSol;
    gsCompositeBasis<2,real_t> * basis = dynamic_cast<gsCompositeBasis<2,real_t> *>(compBasis->clone());
    solvePoisson(basis,bcInfo,domain,f,time,reconstructedSol);
    std::cout << "goodSol: " << reconstructedSol << std::endl;
    std::vector<gsBasis<real_t> *> vc_refine_bases;
    for(unsigned i = 0;i<compBasis->nPatches();i++)
    {
        vc_refine_bases.push_back(new gsCompositeSingleBasis<2,real_t>(compBasis->getCompositeSingleBasis(i)));
    }
    gsCompositeGeom<2,real_t> geo(*compBasis,reconstructedSol);
    gsMultiPatch<real_t> * mp2 = geo.exportToPatches();
    std::vector<gsFunction<real_t>* > solvedfunctions;
    for (unsigned k=0; k<mp2->nPatches(); ++k)
    {
        solvedfunctions.push_back( mp2->patch(k).clone() );
    }
    gsField<real_t> sol(*domain,solvedfunctions);

    if(plot)
    {
        gsFunctionExpr<real_t> g_copy = g;
        gsField<real_t> exact(*domain, g_copy, false);
        gsWriteParaview( sol , "Poisson"+internal::toString(plotNr)+"solutionComposite"+internal::toString(iteration), compBasis->size()*2, true );
        gsWriteParaview( exact , "Poisson"+internal::toString(plotNr)+"solutionExact"+internal::toString(iteration), compBasis->size()*2, false );

        gsField<real_t>::uPtr errorPtr = gsFieldCreator<real_t>::absError(sol,g);
        gsWriteParaview( *errorPtr,"Poisson"+internal::toString(plotNr)+"solutionAbsErrComposite"+internal::toString(iteration), 10000, false);
        for(int i = 0;i<domain->size();i++)
        {
            gsMultiPatch<real_t> pt;
            pt.addPatch(domain->patch(i).clone());
            gsGradientField<real_t> grad(domain->patch(i),sol.function(i));
            gsField<real_t> gradField(pt,grad,true);
            gsWriteParaview(gradField , "Poisson"+internal::toString(plotNr)+"solutionGradFieldIter"+internal::toString(iteration)+"Patch"+internal::toString(i), 10000);
        }
        std::cout << "prints done - ";
    }
    // Find the element size
    h = math::pow( (real_t) vc_refine_bases[0]->size(), -1.0 / vc_refine_bases[0]->dim() );

    l2error = compositeL2Error(sol,g);
    h1error = compositeH1Error(sol,g,dg);

    dof=compBasis->size();

    delete mp2;
    delete basis;
    //cout<<"iteration " << iteration << " done!\n" << std::flush;
}

////======================================== BIHARMONIC EQUATION =======================================//

void solveBiharmonic(gsCompositeBasis<2,real_t> * compBasis,const gsBoundaryConditions<real_t> & bcInfo,
                  gsMultiPatch<real_t> * domain,
                  gsFunctionExpr<real_t>& f,
                  real_t& time,gsMatrix<real_t>& reconstructedSol)
{
    // recipe based version

    gsMatrix<>        dirichlet_values;

    gsStopwatch stopwatch;

    std::vector<const gsBasis<real_t>*> bases;
    for(unsigned i =0;i<compBasis->nPatches();++i)
        bases.push_back(&compBasis->getBase(i));
    gsRecipeAssemblerBiharmonic assembler(*domain,f,bases,compBasis->getMapper());
    assembler.fixBoundaries();
    assembler.assemble();

    gsMatrix<> solVector;
    bool successfulSolve = solveSystemCG(assembler.getSystemMatrix(),assembler.getSystemRhs(),solVector);
    if(!successfulSolve)
        std::cout<<"The solver was not able to reach the desired tolerance.\n";
    GISMO_ASSERT(successfulSolve,"The solver was not able to reach the desired tolerance.");

    dirichlet_values.setZero(assembler.getElimSize(),1);
    reconstructedSol=assembler.reconstructSolution(0,solVector,dirichlet_values);

    time = stopwatch.stop();
}

void solveBiharmonicCollectResults_oneStep(
        gsCompositeBasis<2,real_t> * compBasis,
        const gsBoundaryConditions<real_t> & bcInfo,
        gsMultiPatch<real_t> * domain,
        gsFunctionExpr<real_t>& g,gsFunctionExpr<real_t>& dg,gsFunctionExpr<real_t>& ddg,gsFunctionExpr<real_t>& f,
        int iteration,
        real_t& h,real_t& l2error,real_t& h1error,real_t& h2error,
        unsigned& dof,real_t& time,int& plotNr,bool plot=false)
{
    gsMatrix<real_t> reconstructedSol;
    gsCompositeBasis<2,real_t> * basis = dynamic_cast<gsCompositeBasis<2,real_t> *>(compBasis->clone());
    solveBiharmonic(basis,bcInfo,domain,f,time,reconstructedSol);
    std::vector<gsBasis<real_t> *> vc_refine_bases;
    for(unsigned i = 0;i<compBasis->nPatches();i++)
    {
        vc_refine_bases.push_back(new gsCompositeSingleBasis<2,real_t>(compBasis->getCompositeSingleBasis(i)));
    }
    gsCompositeGeom<2,real_t> geo(*compBasis,reconstructedSol);
    gsMultiPatch<real_t> * mp2 = geo.exportToPatches();
    std::vector<gsFunction<real_t>* > solvedfunctions;
    for (unsigned k=0; k<mp2->nPatches(); ++k)
    {
        solvedfunctions.push_back( mp2->patch(k).clone() );
    }
    gsField<real_t> sol(*domain,solvedfunctions);

    if(plot)
    {
        gsField<real_t> exact(*domain, g, false);
        gsWriteParaview( sol , "Biharmonic"+internal::toString(plotNr)+"solutionComposite"+internal::toString(iteration), 1000, true );
        gsWriteParaview( exact , "Biharmonic"+internal::toString(plotNr)+"solutionExact"+internal::toString(iteration), 1000, false );

        gsField<real_t>::uPtr errorPtr = gsFieldCreator<real_t>::absError(sol,g);
        gsWriteParaview( *errorPtr,"Biharmonic"+internal::toString(plotNr)+"solutionAbsErrComposite"+internal::toString(iteration), 1000//compGeom->getCompBasis().size()*2
                         , false);
        for(int i = 0;i<domain->size();i++)
        {
            gsMultiPatch<real_t> pt;
            pt.addPatch(domain->patch(i).clone());
            gsGradientField<real_t> grad(domain->patch(i),sol.function(i));
            gsField<real_t> gradField(pt,grad,true);
            gsWriteParaview(gradField , "Biharmonic"+internal::toString(plotNr)+"solutionGradFieldIter"+internal::toString(iteration)+"Patch"+internal::toString(i), 1000);
        }
        std::cout << "prints done - ";
    }
    // Find the element size
    h = math::pow( (real_t) vc_refine_bases[0]->size(), -1.0 / vc_refine_bases[0]->dim() );

    l2error = compositeL2Error(sol,g);
    h1error = compositeH1Error(sol,g,dg);
    h2error = compositeH2Error(sol,g,dg,ddg);

    dof=compBasis->size();

    delete mp2;
    //std::cout<<"iteration " << iteration << " done!\n" << std::flush;
}

////========================================== QUALITY MEASURES ========================================//

void solveMeasure(gsCompositeBasis<2,real_t> * compBasis,
                  gsMultiPatch<real_t>& domain,
                  real_t& time,gsMatrix<real_t>& reconstructedSol,
                  gsMatrix<real_t>* qualityMeasureVals,
                  bool fixBoundaries,
                  const gsQualityMeasureWeights& weights)
{
    // recipe based version
    gsStopwatch stopwatch;

    std::vector<const gsBasis<real_t>*> bases;
    for(unsigned i =0;i<compBasis->nPatches();++i)
        bases.push_back(&compBasis->getBase(i));
    gsRecipeAssemblerQMOpt2D assembler(domain,bases,compBasis->getMapper(),weights,
                                           true,true,qualityMeasureVals==NULL,false,false,weights.m_approxPoints!=0);
    if(fixBoundaries)
        assembler.fixBoundaries();
    assembler.assemble();

    gsSparseMatrix<real_t>  stiff = assembler.getSystemMatrix();
    gsMatrix<real_t>        rhs = assembler.getSystemRhs();

    gsMatrix<> solVector;

    bool successfulSolve = solveSystemBiCGSTAB<real_t>(stiff,rhs,solVector,1e-10,10000);
    if(!successfulSolve)
        std::cout<<"The solver was not able to reach the desired tolerance.\n"<<std::endl;
    if(!successfulSolve)
        successfulSolve = solveSystemCG(stiff,rhs,solVector,1e-2,10000);
    GISMO_ASSERT(successfulSolve,"The solver was not able to reach the desired tolerance.");

    gsMatrix<>        dirichlet_values;
    dirichlet_values.setZero(assembler.getElimSize(),1);
    reconstructedSol = assembler.reconstructSolution(0,solVector,dirichlet_values);

    time = stopwatch.stop();
}

void calculateAreas(gsCompositeBasis<2,real_t> * compBasis,
                    gsMultiPatch<real_t>& domain,
                    gsMatrix<real_t>& areas)
{
    // recipe based version
    std::vector<const gsBasis<real_t>*> bases;
    for(unsigned i =0;i<compBasis->nPatches();++i)
        bases.push_back(&compBasis->getBase(i));
    gsQualityMeasureWeights weights(compBasis->nPatches());
    gsRecipeAssemblerQMOpt2D assembler(domain,bases,compBasis->getMapper(),weights,
                                           false,false,false,true,false,false);
    assembler.assemble();
    assembler.getAreas(areas);
}

void calculateMeasures(gsCompositeBasis<2,real_t> * compBasis,
                        gsMultiPatch<real_t>& domain,
                        gsMatrix<real_t>& qualityMeasures,
                       const gsQualityMeasureWeights& weights)
{
    // recipe based version
    std::vector<const gsBasis<real_t>*> bases;
    for(unsigned i =0;i<compBasis->nPatches();++i)
        bases.push_back(&compBasis->getBase(i));
    gsRecipeAssemblerQMOpt2D assembler(domain,bases,compBasis->getMapper(),weights,
                                           false,false,true,false,false,false);
    assembler.assemble();
    assembler.getQualityMeasureVals(qualityMeasures);
}

} // end namespace gismo
