/** @file gsCompositeBasis.h

    @brief Provides declaration of Basis abstract interface.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): F. Buchegger
*/

#pragma once

#include <gsCore/gsBoxTopology.h>
#include <gsCore/gsMultiPatch.h>

#include <gsSmoothPatches/gsCompositeSingleBasis.h>
#include <gsMapUtils/gsMapper.h>
#include <gsMapUtils/gsMapFactoryFile.h>
#include <gsCore/gsDomainIterator.h>

namespace gismo
{

// forward declarations of the mapper classes
template<unsigned d,class T>
class gsCompositeSingleBasis;

template<unsigned d,class T>
class gsCompositeBasis
{
private:
    typedef T weightType;
    typedef index_t indexType; //indizes of gsMatrix
    typedef std::vector<std::pair<int,int> >::iterator step_iter;
    typedef gsBasis<T> BasisType;
    typedef typename std::vector<BasisType *>::const_iterator ConstBasisIter;
    typedef typename std::vector<BasisType *>::iterator BasisIter;
    typedef typename std::vector<gsMatrix<T> *>::const_iterator ConstMatrixPtrIter;
    typedef typename gsSparseMatrix<weightType,0,indexType>::InnerIterator InIterMat;
    typedef typename std::vector<indexType> IndexContainer;
    typedef typename std::vector<indexType>::const_iterator ConstIndexIter;
    typedef typename std::vector<weightType> WeightContainer;
    typedef typename std::vector<weightType>::const_iterator ConstWeightIter;
    typedef Eigen::PermutationMatrix<Dynamic,Dynamic,index_t> gsPermutationMatrix;

    typedef memory::auto_ptr< gsDomainIterator<T> > domainIter;

public:
    gsCompositeBasis()
    {
        m_mapFactory = NULL;
        m_mapper = NULL;
    }

    gsCompositeBasis( gsMultiPatch<T> const & mp, std::string pathToMap );

    gsCompositeBasis(gsMapFactory * factory)
    {
        m_mapFactory = factory;
        m_mapper = factory->makeMapper();
    }

    gsCompositeBasis( const gsCompositeBasis& other );

    virtual ~gsCompositeBasis();

public:
    //////////////////////////////////////////////////
    // getters for the private fields
    //////////////////////////////////////////////////

    /// getter for (const) m_bases[i]
    BasisType const & getBase(int i) const
    { return *m_bases[i]; }

    /// getter for m_bases[i]
    BasisType & getBase(int i)
    { return *m_bases[i]; }

     /// getter for m_bases
    std::vector<const BasisType*> getBases() const;

    /// getter for m_mapper
    gsMapper const & getMapper() const
    { return *m_mapper; }

    /// getter for m_mapper
    gsMapper * getMapPointer() const
    { return m_mapper; }

    /// getter for m_mapper
    gsMapFactory const & getMapFactory() const
    { return *m_mapFactory; }

    /// getter for m_topol
    gsBoxTopology const & getTopol() const
    { return m_topol; }

public:
    //////////////////////////////////////////////////
    // general functions for interacting with this class
    //////////////////////////////////////////////////

    /// Clone function. Used to make a copy of a derived basis
    virtual gsCompositeBasis * clone() const
    { return new gsCompositeBasis<d,T>(*this); }

    /// Prints the object to the stream
    void print(std::ostream & os) const { }

    unsigned dim() const
    {
        GISMO_ASSERT(m_bases.size()>0,"there should be at least one basis provided.");
        return m_bases[0]->dim();
    }

    /// Returns the approximate number of (global) basis functions in the patch with given \a index
    /// all the sizes summed up should give the total number of (global) basis functions
    /// the default argument -1 will give the total number of basis functions for all patches
    unsigned size(const int index=-1) const;

    /// Returns the number of local basis functions in the basis
    unsigned localSize() const { return m_mapper->getNrOfSources(); }

    /// Returns the number of local basis functions of the patch with given \a index in the basis
    unsigned localSize(const int index) const
    { return m_bases[index]->size(); }

    /// Returns the maximal polynomial degree of the patches.
    int maxDegree() const;

    /// Returns the polynomial degree of direction \a i of \a patch
    int degree(const unsigned patch,const int i) const
    { return m_bases[patch]->degree(i); }

    /// returns the amount of patches of the multi patch
    unsigned nPatches() const
    { return m_bases.size(); }

private:
    /// helper function for boundary and innerBoundaries
    void addLocalIndizesOfPatchSide(const patchSide& ps,unsigned offset,std::vector<index_t>& locals) const;

public:
    void reorderDofs(const gsPermutationMatrix& permMatrix)
    {
        (*m_mapper)*=permMatrix;
        m_mapper->optimize(gsMapper::optSourceToTarget);
    }

    void reorderDofs_withCoef(const gsPermutationMatrix& permMatrix,gsMatrix<T>& coefs)
    {
        reorderDofs(permMatrix);
        coefs*=permMatrix;
    }

    /// gets all indices of global basis functions on the boundary
    /// upto a given offset
    void boundary(std::vector<index_t> & indices,unsigned offset = 0) const;

    /// gets all indices of global basis functions on the inner boundary
    /// upto a given offset
    void innerBoundaries(std::vector<index_t> & indices,unsigned offset = 0) const;

    /// gives back the gsCompositeSingleBasis object set to the patch i, which
    /// ressembles the composite basis on one patch
    gsCompositeSingleBasis<d,T> getCompositeSingleBasis(const unsigned i)
    { return gsCompositeSingleBasis<d,T>(this,i); }

    /// gives back the domain iterator of the boundary side \a s of a given \a patch
    domainIter makeDomainIterator(const int patch,const boxSide & s) const
    { return m_bases[patch]->makeDomainIterator(s); }

    /** exports the patch \a i of this geometry (with coefs) to a Geometry object
     *  of the underlying basis. The ownership of this Geometry will go to the
     *  caller of this function.
     *
     * \param[in] localCoef : the coefficients to the local basis functions
     */
    gsGeometry<T>* exportPatch(const int i,gsMatrix<T> const & localCoef) const;

    /** exports this geometry (with coefs) to a gsMultiPatch object. The
     *  ownership of this gsMultipatch will go to the caller of this function.
     *
     * \param[in] localCoef : the coefficients to the local basis functions
     */
    gsMultiPatch<T>* exportToPatches(gsMatrix<T> const & localCoef) const;

public:
    //////////////////////////////////////////////////
    // functions for evaluating and derivatives
    //////////////////////////////////////////////////

    /** \brief Returns the indices of active (non-zero) basis functions of \a patch
     * at points <em>u</em>, as a list of indices, in <em>result</em>.
     *
     * \param[in] u  gsMatrix containing evaluation points. Each column represents one evaluation point.
     * \param[out]  result For every column \a i of \a u, a column containing the indices of the
     *   active basis functions at evaluation point <em>u</em>.col(<em>i</em>).
     */
    void active_into(const unsigned patch, const gsMatrix<T> & u,
                     gsMatrix<unsigned>& result) const; //global BF active on patch at point

    /// Returns the number of active (nonzero) basis functions at points \a u in \a result.
    virtual void numActive_into(const unsigned patch,const gsMatrix<T> & u, gsVector<unsigned>& result) const
    {
        GISMO_NO_IMPLEMENTATION
    }

    /// @name Evaluation functions
    /// @{

    /// \brief Evaluates nonzero basis functions of \a patch at point \a u into \a result.
    ///
    /// Let...\n
    /// \a d denote the dimension of the parameter domain.\n
    /// \a k denote the number of active (i.e., non-zero) basis functions (see active_into()).
    /// \a n denote the number of evaluation points.\n
    ///
    /// The \a n <b>evaluation points \a u</b> are given in a gsMatrix of size <em>d</em> x <em>n</em>.
    /// Each column of \a u represents one evaluation point.\n
    /// \n
    /// The gsMatrix <b>\a result</b> contains the computed function values in the following form:\n
    /// Column \a j of \a result corresponds to one evaluation point (specified by the <em>j</em>-th column of \a u).
    /// The column contains the values of all active functions "above" each other.\n
    ///
    /// For example, for scalar basis functions \a Bi : (x,y,z)-> R, a colum represents\n
    /// (B1, B2, ... , Bn)^T,\n
    /// where the order the basis functions \a Bi is as returned by active() and active_into().
    ///
    /// \param[in] u Evaluation points given as gsMatrix of size <em>d</em> x <em>n</em>.
    /// See above for details.
    /// \param[in,out] result gsMatrix of size <em>k</em> x <em>n</em>.
    /// See above for details.
    ///
    void eval_into(const unsigned patch, const gsMatrix<T> & u, gsMatrix<T>& result ) const;
    void deriv_into(const unsigned patch, const gsMatrix<T> & u, gsMatrix<T>& result ) const;
    void deriv2_into(const unsigned patch, const gsMatrix<T> & u, gsMatrix<T>& result ) const;

    /// Evaluate the \a global_BF-th basis function on \a patch at points \a u into \a result.
    void evalSingle_into(const unsigned patch, const int global_BF, const gsMatrix<T> & u, gsMatrix<T>& result ) const;
    void derivSingle_into(const unsigned patch, const int global_BF, const gsMatrix<T> & u, gsMatrix<T>& result ) const;
    void deriv2Single_into(const unsigned patch, const int global_BF, const gsMatrix<T> & u, gsMatrix<T>& result ) const;

    /// @brief Evaluate the nonzero basis functions of \a patch and their derivatives up
    /// to order \a n at points \a u into \a result.
    void evalAllDers_into(const unsigned patch, const gsMatrix<T> & u, 
                          const int n, std::vector<gsMatrix<T> >& result ) const;

    /// @brief Evaluate the basis function \a global_BF at \a patch and its derivatives up
    /// to order \a n at points \a u into \a result.
    void evalAllDersSingle_into(const unsigned patch,const unsigned global_BF, const gsMatrix<T> & u,const int n,gsMatrix<T> & result ) const;

    /// @}

public:
    //////////////////////////////////////////////////
    // functions for converting global to local coefs and back
    //////////////////////////////////////////////////

    /** converts the global coefficients to the local ones
     *
     * \param[in] globalCoefs : the coefficients to the global basis functions
     * \param[out] localCoefs : the coefficients to the local basis functions
     */
    void global_coef_to_local_coef(gsMatrix<T> const & globalCoefs,gsMatrix<T> & localCoefs) const
    { m_mapper->mapToSourceCoefs(globalCoefs,localCoefs); }

    /** converts the local coefficients to the global ones
     *
     * \param[in] localCoefs : the coefficients to the local basis functions
     * \param[out] globalCoefs : the coefficients to the global basis functions
     */
    void local_coef_to_global_coef(gsMatrix<T> const & localCoefs,gsMatrix<T> & globalCoefs) const
    {  m_mapper->mapToTargetCoefs(localCoefs,globalCoefs); }

protected:
    //////////////////////////////////////////////////
    // functions for working with Indexes
    //////////////////////////////////////////////////

    /** gets the number of the patch, where the local basis function with the
     *  given index is on.
     *
     * \param[in] localIndex : the accumulated index of the basis function
     */
    unsigned _getPatch(unsigned localIndex) const;

    /** gets the patchIndex of the basis function specified by its localIndex.
     *  This is the index of this basis function in the basis of this patch.
     *
     * \param[in] localIndex : the accumulated index of the basis function
     */
    unsigned _getPatchIndex(const unsigned localIndex) const;

    /** gets the local index of the basis function of the patch with the
     *  given number.
     *
     * \param[in] patch : the number of the patch
     * \param[in] patchIndex : the index of the basis function on the patch
     */
    unsigned _getLocalIndex(unsigned const patch,unsigned const patchIndex) const
    { return _getFirstLocalIndex(patch)+patchIndex; }

    /** gets the local index of the first basis function of the patch with the
     *  given number.
     *
     * \param[in] patch : the number of the patch
     */
    unsigned _getFirstLocalIndex(unsigned const patch) const;

    /** gets the local index of the last basis function of the patch with the
     *  given number.
     *
     * \param[in] patch : the number of the patch
     */
    unsigned _getLastLocalIndex(unsigned const patch) const
    { return _getFirstLocalIndex(patch)+m_bases[patch]->size()-1; }

    // Data members
protected:
    /// topology, specifying the relation (connections) between the patches
    gsBoxTopology m_topol;
    /// vector of patches (bases)
    std::vector<BasisType *> m_bases;
    /// map between the local basis functions and the newly created ones
    gsMapper * m_mapper;
    /// map factory giving back the mapping
    gsMapFactory * m_mapFactory;
};

}


//////////////////////////////////////////////////
//////////////////////////////////////////////////

#ifndef GISMO_BUILD_LIB
#include GISMO_HPP_HEADER(gsCompositeBasis.hpp)
#endif
