/** @file gsCompositeGeom.h

    @brief Provides declaration of Basis abstract interface.

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): F. Buchegger
*/

#pragma once
#include <gsCore/gsGeometry.h>
#include <gsCore/gsMultiPatch.h>
#include <gsSmoothPatches/gsCompositeSingleBasis.h>
#include <gsSmoothPatches/gsCompositeBasis.h>
#include <gsCore/gsDomainIterator.h>

namespace gismo
{
template<unsigned d,class T>
class gsCompositeGeom : public gsGenericGeometry< gsCompositeSingleBasis<d,T> >
{

private:
    typedef gsCompositeSingleBasis<d,T> Basis;
    typedef gsGenericGeometry<Basis> Base;
public:
    /// Default empty constructor
    gsCompositeGeom() : Base() { }

    gsCompositeGeom( gsMultiPatch<T> const & mp,std::string pathToMap );

    /// Construct Geom by basis and coefficient matrix
    gsCompositeGeom( const gsCompositeBasis<d,T> & basis, const gsMatrix<T> & coefs );

    gsCompositeGeom( const gsCompositeGeom& other );

    gsCompositeGeom<d,T> & operator=( const gsCompositeGeom& other );

    ~gsCompositeGeom()
    { delete m_compBasis; } //destructor

public:

//////////////////////////////////////////////////
// Virtual base members with a new implementation
//////////////////////////////////////////////////

//////////////////////////////////////////////////
// Virtual member functions required by the base class
//////////////////////////////////////////////////

    /// Clone function. Used to make a copy of the (derived) geometry
    virtual gsCompositeGeom * clone() const
    { return new gsCompositeGeom(*this); }

    /// Prints the object as a string.
    std::ostream &print(std::ostream &os) const
    {
        m_compBasis->print(os);
        return os;
    }

    const gsCompositeBasis<d,T> & getCompBasis() const
    { return *m_compBasis; }

    gsCompositeBasis<d,T> & getCompBasis()
    { return *m_compBasis; }

    /// The gsBasisFun points to the i-th basis function of m_basis
    /// after calling this setter.
    void setBasis( unsigned const & i ) const
    { this->basis().setBasis(i); }


    /// Point to the first basis function of the basis
    void first()
    { this->basis().first(); }

    /// Points to the first basis function of the basis
    /// or return false if this is the last basis function
    bool next() const
    {  return this->basis().next(); }

    gsMultiPatch<T>* exportToPatches() const;

    gsGeometry<T> * exportPatch(int i,gsMatrix<T> const & localCoef) const
    { return m_compBasis->exportPatch(i,localCoef); }

// Data members
protected:
    gsCompositeBasis<d,T> * m_compBasis;

}; // class gsCompositeGeom

} // namespace gismo
