/** @file gsIETIScaledDirichlet.h
  
    @brief
    
    
    This file is part of the G+Smo library.
    
    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.
    
    Author(s): C. Hofer
    Created on: 2015-01-19
*/


#pragma once

#include <gsSolver/gsLinearOperator.h>
#include <gsAssembler/gsIETIUtils.h>
#include <gsSolverDev/gsIETISolver.h>
#include <gsAssembler/gsIETIAssembler.h>


namespace gismo {


/**
 * @brief This class represents the preconditioner for the IETI method.
 * It supports several strategies for scaling:
 *      none , multiplicity, coefficient.
 * Note that the coefficient scaling is currently only available for poisson equation
 *
 *let H be the maximal diameter of a patch
 *  none: in the worst case leads to a condition number of the systemmatrix of O( (max(1+log(H_i/h_i))^3 )
 *  multiplicity, coefficient: leads to a condition number of the systemmatrix of O( (max(1+log(H_i/h_i))^2 )
 *
 */
template<class T>
class gsScaledDirichletPrecond : public gsLinearOperator
{

    /// \brief shortcut for LU factorizations
    typedef typename gsIETIAssembler<T>::sparseLUfact sparseLUfact;

public:

    /**
     * @brief gsScaledDirichletPrecond The constructor for the preconditioner, requires the
     *  a gsIETIAssembler class and the strategy for scaling.
     * @param ass An instance of the gsIETIAssembler class
     * @param strat The desired method for scaling
     */
    gsScaledDirichletPrecond(const gsIETIAssembler<T>& ass,const IETIPrecondScaling::strategy& strat)
        : m_ass(ass),m_strat(strat) ,m_B(ass,strat), m_info(ass.getInfo())
    {
        m_Kib.resize(m_info.numberPatches);
        m_Kbb.resize(m_info.numberPatches);
        for(std::size_t np = 0; np<m_info.numberPatches;np++)
        {
            m_Kib[np] = m_ass.getKib(np);
            m_Kbb[np] = m_ass.getKbb(np);
        }

    }

    /**
     * @brief apply The main routine of the class, represents the application of the preconditioner
     * The algorithm is:
     * 1) apply the transposed jump operator (without embedding) input -> {w} on each patch
     * 2) scale the values of the boundary
     * 3) apply the Schur complement of the stiffness matrix (Kbb - Kbi*Kii^-1*Kib)
     * 4) scale the result again
     * 5) apply the jump operator  (without embedding) {w} on each patch -> {result}
     *
     * @param input lagrange multiplier (lagrangeMult x nRhs)
     * @param result the application of the preconditioner (lagrangeMult x nRhs)
     */
    void apply(const gsMatrix<T> & input, gsMatrix<T> & result) const
    {

        std::vector<gsMatrix<T> > w(m_info.numberPatches);


        //Apply transposed scaled jump operator
        //m_B.calculateScaling(w);
        m_B.applyTrans(input,w);


        //Apply S
        Eigen::setNbThreads(1);
#pragma omp parallel
        {
            gsMatrix<T> temp, wi, w_temp;
#pragma omp for schedule(static, 1) nowait
            for(std::size_t np =0; np<m_info.numberPatches;np++)
            {
                temp.noalias() = m_Kib[np]*w[np];
                m_ass.solveKii(np,temp,wi);
                wi.transposeInPlace();

                w_temp.noalias() = wi*m_Kib[np];
                w_temp.transposeInPlace();

                w[np] = m_Kbb[np]*w[np]-w_temp;
            }
        }
        Eigen::setNbThreads(0);
        //Apply scaled jump operator
        //m_B.calculateScaling(w);
        m_B.apply(w,result);

    }


    /// \brief return the number of rows
    index_t rows() const {return m_info.lagrangeMult;}

    /// \brief return the number of columns
    index_t cols() const {return m_info.lagrangeMult;}

    /**
     * @brief calculateMatrixForm The function provides the possibility to get the matrix representation of the apply function.
     * This is done by calling the apply function with all unit vector. The use is not recommended.
     * @param PF The matrix representation of the apply function. (lagrangeMult x lagrange Mult)
     */
    void calculateMatrixForm(gsMatrix<T>& PF)
    {
        PF.setZero(m_info.lagrangeMult,m_info.lagrangeMult);
        gsMatrix<T> I(m_info.lagrangeMult,m_info.lagrangeMult);
        I.setIdentity(m_info.lagrangeMult,m_info.lagrangeMult);

        apply(I,PF);
    }

    void calculateMatrixForm(std::vector<gsMatrix<T> >& S)
    {
        gsMatrix<> I,temp, wi, w_temp;
        for(std::size_t np =0; np<m_info.numberPatches;np++)
        {
            I.setIdentity(m_info.dofsB[np],m_info.dofsB[np] );

            temp.noalias() = m_Kib[np]*I;
            m_ass.solveKii(np,temp,wi);

            wi.transposeInPlace();

            w_temp.noalias() = wi*m_Kib[np];
            w_temp.transposeInPlace();

            S.push_back(m_Kbb[np]*I-w_temp);
        }

    }

    void calculateMatrixForm(gsMatrix<T>& B, gsMatrix<T>& BT)
    {
        int nB=0;
        for(unsigned np = 0;np<m_info.numberPatches;np++)
            nB+=m_info.dofsB[np];

        int nL = m_info.lagrangeMult;

        B.setZero(nL,nB);

        std::vector< gsMatrix<T> >inp(m_info.numberPatches);
        int idx=0;
        for(std::size_t np = 0; np< m_info.numberPatches;np++)
        {
            inp[np].setZero(m_info.dofsB[np],nB);
            inp[np].block(0,idx,m_info.dofsB[np],m_info.dofsB[np]).setIdentity(m_info.dofsB[np],m_info.dofsB[np]);

            idx+=m_info.dofsB[np];
        }

        m_B.apply(inp,B);

        gsMatrix<T> id(nL,nL);
        id.setIdentity(nL,nL);
        std::vector<gsMatrix<T> > res(m_info.numberPatches);
        m_B.applyTrans(id,res);

        BT.setZero(nB,nL);
        idx=0;
        for(std::size_t np = 0; np< m_info.numberPatches;np++)
        {
            BT.block(idx,0,m_info.dofsB[np],nL)=res[np];
            idx+=m_info.dofsB[np];
        }
    }
private:
    /// \brief reference of the gsIETIAssembler class
    const gsIETIAssembler<T>& m_ass;

    /// \brief strategy for scaling
    const IETIPrecondScaling::strategy& m_strat;

    /// \brief jump operator
    const gsIETIJumpOperator<T>  m_B;

    /// \brief info structure
    gsIETIInfo m_info;

    /// \brief for each patch the matrix Kib
    std::vector<gsSparseMatrix<T> > m_Kib;

    /// \brief for each patch the matrix Kbb
    std::vector<gsSparseMatrix<T> > m_Kbb;
};






} // namespace gismo

