/** @file gsIETISolver.h

    @brief


    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): C. Hofer
    Created on: 2014-12-02
*/


#pragma once

#include <Eigen/SparseLU>
#include <gsAssembler/gsIETIUtils.h>
#include <gsAssembler/gsIETIAssembler.h>
#include <gsAssembler/gsIETIdGAssembler.h>

#include <gsSolver/gsLinearOperator.h>



namespace gismo {


/**
 *  This class represents the jump operator and calculates the jumps of a given function across the interfaces.
 *   It proived the routine for calling the the operator, its transposed and supports a scaled verion suited for precontioning.
 */
template<typename T>
class gsIETIJumpOperator
{

    typedef typename gsIETIAssembler<T>::patchDof patchDof; //patch + local number

    /// \brief  the corresponding IETI assembler
    const gsIETIAssembler<T>& m_ass;

    /// \brief  scaling strategy
    IETIPrecondScaling::strategy scaling;

    /// \brief reference to the info structure of gsIETIAssembler
    const gsIETIInfo & m_info;

    /// \brief  contains for each interface dof on each patch its scaling
    std::vector<std::vector<T> > scalings;

    /// \brief  Stores for each patch the coefficient for scaling strategy
    std::vector<T> m_coeffs;

public:
    gsIETIJumpOperator(const gsIETIAssembler<T>& ass, IETIPrecondScaling::strategy sc = IETIPrecondScaling::none);

    /**
     * @brief applyTrans this function maps the lagrange multipliers (the jumps of the corresponding interface nodes)
     *   to the boundary values including a transformation to primal and remaining. This function combines applyBTrans(..) and gsIETIAssember.embeddingTrans
     * @param input the lagrange multipliers (lagrangeMult x nRhs)
     * @param uP the primal variables (totalPdofs x nRhs)
     * @param u2 the remaining variable for each patch (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     */
    void applyTransTil(const gsMatrix<T>& input,gsMatrix<T>&  uP, std::vector<gsMatrix<T> > &  u2) const;

    /**
     * @brief apply this function maps the primal variables and the remaining to the jumnps across the interface, (= the lagrange multipliers).
     * This functions combines applyB and gsIETIAssember.embedding
     * @param xP the primal variables (totalPdofs x nRhs)
     * @param x2 the remaining variable for each patch (if isMinimumEnergy: dofsB[np] x nRhs ; else dofsR[np] x nRhs)
     * @param result the lagrange multipliers (lagrangeMult x nRhs)
     */
    void applyTil(const gsMatrix<T>& xP, const std::vector<gsMatrix<T> > & x2, gsMatrix<T> & result) const;

    /**
     * @brief applyB this functions maps the values/coefficients on each patch to the jumps on the boundary
     * @param w for each patch it contains the inferface values/coefficients (dofsB x nRhs)
     * @param result the lagrange multipliers (lagrangeMult x nRhs)
     */
    void apply(const std::vector<gsMatrix<T> >& w, gsMatrix<T> & result) const;

    /**
     * @brief applyBTrans maps the lagrange multipliers (the jumps of the corresponding interface nodes)
     *   to the boundary values.
     * @param input the lagrange multipliers (lagrangeMult x nRhs)
     * @param w for each patch it contains the inferface values/coefficients (dofsB x nRhs)
     */
    void applyTrans(const gsMatrix<T>& input, std::vector<gsMatrix<T> >& w) const;


    //void calculateScaling(std::vector<gsMatrix<T> >& w) const;


    /**
     * @brief calculateMatrixForm calculates the matrix representation of applyB and applyBTrans included the scaling.
     * @param B the matrix representation for B
     * @param BT the matrix representation for B^T
     */
    void calculateMatrixForm(gsMatrix<T>& B, gsMatrix<T>& BT);

protected:

    /**
     * @brief calculateScaling given the opposite patch dof  (w.r.t. the lagrange multiplier), this function scales the values accoring to the set strategy for scaling. (see Dirichletpreconditioner).
     * If no preconditioning happens, all entries of scalings are 1, i.e. no scaling happends.
     * @param pd the opposite patch dof  (w.r.t. the lagrange multiplier)
     */
    inline T getScaling(const patchDof& pd) const;

    /**
     * @brief setCoeffs calculates the coefficients for scaling strategy. This factor is PDE dependent and may not be implemented.
     */
    void setCoeffs();
};





/**
 * \brief This class represents the application of the IETI system matrix and is
 * designed to fit as input for the gsConjugateGradient class. This means that this class is inherited from the abs It also returns the
 * corresponding right hand side.
 */
template<class T>
class gsIETISolver : public gsLinearOperator
{
public:
    /// \brief shortcut for LU-factorization
    typedef typename gsIETIAssembler<T>::sparseLUfact sparseLUfact;

    typedef  T Scalar;
    typedef real_t RealScalar;

protected:
    /**
     * @brief applyMinEnergy This funciton performs the application of the inverse Schur complement S
     * of the stiffness matrix via the minimal energy algorithm. It is part of the apply(..) function.
     *
     * @param inputP The primal part of the input argument. (dofsTotalP x nRhs)
     * @param inputD The dual part of the input argument. For each patch (dofsB x nRhs)
     * @param resultP The primal part of the result. (dofsTotalP x nRhs)
     * @param resultD The dual part of the result  For each patch (dofsB x nRhs)
     */
    void applyMinEnergy(const gsMatrix<T> & inputP, const std::vector<gsMatrix<T> > & inputD, gsMatrix<T> & resultP, std::vector<gsMatrix<T> > & resultD) const;


    /**
     * @brief applyStdIETI his funciton performs the application of the inverse of the stiffness matrix K
     * via the the standard algorithm. It is part of the apply(..) function.
     *
     * @param inputP The primal part of the input argument. (dofsTotalP x nRhs)
     * @param inputR The remaining part of the input argument. For each patch (dofsR x nRhs)
     * @param resultP The primal part of the result. (dofsTotalP x nRhs)
     * @param resultR The remaining part of the result  For each patch (dofsB x nRhs)
     */
    void applyStdIETI(const gsMatrix<T> & inputP, const std::vector<gsMatrix<T> > & inputR, gsMatrix<T> & resultP, std::vector<gsMatrix<T> > & sresultR) const;


    /*
     *This functions are currently not used and needed.
     *
    void inline parseArray(const gsMatrix<T> & input, gsMatrix<T> &result1, std::vector<gsMatrix<T> >&result2) const
    {
        result1 = input.topRows(info.dofTotalP);
        parseStdVec(input.bottomRows(info.dofTotal-info.dofTotalP), result2);
    }

    void mergeArray(const gsMatrix<T> & x1,const std::vector<gsMatrix<T> > &x2, gsMatrix<T> &x) const
    {
        x.topRows(info.dofTotalP)= x1;
        mergeStdVec(x2, x.bottomRows(info.dofTotal-info.dofTotalP));
    }

    void inline parseStdVec(const gsMatrix<T>& input, std::vector<gsMatrix<T> >& result ) const
    {
        result.reserve(info.numberPatches);

        int indx = 0;
        for(index_t k=0; k<info.numberPatches;k++)
        {
            result.push_back(input.block(0,indx,info.dofs[k],info.nRhs));
            indx+= info.dofs[k];
        }
    }

    void inline mergeStdVec(const std::vector<gsMatrix<T> >& input, gsMatrix<T>& result )const
    {
        result.reserve(info.dofTotal);

        int indx = 0;
        for(index_t k=0; k<info.numberPatches;k++){
            result.block(0,indx,info.dofs[k],info.nRhs) = input[k];
            indx+= info.dofs[k];
        }
    }
    */


public:


    /**
     * @brief gsIETISolver Constructor of the IETISolver class.
     * @param assembler An instance of the gsIETIAssembler class. Based on this class gsIETISolver creates
     *  the jump operator B (gsIETIJumpOperator) and computes the appropriate rhs.
     */
    gsIETISolver(gsIETIAssembler<T>& assembler): m_B(assembler), m_ass(assembler)
    {

        info=assembler.getInfo();

        m_Krp.resize(info.numberPatches);
        m_rhs.resize(info.lagrangeMult,info.nRhs);


        gsMatrix<T> rhs_p, rhs_p_t;
        std::vector<gsMatrix<T> > rhs_r(info.numberPatches);
        std::vector<gsMatrix<T> > rhs_r_t(info.numberPatches);


        assembler.getRawRhs(rhs_p,rhs_r);

        if(info.isMinimumEnergy){
            //process rhs
            applyMinEnergy(rhs_p,rhs_r,rhs_p_t,rhs_r_t);
        }
        else
        {
            m_Krp.resize(info.numberPatches);
            for(std::size_t k = 0;k < info.numberPatches;k++)
                m_Krp[k] = assembler.getKrp(k);

            applyStdIETI(rhs_p,rhs_r,rhs_p_t,rhs_r_t);
        }

        m_B.applyTil(rhs_p_t,rhs_r_t,m_rhs);
        if(!m_ass.nCoupledElimDofs())
            m_rhs-= m_ass.getDirRhs();

    }

    /**
     * @brief Destructor, does nothing. Contained pointers are freed in gsIETIAssembler
     */
    ~gsIETISolver()
    {
    }

    /**
     * @brief getRhs returns the rhs used for solving the system.
     * @return the corresponding rhs
     */
    gsMatrix<T>& getRhs()
    {
        return m_rhs;
    }

    /**
     * @brief apply The main routine of the class, performs the application of the system matrix. The application work in the following way:
     *
     * 1) B^T(input) -> u;
     * 2) applyMinEnergy(u) or applyStdIETI(u) -> v;
     * 3) B(v) -> result;
     *
     * The application of B includes:
     * 1) Embedding : inputP,inputD -> {input} on each patch
     * 2) Application of B: {input} on each patch -> result
     *
     * The application of B^T is vice verca
     * @param input The lagrange multipliers (lagrangeMult x nRhs)
     * @param result The application of F on the lagrange multipliers (lagrangeMult x Rhs)
     */
    void apply(const gsMatrix<T> & input, gsMatrix<T> & result) const
    {
        gsMatrix<T> u1(info.dofTotalP,input.cols());
        std::vector<gsMatrix<T> > u2(info.numberPatches);

        gsMatrix<T> x1(info.dofTotalP,input.cols());
        std::vector<gsMatrix<T> > x2(info.numberPatches);

        m_B.applyTransTil(input, u1, u2);

        if(info.isMinimumEnergy){
            applyMinEnergy(u1,u2, x1,x2);
        }
        else{
            applyStdIETI(u1, u2, x1, x2);
        }

        m_B.applyTil(x1, x2, result);

    }

    /**
     * @brief rows returns the rows of the matrix. Inherited from base class
     * @return number of rows
     */
    index_t rows() const {return info.lagrangeMult;}

    /**
     * @brief cols returns the columns of the matrix. Inherited from base class
     * @return number of columns
     */
    index_t cols() const {return info.lagrangeMult;}


    /**
     * @brief calculateSolution After the solution of the system is found (e.g. via PCG), the lagrange multipliers are
     * used to calculate the solution in the form (primal, {dual} ). In order to receive a solution which can be plotted
     * use the function processSolution of gsIETIAssembler as with this data.
     * @param lambda The lagrange multipliers, i.e. solution where the PCG uses the apply function
     * @param resP The primal part of the solution
     * @param resR The dual/remaining part of the solution    /// \brief  Stores for each patch the coefficient for scaling strategy
    std::vector<T> m_coeffs;
     */
    void calculateSolution(const gsMatrix<T> & lambda, gsMatrix<T> & resP, std::vector<gsMatrix<T> > & resR) const;

    /**
     * @brief calculateSolution combines the functions calculateSolution(const gsMatrix<T> & lambda, gsMatrix<T> & resP, std::vector<gsMatrix<T> > & resR) and
     * gsIETIAssembler::processSolution(gsMatrix<T> &  resP,std::vector<gsMatrix<T> > & resR, gsMatrix<T> & solVec). Returns for given lagrange multipliers,
     * the solution of system in vector/matrix form. (like the output of the solver without using IETI)
     * @param lambda the lagrange multipliers
     * @param solVec the solution in matrix form
     */
    void calculateSolution(const gsMatrix<T> &lambda, gsMatrix<T>& solVec) const
    {
        gsMatrix<T> resP;
        std::vector<gsMatrix<T> > resR(info.numberPatches);

        calculateSolution(lambda,resP,resR);

        m_ass.processSolution(resP,resR,solVec);
    }

    /**
     * @brief calculateSolution combines the functions calculateSolution(const gsMatrix<T> & lambda, gsMatrix<T> & resP, std::vector<gsMatrix<T> > & resR) and
     * gsIETIAssembler::processSolution(gsMatrix<T> &  resP,std::vector<gsMatrix<T> > & resR, int rhs). Returns for given lagrange multipliers,
     * the solution of system as a gsField*.
     * ASSUMES A SPECIFIC TREATMENT OF THE SOLUTION VECTOR, THIS MAY NOT WORK FOR EACH ASSEMBLER. Use instead calculateSolution(const gsMatrix<T> &lambda, const gsMatrix<T>& solVec)
     * and then a assembler specific constructSolution(..) routine.
     * @param lambda the lagrange multipliers
     * @param rhs in case of multiple rhs, which one shoukd be plotted.
     * @return a pointer to a gsField representing the solution
     */
   gsField<T>* calculateSolution(const gsMatrix<T> &lambda, int rhs=0) const
   {
       gsMatrix<T> resP;
       std::vector<gsMatrix<T> > resR(info.numberPatches);

       calculateSolution(lambda,resP,resR);
       return m_ass.processSolution(resP,resR,rhs);
   }


    /**
     * @brief calculateMatrixForm The function provides the possibility to get the matrix representation of the apply function.
     * This is done by calling the apply function with all unit vector. The use is not recommended.
     * @param F The matrix representation of the apply function. (lagrangeMult x lagrange Mult)
     */
    void calculateMatrixForm(gsMatrix<T>& F)
    {
        F.setZero(info.lagrangeMult,info.lagrangeMult);
        gsMatrix<T> I(info.lagrangeMult,info.lagrangeMult);
        I.setIdentity(info.lagrangeMult,info.lagrangeMult);

        apply(I,F);
    }

    void calculateMatrixForm(gsMatrix<T>& B, gsMatrix<T>& BT)
    {
        m_B.calculateMatrixForm(B,BT);
    }

    /*
    void calculateMatrixForm(std::vector<gsMatrix<T> >& S)
    {
        gsMatrix<T> IP,SP;
        std::vector<gsMatrix<T> > ID(info.numberPatches);
        std::vector<gsMatrix<T> > SD(info.numberPatches);


        std::vector<gsMatrix<T> >I(info.numberPatches);

        S.resize(info.numberPatches);
        for(unsigned np = 0; np< info.numberPatches;np++)
        {
            I[np].setIdentity(info.dofsB[np],info.dofsB[np]);
            S[np].setZero(info.dofsB[np],info.dofsB[np]);
         }

         m_ass.embeddingTrans(I,IP,ID);
         if(info.isMinimumEnergy){
             applyMinEnergy(IP,ID, SP,SD);
         }
         else{
             applyStdIETI(IP, ID, SP, SD);
         }

         m_ass.embedding(SP,SD,S);

    }
*/

protected:
    /// \brief the info structure of gsIETIAssembler
    gsIETIInfo info;

    /// \brief the jump operator
    gsIETIJumpOperator<T> m_B;

    /// \brief reference to the assembler class, if some informaton is required
    const gsIETIAssembler<T>& m_ass;

    /// \brief the right hand side of the system
    gsMatrix<T> m_rhs;

    /// \brief for each patch the matrix Krp
    std::vector<gsSparseMatrix<T> > m_Krp;



};





} // namespace gismo

//#ifndef GISMO_BUILD_LIB
//#include GISMO_HPP_HEADER(gsIETISolver.hpp)
//#endif

#include <gsSolverDev/gsIETISolver.hpp>

