//
// gsMultiGrid.h
//
// Multigrid solver for isogeometric discretizations.
//
// Clemens Hofreither
//
 
#pragma once

#include <gsCore/gsField.h>
#include <gsCore/gsDofMapper.h>
#include <gsCore/gsDebug.h>
#include <gsUtils/gsStopwatch.h>
#include <gsSolver/gsSmoother.h>
#include <gsSolver/gsLinearOperator.h>

namespace gismo
{

/** @brief
    Multigrid solver for single-patch isogeometric problems.

    (This documentation is partially outdated.)

    This class implements geometric multigrid for isogeometric patches.
    To use it, pass the gsBVProblem as well as the coarse-grid gsBasis to the
    constructor and tell it how many refinement steps you want to perform. The
    class supports many different options, but in the simplest case, you then
    simply call gsMultiGrid::solvePCG() to solve the problem using
    multigrid-preconditioned CG iteration and obtain a solution field.

    It is also possible to use gsMultiGrid::solve() to solve using pure multigrid
    iteration without CG, however, this can be quite slow for high-dimensional
    problems with high spline degrees. Using CG iteration is usually recommended.

    \par Additional options

    The solver can be configured to use V- or W-cycles using gsMultiGrid::setNumCycles().

    The smoother to use can be set with gsMultiGrid::setSmoother(). See also
    Smoother::SmootherType for possible arguments.

    The number of pre- and post-smoothing steps can be configured with
    gsMultiGrid::setNumPreSmooth() and gsMultiGrid::setNumPostSmooth(), respectively.

    Full multigrid (FMG) is supported and sometimes extremely efficient.
    It can be invoked by passing the proper \c bool flag to the constructor
    and then calling gsMultiGrid::solveFMG(). It should however be considered
    experimental since the theory is not well understood at this point.

    \par Nonsymmetric problems

    Nonsymmetric problems are partially supported. Things left to do:
      - the Gauss-Seidel smoother assumes a symmetric matrix right now, a bit more
        care has to be taken there. In the meanwhile, use the Jacobi smoother.
      - there should be a solveGMRES() method for nonsymmetric problems to speed
        up the convergence. In the meanwhile, use the basic gsMultiGrid::solve() iteration.
*/



/** Takes a matrix \a m which maps from all basis functions of a coarse level to a fine level
 *  and returns a matrix which is restricted to the free dofs. */
GISMO_EXPORT void restrictToFreeDofs(const Eigen::SparseMatrix<real_t>& m, const gsDofMapper& mapper_c, const gsDofMapper& mapper_f, Eigen::SparseMatrix<real_t>& result);



class GISMO_EXPORT gsMultiGrid : public gsLinearOperator
{

public:
    //typedef Eigen::SparseMatrix<real_t> SpMatrix;
    typedef gsSparseMatrix<real_t> SpMatrix;


    /// Default empty constructor
    gsMultiGrid() { }
    
    gsMultiGrid( const SpMatrix & fineMatrix, const std::vector<const gsDofMapper *>& dofMapper,
                 std::vector< gsSparseMatrix<real_t, RowMajor> > transferMatrices);

    /// Destructor
    ~gsMultiGrid();

public:

    /// Solve the problem using multigrid iteration and return the resulting solution matrix.
    gsMatrix<real_t> solve(const gsVector<real_t>& rhs_, real_t tol = 1e-8);

    /// Solve the problem using CG iteration using a multigrid preconditioner and return the resulting solution matrix.
    gsMatrix<real_t> solvePCG(const gsVector<real_t>& rhs_, real_t tol = 1e-8);

    /// Solve the problem using a single full multigrid cycle and return the resulting solution matrix.
    gsMatrix<real_t> solveFMG()
    {
        gsMatrix<real_t> result;
        this->fullMultiGrid( result );
        return result;
    }

    /// Solve the problem using cascadic multigrid (experimental).
    gsMatrix<real_t> solveCascadic()
    {
        gsMatrix<real_t> result;
        this->cascadicMultiGrid( result );
        return result;
    }

    void smoothingStep(gsMatrix<real_t>& x, const gsMatrix<real_t>& rhs, int level) const;
    void smoothingStep(gsMatrix<real_t>& x, const gsMatrix<real_t>& rhs) const
      { smoothingStep(x, rhs, finestLevel()); }

    void twoGridStep(const gsMatrix<real_t>& rhs_, gsMatrix<real_t>& x) const;

    /// Perform one multigrid cycle on the iterate \a x with right-hand side \a f at the given \a level.
    void multiGridStep(gsMatrix<real_t>& x, const gsMatrix<real_t>& f, int level) const;

    /// Perform one multigrid cycle on the iterate \a x on the finest level.
    void multiGridStep(gsMatrix<real_t>& x, const gsMatrix<real_t>& f) const
      { multiGridStep(x, f, finestLevel()); }

    /*
    /// @brief Apply one multigrid cycle to obtain an approximation to <em>A^-1 rhs</em> and store it in \a result.
    ///
    /// This is useful when using multigrid as a preconditioner for an iterative method like CG or GMRES.
    ///
    /// \param rhs  the right-hand side for which to apply the multigrid preconditioner
    /// \param[out] result  receives the result of the multigrid cycle
    template <class Rhs, class Result>
    void apply(const Eigen::MatrixBase<Rhs>& rhs, Eigen::MatrixBase<Result>& result) const
    {
        gsMatrix<real_t> resultMat(nDofs(), 1);
        resultMat.setZero();

        //if (n_levels == 1)
        //    smoothingStep(resultMat, rhs);          // only smooth
        //else
            multiGridStep(resultMat, rhs, finestLevel());

        result = resultMat;
    }
    */

    /// @brief Apply one multigrid cycle to obtain an approximation to <em>A^-1 rhs</em> and store it in \a result.
    ///
    /// This is useful when using multigrid as a preconditioner for an iterative method like CG or GMRES.
    ///
    /// \param rhs  the right-hand side for which to apply the multigrid preconditioner
    /// \param[out] result  receives the result of the multigrid cycle
    void apply(const gsMatrix<real_t>& rhs, gsMatrix<real_t>& result) const
    {
        GISMO_ASSERT(rhs.cols() == 1, "Implemented only for single column right hand side matrix");

        result.resize(nDofs(), 1);
        result.setZero();

        multiGridStep(result, rhs, finestLevel());
    }


    /// Perform one full mmultigrid cycle and store the resulting solution vector in \a result.
    void fullMultiGrid(gsMatrix<real_t>& result) const;

    void cascadicMultiGrid(gsMatrix<real_t>& result) const;

    /// Restrict a vector of coefficients from the level \a lf, given by \a fine, to the next coarser level.
    void restrictVector(const gsMatrix<real_t>& fine, gsMatrix<real_t>& coarse, int lf) const;

    /// Prolong a vector of coefficients from the level \a lc, given by \a coarse, to the next finer level.
    void prolongVector(const gsMatrix<real_t>& coarse, gsMatrix<real_t>& fine, int lc) const;

    /// Given a fine-grid matrix, compute all coarser levels of the matrix by Galerkin projection.
    void computeMatrixLevels(SpMatrix& M, std::vector< SpMatrix >& Mlevels) const;

    void solveCoarse(const gsMatrix<real_t>& rhs, gsMatrix<real_t>& result) const;

    int numLevels() const    { return n_levels; }               ///< Number of levels in the multigrid construction.

    int finestLevel() const  { return n_levels - 1; }           ///< The index of the finest level (0-based).

    /// Stiffness matrix at given level.
    const SpMatrix& matrix(int lvl) const  { return m_matrix[lvl]; }
    /// Stiffness matrix at finest level.
    const SpMatrix& matrix() const         { return matrix( finestLevel() ); }

    int nDofs(int lvl) const { return matrix(lvl).cols( ); }    ///< Number of dofs at the given level.
    int nDofs()        const { return nDofs( finestLevel() ); }    ///< Number of dofs at the finest level.

    index_t rows() const { return matrix().rows(); }
    index_t cols() const { return matrix().cols(); }

    /// Right-hand side of the lvl-level linear system. Only used for FMG.
    const gsVector<real_t>& rhs(int lvl) const { return *m_rhs[ lvl ]; }
    const gsVector<real_t>& rhs() const { return rhs(finestLevel() ); }

    void setSmoother(int lvl, gsSmoother * sm);
    void setNumPreSmooth(int n)    { m_numPreSmooth = n; }      ///< Set number of pre-smoothing steps to perform.
    void setNumPostSmooth(int n)   { m_numPostSmooth = n; }     ///< Set number of post-smoothing steps to perform.
    void setNumCycles(int n)       { m_numCycles = n; }         ///< Set number of cycles (usually 1 for V-cycle or 2 for W-cycle).

    void testSmoother(const gsMatrix<real_t>& rhs_, int numIt) const;

    double assemblingTime() const  { return m_assemblingTime; } ///< The time in seconds used for assembling the problem.
    double setupTime() const       { return m_setupTime; }      ///< The time in seconds used for the multigrid setup.
    double coarseSolveTime() const { return m_coarseSolveTime; }///< The time in seconds spent in the direct coarse-grid solver.

private:

    void init( SpMatrix& A, const std::vector< gsSparseMatrix<real_t, RowMajor> >& transferMatrices, const std::vector< const gsDofMapper* >& dofMappers );

    // disable copying
    gsMultiGrid( const gsMultiGrid& other );
    gsMultiGrid& operator= (const gsMultiGrid& other);

private:

    // Number of levels
    int n_levels;

    // Stiffness matrix for each grid
    std::vector< SpMatrix > m_matrix;

    // Smoothers on each level
    std::vector< gsSmoother * > m_smoother;

    // Prolongation matrix for each grid transition
    std::vector< SpMatrix > m_prolong;

    // solver for the coarsest-grid problem
    //Eigen::LLT< Eigen::Matrix<real_t,Dynamic,Dynamic> > m_coarseSolver;        // symmetric problems only
    Eigen::PartialPivLU< Eigen::Matrix<real_t,Dynamic,Dynamic> > m_coarseSolver;

    // Right-hand side on each level -- only needed for FMG
    std::vector<gsVector<real_t> *> m_rhs;

    // Strongly imposed dirichlet data -- unused, only needed for FMG
    std::vector<gsMatrix<real_t> > m_fixedDofs;

    // contributions of the boundary values to the prolongation of a full function; needed for FMG
    std::vector< gsMatrix<real_t> > m_dirichletIntp;

    int m_numPreSmooth;
    int m_numPostSmooth;
    int m_numCycles;

    double m_setupTime;
    double m_assemblingTime;
    mutable double m_coarseSolveTime;

}; // class gsMultiGrid



GISMO_EXPORT
void buildDefaultHierarchy(const gsBasis<real_t>& basis0, int lvls, int refineKnots,
                           std::vector<gsBasis<real_t>*> &bases,
                           std::vector< gsSparseMatrix<real_t, RowMajor> >& transferMatrices);


}  // namespace gismo

