/** @file gsSolver.h

    @brief abstract interfaces for solvers and wrapper around Eigen solvers

    This file is part of the G+Smo library.

    This Source Code Form is subject to the terms of the Mozilla Public
    License, v. 2.0. If a copy of the MPL was not distributed with this
    file, You can obtain one at http://mozilla.org/MPL/2.0/.

    Author(s): J. Sogn, A. Bressan
**/

#pragma once

#include <gsSolver/gsMatrixOperator.h>

namespace gismo {


/// @brief Abstract class for solvers.
///
/// \ingroup Solver
class gsSolver
{
public:
    typedef gsSparseMatrix<real_t> MatrixT;
    typedef gsMatrix<real_t>       VectorT;
public:
    virtual ~gsSolver(){}

    virtual gsSolver& compute (const MatrixT &matrix) = 0;
    virtual VectorT solve   (const VectorT &rhs)    = 0;
    virtual bool succeed() = 0;
};


template <typename eigenSolver>
class gsSolverWrapper : public gsSolver, public eigenSolver
{
public:
    gsSolverWrapper()
    {}

    gsSolverWrapper(const MatrixT &matrix)
        : eigenSolver(matrix)
    {}

    gsSolverWrapper<eigenSolver>& compute   (const MatrixT &matrix)
    {
        eigenSolver::compute(matrix);
        return *this;
    }
    VectorT solve  (const VectorT &rhs)
    {
        return eigenSolver::solve(rhs);
    }
    bool succeed()
    {
        return eigenSolver::info()==Eigen::Success;
    }
};

typedef gsSolverWrapper<gsSparseSolver<real_t>::CGIdentity>     gsEigenCGIdentity;
typedef gsSolverWrapper<gsSparseSolver<real_t>::CGDiagonal>     gsEigenCGDiagonal;
typedef gsSolverWrapper<gsSparseSolver<real_t>::BiCGSTABILUT>   gsEigenBiCGSTABILUT;
typedef gsSolverWrapper<gsSparseSolver<real_t>::LU>             gsEigenLU;
typedef gsSolverWrapper<gsSparseSolver<real_t>::QR>             gsEigenQR;
typedef gsSolverWrapper<gsSparseSolver<real_t>::SimplicialLDLT> gsEigenSimplicialLDLT;

}
